Float = Float64

include("utilities.jl")
using Parameters
using LinearAlgebra
using TensorOperations
using NLsolve
using MPI
using PyCall

Cluster = pyimport("scipy.cluster.vq")


"""
    eye(n::Int64)

Constructs the identity matrix with n rows and columns
"""
function eye(n::Int64)
    return diagm(0=>ones(n))
end

#Some helpful types
JacobianDict = Dict{Symbol,Matrix{Float}}
HessianDict = Dict{Tuple{Symbol,Symbol},Array{Float,3}}
AgentDist = Tuple{Matrix{Float},Vector{Float}} #Distribution of agents
GMatrix = Union{Matrix{Float},UniformScaling{Bool}} #Allow for UniformScaling

import Base.convert
convert(::Type{GMatrix},x::Matrix{Float64}) = convert(Matrix{Float},x)

"""
Struct holding all of the data for the approximation
"""
@with_kw mutable struct Approximation{Para}
    #Hold the parameter values
    para::Para
    comm::MPI.Comm = MPI.COMM_WORLD

    #Some variables to store
    N::Int = 0 #number of agents
    ω::Vector{Float} = zeros(Float,0)  #mass of each point
    x̄::Matrix{Float} = zeros(Float,0,0) #steady state values of x
    z̄::Matrix{Float} = zeros(Float,0,0) #steady state values of z
    w̄::Matrix{Float} = zeros(Float,0,0) #steady state values of w
    X̄::Vector{Float} = zeros(Float,0) #steady state values for X

    #
    Ω::AgentDist= (zeros(Float,0,0),zeros(Float,0)) #The full distribution of points
    labels::Vector{Int} = zeros(Int,0)
    my_i::Vector{Int} = zeros(Int,0)


    #derivatives of F and R
    DF::Vector{JacobianDict} = Vector{JacobianDict}()
    HF::Vector{HessianDict} = Vector{HessianDict}()
    DR::Vector{JacobianDict} = Vector{JacobianDict}()
    HR::Vector{HessianDict} = Vector{HessianDict}()

    #derivatives of policy rules
    dx::Vector{JacobianDict} = Vector{JacobianDict}()
    ∂x::Vector{JacobianDict} = Vector{JacobianDict}()
    d2x::Vector{HessianDict} = Vector{HessianDict}()
    dX::JacobianDict = JacobianDict()
    dX′::JacobianDict = JacobianDict() #capture effect on distributions
    d2X::HessianDict = HessianDict()
    d2X′::HessianDict = HessianDict()
    ∂X::Vector{Matrix{Float}} = Vector{Matrix{Float}}()

    Dinv::Matrix{Float} = zeros(Float,0,0)

    #Variables to be used along the path
    K::Int = 0
    C::Vector{Matrix{Float}} = Vector{Matrix{Float}}()
    E::Matrix{Float} = zeros(Float,0,0)
    finalshock::Bool = false
end

"""
    setupApproximate!(approx::Approximation,ω,x̄,X̄)

Creates all the empty variables for the approximate object
"""
function setupApproximation!(approx::Approximation,ω)
    approx.ω = ω
    approx.N = length(ω)
    #Setup and compute derivatives
    approx.DF = [JacobianDict() for i in 1:approx.N]
    approx.HF = [HessianDict() for i in 1:approx.N]
    approx.DR = [JacobianDict() for i in 1:approx.N]
    approx.HR = [HessianDict() for i in 1:approx.N]
    approx.dx = [JacobianDict() for i in 1:approx.N]
    approx.∂x = [JacobianDict() for i in 1:approx.N]
    approx.∂X = [zeros(Float,0,0) for i in 1:approx.N]
    approx.d2x = [HessianDict() for i in 1:approx.N]
    approx.d2X = HessianDict()
    approx.C = [zeros(Float,0,0) for i in 1:approx.N]
    #Compute the derivatives of F
    computeDeriatives!(approx)
end

"""
    approximate(approx::Approximation,Ω::AgentDist)

Approximates policy rules around current distribution of agents
"""
function approximate!(approx::Approximation,Ω::AgentDist)
    para = approx.para
    comm = MPI.COMM_WORLD
    r,s,N = MPI.Comm_rank(comm),MPI.Comm_size(comm),length(Ω[2])
    #Store distribution
    approx.Ω = (Ω[1][MPI_slice(r,s,N),:],Ω[2][MPI_slice(r,s,N)])
    approx.ω = Ω[2][MPI_slice(r,s,N)]
    approx.N = length(approx.ω)

    #Solve for steady state
    approx.x̄,approx.X̄ = para.steadyState(approx.Ω[1],approx.ω)
    setw̄!(approx,approx.x̄,approx.X̄)

    #compute derivatives
    approx.DF = [JacobianDict() for i in 1:approx.N]
    approx.HF = [HessianDict() for i in 1:approx.N]
    approx.DR = [JacobianDict() for i in 1:approx.N]
    approx.HR = [HessianDict() for i in 1:approx.N]
    approx.dx = [JacobianDict() for i in 1:approx.N]
    approx.∂x = [JacobianDict() for i in 1:approx.N]
    approx.∂X = [zeros(Float,0,0) for i in 1:approx.N]
    approx.d2x = [HessianDict() for i in 1:approx.N]
    approx.d2X = HessianDict()
    approx.C = [zeros(Float,0,0) for i in 1:approx.N]
    #Compute the derivatives of F
    computeDeriatives!(approx)

    #Compute Policy Rule Derivatives
    approximate_firstorder!(approx)
    approximate_secondorder!(approx)
end

"""
    approximateDistribution(Ω::AgentDist,k::Int)

Approximates the distribution using the K means algorithm
"""
function approximateDistribution!(comm,Ω::AgentDist,k::Int)
    r = MPI.Comm_rank(comm)
    s = MPI.Comm_size(comm)
    if r == 0
        Ω̄,labels = Cluster.kmeans2(Ω[1],Ω[1][1:k,:],minit="matrix")
        labels = labels .+ 1
        ω̄ = zeros(Float,size(Ω̄,1))
        for i in 1:length(ω̄)
            ω̄[i] = sum(Ω[2][labels .== i])
        end
        Ω̄ = (Ω̄,ω̄)

        Ω̄ = MPI.bcast(Ω̄,0,comm)
        labels = MPI.bcast(labels,0,comm)
    else
        Ω̄ = MPI.bcast(Nothing,0,comm)
        labels = MPI.bcast(Nothing,0,comm)
    end
    N = size(Ω̄[1],1)
    my_labels = MPI_index.(labels,r,s,N)
    mask = my_labels .> 0
    my_labels = my_labels[mask]
    my_Ω = (Ω[1][mask,:],Ω[2][mask])
    my_i = Array(1:size(Ω[1],1))[mask]

    return Ω̄::AgentDist,my_Ω,my_labels,my_i
end


"""
    approximateDistribution(Ω::AgentDist,k::Int)

Approximates the distribution using the K means algorithm
"""
function approximateDistribution_old!(comm,Ω::AgentDist,k::Int)
    r = MPI.Comm_rank(comm)
    s = MPI.Comm_size(comm)
    if r == 0
        cluster = kmeans(Ω[1]',k,weights = Ω[2],init=1:k)
        Ω̄ = (Matrix(cluster.centers'),wcounts(cluster))

        Ω̄ = MPI.bcast(Ω̄,0,comm)
        labels = MPI.bcast(cluster.assignments,0,comm)
    else
        Ω̄ = MPI.bcast(Nothing,0,comm)
        labels = MPI.bcast(Nothing,0,comm)
    end
    N = size(Ω̄[1],1)
    my_labels = MPI_index.(labels,r,s,N)
    mask = my_labels .> 0
    my_labels = my_labels[mask]
    my_Ω = (Ω[1][mask,:],Ω[2][mask])
    my_i = Array(1:size(Ω[1],1))[mask]

    return Ω̄::AgentDist,my_Ω,my_labels,my_i
end

"""
    approximate_firstorder!(approx::Approximation)

Do the first order approximation
"""
function approximate_firstorder!(approx::Approximation)
    compute_dx_z!(approx)
    compute_dx_ϵ!(approx)
    compute_dΛ!(approx)
    compute_∂!(approx)
    compute_dΘ!(approx)
    compute_dΕ!(approx)
end

"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder!(approx::Approximation)
    compute_dx_zz!(approx)
    compute_zΛ!(approx)
    compute_z∂!(approx)
    compute_dx_ϵϵ!(approx)
    compute_zΘ!(approx)
    compute_ϵΘ!(approx)
    compute_dΛΛ!(approx)
    compute_ΛΘ!(approx)
    compute_dx_dΘΘ!(approx)
    compute_dx_d2Ε!(approx)
    compute_σσ!(approx)
end

"""
    simulatepolicies(approx::Approximation, Ω::AgentDist,Θ,Ε,ϵ)

Computes the policy rules for X and for x
"""
function simulatepolicies(approx::Approximation,Θ,Ε,ϵ)
    @unpack x̄,X̄,para,dX,d2X,comm = approx
    @unpack IΘX,nx = para

    Θ̂ = Θ - IΘX*X̄

    #compute X
    X = X̄ + dX[:Θ]*Θ̂ + dX[:Ε]*Ε + 0.5*(d2X[:Θ,:Θ]⋅(Θ̂,Θ̂)+d2X[:Θ,:Ε]⋅(Θ̂,Ε)+d2X[:Ε,:Θ]⋅(Ε,Θ̂)+d2X[:Ε,:Ε]⋅(Ε,Ε)+d2X[:σ,:σ][:])

    #compute x
    N = size(x̄,2)
    my_x = similar(x̄)
    for i in 1:N
        dx,d2x = approx.dx[i],approx.d2x[i]
        ϵ̂ = ϵ[:,i]
        temp = (0.5*d2x[:ϵ,:ϵ]⋅(ϵ̂,ϵ̂) + d2x[:ϵ,:Ε]⋅(ϵ̂,Ε) + d2x[:ϵ,:Θ]⋅(ϵ̂,Θ̂)
                +0.5*d2x[:Ε,:Ε]⋅(Ε,Ε) + d2x[:Ε,:Θ]⋅(Ε,Θ̂) + 0.5*d2x[:Θ,:Θ]⋅(Θ̂,Θ̂)
                + 0.5*d2x[:σ,:σ][:])
        my_x[:,i] .= x̄[:,i] + dx[:ϵ]*ϵ̂ + dx[:Θ]*Θ̂ + dx[:Ε]*Ε + temp
    end
    x = MPI.Gather(my_x,0,comm)
    if MPI.Comm_rank(comm) == 0
        return X,reshape(x,nx,:)
    else
        return X,my_x
    end
end


"""
    simulatepolicies_linear(approx::Approximation, Ω::AgentDist,Θ,Ε,ϵ)

Computes the policy rules for X and for x
"""
function simulatepolicies_linear(approx::Approximation,Θ,Ε,ϵ)
    @unpack x̄,X̄,para,dX,d2X,comm = approx
    @unpack IΘX,nx = para

    Θ̂ = Θ - IΘX*X̄

    #compute X
    X = X̄ + dX[:Θ]*Θ̂ + dX[:Ε]*Ε + 0.5*(d2X[:σ,:σ][:])

    #compute x
    N = size(x̄,2)
    my_x = similar(x̄)
    for i in 1:N
        dx,d2x = approx.dx[i],approx.d2x[i]
        ϵ̂ = ϵ[:,i]
        temp = (0.5*d2x[:ϵ,:ϵ]⋅(ϵ̂,ϵ̂) + d2x[:ϵ,:Ε]⋅(ϵ̂,Ε) + d2x[:ϵ,:Θ]⋅(ϵ̂,Θ̂)
                + 0.5*d2x[:σ,:σ][:])
        my_x[:,i] .= x̄[:,i] + dx[:ϵ]*ϵ̂ + dx[:Θ]*Θ̂ + dx[:Ε]*Ε + temp
    end
    x = MPI.Gather(my_x,0,comm)
    if MPI.Comm_rank(comm) == 0
        return X,reshape(x,nx,:)
    else
        return X,my_x
    end
end



"""
    getw̄(approx::Approximation,x̄,X̄)

Solves for the vector of w̄  given x̄ and X̄
"""
function setw̄!(approx::Approximation,x̄,X̄)
    @unpack para,N = approx
    @unpack Izx,Ix′x,Ix_x,IΘX,IΛX,IX′X,nϵ,nΕ = para

    X̄vec = X̄.*ones(N)'

    approx.w̄ = [Izx*x̄;x̄;Ix_x*x̄;Ix′x*x̄;IΘX*X̄vec;IΛX*X̄vec;X̄vec;IX′X*X̄vec;zeros(Float,nϵ,N);zeros(Float,nΕ,N)]
end

"""
    computeDeriatives(approx)

 Computes derivatives of F and R
"""
function computeDeriatives!(approx::Approximation)
    @unpack para,N,DF,HF,DR,HR,w̄ = approx
    @unpack nx,nX,nw,ind = para

    Fvars = [:z,:x,:x_,:x′,:X,:Θ,:Λ,:ϵ,:Ε] #arguments of F
    Rvars = [:z,:x,:X,:Θ,:Λ,:X′,:ϵ,:Ε]

    DFtemp = zeros(Float,nx,nw)
    HFtemp = zeros(Float,nx,nw,nw)
    DRtemp = zeros(Float,nX,nw)
    HRtemp = zeros(Float,nX,nw,nw)

    for i in 1:N

        #compute derivatives w.r.t F
        para.DF(DFtemp,w̄[:,i])
        para.HF(HFtemp,w̄[:,i])
        for var1 in Fvars
            DF[i][var1] = DFtemp[:,ind[var1]]
            for var2 in Fvars
                HF[i][var1,var2] = HFtemp[:,ind[var1],ind[var2]]
            end
        end

        #compute derivatives w.r.t R
        para.DR(DRtemp,w̄[:,i])
        para.HR(HRtemp,w̄[:,i])
        for var1 in Rvars
            DR[i][var1] = DRtemp[:,ind[var1]]
            for var2 in Rvars
                HR[i][var1,var2] = HRtemp[:,ind[var1],ind[var2]]
            end
        end

    end
end

"""
    compute_d_Λ_alt!(approx)

Computes the derivative with respect to Lambda by solving a non-linear equation
in X̄
"""
function compute_dΛ!(approx,T=15)
    @unpack dx,N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,IX′X,IΛX,nx,nX,nΛ = para

    aT,a,bT,c = zeros(Float,nx,nX,N),zeros(Float,nx,nX,N),zeros(Float,nx,nΛ,N),zeros(Float,nx,nΛ,N)
    AT,A,BT = zeros(Float,nX,nX),zeros(Float,nX,nX),zeros(Float,nX,nΛ)
    temp = zeros(Float,nx,nx,N)
    for i in 1:N
        DF,DR = approx.DF[i],approx.DR[i]
        temp[:,:,i] .= inv(DF[:x_]*Ix_x+DF[:x])
        aT[:,:,i] .= -temp[:,:,i]*DF[:X]
        bT[:,:,i] .= -temp[:,:,i]*DF[:Λ]

        AT .+= ω[i]*(DR[:x]*aT[:,:,i] + DR[:X])
        BT .+= ω[i]*(DR[:x]*bT[:,:,i] + DR[:Λ])
    end
    MPI.Allreduce!(AT,+,comm)
    MPI.Allreduce!(BT,+,comm)
    C = -AT\BT
    C_ = copy(C)
    diff = 1
    while diff > 1e-10
        A .= 0.
        for i in 1:N
            DF,DR = approx.DF[i],approx.DR[i]
            c[:,:,i] .= (aT[:,:,i].+a[:,:,i])*C + bT[:,:,i]
            a[:,:,i] .= -temp[:,:,i]*DF[:x′]*Ix′x*c[:,:,i]*IΛX

            A .+= ω[i]*(DR[:x]*a[:,:,i]+DR[:X′]*IX′X*C*IΛX)
        end
        MPI.Allreduce!(A,+,comm)
        C .= -(AT+A)\BT
        diff = norm(C-C_,Inf)
        C_ .= C
    end

    approx.dX[:Λ] = C
    for i in 1:N
        approx.dx[i][:Λ] = (aT[:,:,i].+a[:,:,i])*C + bT[:,:,i]
    end
    Λ̄_Λ = IΛX*C
end


"""
    compute_dx_Λ!(approx,X̄_Λ)

Computes the derivatives of x w.r.t Λ for a given value of X̄_Λ
"""
function compute_dx_dΛ!(approx,X̄_Λ)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,IΛX = para
    Λ̄_Λ = IΛX*X̄_Λ

    for i in 1:N
        DF = approx.DF[i]
        A = convert(Matrix{Float64},inv(DF[:x_]*Ix_x+DF[:x])*DF[:x′]*Ix′x)
        B = convert(Matrix{Float64},inv(Λ̄_Λ))
        C = convert(Matrix{Float64},inv(DF[:x_]*Ix_x+DF[:x])*(DF[:Λ]+DF[:X]*X̄_Λ)*inv(Λ̄_Λ))

        approx.dx[i][:Λ] = sylvester(A,B,C)
    end
end

"""
    compute_d_Λ!(approx)

Computes the derivative with respect to Lambda by solving a non-linear equation
in X̄
"""
function compute_dΛ_alt!(approx)
    @unpack dx,N,para,ω,comm = approx
    @unpack Ix_x,IX′X,Izx,IΛX,nX,nΛ,X̄_Λ0 = para
    r = MPI.Comm_rank(comm)
    function res!(F,X̄_Λ)
        Λ̄_Λ = IΛX*X̄_Λ
        compute_dx_dΛ!(approx,X̄_Λ)

        F[:] .= 0.
        for i in 1:N
            DR = approx.DR[i]
            dx = approx.dx[i]
            F .+= ω[i]*(DR[:x]*dx[:Λ] + DR[:X]*X̄_Λ + DR[:X′]*IX′X*X̄_Λ*Λ̄_Λ + DR[:Λ])
        end
        MPI.Allreduce!(F,+,comm)
    end

    #Now find the root
    while true
        MPI.Bcast!(X̄_Λ0,0,comm)
        result = nlsolve(res!,convert(Matrix{Float64},X̄_Λ0)) #make sure to use the same starting guess

        if result.f_converged || result.x_converged
            X̄_Λ = result.zero
            Λ̄_Λ = IΛX*X̄_Λ
            if r == 0
                println(abs.(eigen(Λ̄_Λ).values))
            end
            if all(abs.(eigen(Λ̄_Λ).values) .< 1+1e-5)
                approx.dX[:Λ] = X̄_Λ
                para.X̄_Λ0 = X̄_Λ
                break
            end
        else
            if r == 0
                println("Root Failed")
            end
        end
        X̄_Λ0 .= rand(nX,nΛ)
    end

    compute_dx_dΛ!(approx,approx.dX[:Λ])
end # function

"""
    compute_dy_z!(approx)

Computes the derivatives of x w.r.t. z
"""
function compute_dx_z!(approx)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,Izx,nz = para

    for i in 1:N
        DF = approx.DF[i]
        temp = convert(Matrix{Float64},[(DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x);Izx]) #last term imposes that x_z = 1
        approx.dx[i][:z] = pinv(temp)*[-DF[:z];diagm(0=>ones(nz))]
    end
end

"""
    compute_dy_ϵ!(approx)

Computes the derivatives of x w.r.t. ϵ
"""
function compute_dx_ϵ!(approx)
    @unpack N,para = approx
    @unpack Ix_x,Ix′x,Izx,Fmeasure = para
    nF_ = size(Fmeasure,1) #number of variables
    for i in 1:N
        DF = approx.DF[i]
        dx = approx.dx[i]
        temp = DF[:x]+DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] .= Fmeasure

        dx[:ϵ] = -temp\DF[:ϵ] #first nF_ equations won't depend on epsilon
    end
end



"""
    compute_∂!(approx)

Compute the Frechet derivative of x and X
"""
function compute_∂!(approx)
    @unpack N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nX = para

    #Compute C(z) from the appendix
    for i in 1:N
        DF = approx.DF[i]
        dx = approx.dx[i]
        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x+DF[:x′]*Ix′x*dx[:z]*Izx

        approx.∂x[i][:X] = -temp\(DF[:X]+DF[:x′]*Ix′x*dx[:Λ]*IΛX)
    end

    #compute D from the appendix
    D = zeros(Float,para.nX,para.nX)
    for i in 1:N
        DR = approx.DR[i]
        C = approx.∂x[i][:X]
        D .+= ω[i]*(DR[:x]*C + DR[:X′]*IX′X*(I + approx.dX[:Λ]*IΛX) + DR[:X])
        approx.C[i] = C
    end
    MPI.Allreduce!(D,+,comm)

    approx.Dinv = inv(D)
    #compute A(z) from the appendix
    for i in 1:N
        DR = approx.DR[i]
        dx = approx.dx[i]

        approx.∂X[i] = -approx.Dinv*(DR[:z] + DR[:x]*dx[:z])
    end
end

"""
    compute_dΘ!(approx::Approximation)

Computes the derivative w.r.t the aggregate state Θ, assume that ρ_Θ is diagonal
"""
function compute_dΘ!(approx::Approximation)
    @unpack N,para,ω,dX,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΘ,nx,nX = para

    x_Θ = zeros(Float,N,nx,nΘ)
    x_XΘ = zeros(Float,N,nx,nX,nΘ)
    x_X′Θ = zeros(Float,N,nx,nX,nΘ)

    #K and J from dΕ note
    K1 = zeros(Float,nX,2*nX,nΘ)
    K2 = zeros(Float,nX,2*nX,nΘ)
    J1 = zeros(Float,nX,nΘ)
    J2 = zeros(Float,nX,nΘ)

    for i in 1:N
        DF,DR,dx,∂x,∂X = approx.DF[i],approx.DR[i],approx.dx[i],approx.∂x[i],approx.∂X[i]

        for j in 1:nΘ
            ρ_θ = para.ρ_Θ[j,j] #compute for each $ρ_Θ#
            temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*(dx[:z]*Izx+ρ_θ*I))
            #compute coefficients on X
            x_Θ[i,:,j] .= -temp*DF[:Θ][:,j]
            x_XΘ[i,:,:,j] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
            x_X′Θ[i,:,:,j] .= -temp*DF[:x′]*Ix′x*∂x[:X]

            #Compute K and J
            K1[:,1:nX,j] .+= ω[i].*(DR[:x]*x_XΘ[i,:,:,j] + DR[:X] + ρ_θ*DR[:X′]*IX′X + DR[:X′]*IX′X*dX[:Λ]*IΛX)
            K1[:,nX+1:end,j] .+= ω[i].*(DR[:x]*x_X′Θ[i,:,:,j] + DR[:X′]*IX′X)
            K2[:,1:nX,j] .+=  ω[i].*(-∂X*Izx*x_XΘ[i,:,:,j])
            K2[:,nX+1:end,j] .+=  ω[i].*(I-∂X*Izx*x_X′Θ[i,:,:,j])
            J1[:,j] .+= -ω[i].*(DR[:x]*x_Θ[i,:,j] + DR[:Θ][:,j] )
            J2[:,j] .+=  ω[i].*(∂X*Izx*x_Θ[i,:,j])
        end
    end
    MPI.Allreduce!(K1,+,comm)
    MPI.Allreduce!(K2,+,comm)
    MPI.Allreduce!(J1,+,comm)
    MPI.Allreduce!(J2,+,comm)

    X_Θ = zeros(Float,nX,nΘ)
    X′_Θ = zeros(Float,nX,nΘ)
    for j in 1:nΘ
        XX′_Θ = [K1[:,:,j] ; K2[:,:,j]] \ [J1[:,j] ; J2[:,j]]
        X_Θ[:,j] .= XX′_Θ[1:nX]
        X′_Θ[:,j] .= XX′_Θ[nX+1:end]
    end
    approx.dX[:Θ] = X_Θ
    approx.dX[:Ω_Θ] = X′_Θ
    approx.dX′[:Θ] = X′_Θ
    #store results
    for i in 1:N
        approx.dx[i][:Θ] = zeros(Float,nx,nΘ)
        for j in 1:nΘ
            approx.dx[i][:Θ][:,j] = x_Θ[i,:,j] + x_XΘ[i,:,:,j]*X_Θ[:,j] + x_X′Θ[i,:,:,j]*X′_Θ[:,j]
        end
    end
end


"""
    compute_dΕ!(approx::Approximation)

Computes the derivative w.r.t the aggregate state Ε
"""
function compute_dΕ!(approx::Approximation)
    @unpack N,para,ω,dX,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΕ,nx,nX,ρ_Θ,Fmeasure,Gmeasure = para
    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    x_Ε = zeros(Float,N,nx,nΕ)
    x_XΕ = zeros(Float,N,nx,nX)
    x_X′Ε = zeros(Float,N,nx,nX)

    #K and J from dΕ note
    K1 = zeros(Float,nX,2*nX)
    K2 = zeros(Float,nX,2*nX)
    J1 = zeros(Float,nX,nΕ)
    J2 = zeros(Float,nX,nΕ)

    for i in 1:N
        DF,DR,dx,dX,∂x,∂X = approx.DF[i],approx.DR[i],approx.dx[i],approx.dX, approx.∂x[i], approx.∂X[i]

        temp = DF[:x]+DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] .= Fmeasure #apply the measurablility restriction on subset of the variables
        temp = inv(temp)
        #compute coefficients on X
        x_Ε[i,:,:] .= -temp*(DF[:Ε]+DF[:x′]*Ix′x*dx[:Θ])
        x_XΕ[i,:,:] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
        x_X′Ε[i,:,:] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        #Compute K and J
        K1[:,1:nX] .+= ω[i].*(DR[:x]*x_XΕ[i,:,:] + DR[:X] + DR[:X′]*IX′X*dX[:Λ]*IΛX)
        K1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′Ε[i,:,:] + DR[:X′]*IX′X)
        K2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_XΕ[i,:,:])
        K2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′Ε[i,:,:])
        J1 .+= -ω[i].*(DR[:x]*x_Ε[i,:,:] + DR[:Ε] + DR[:X′]*IX′X*dX[:Θ])
        J2 .+= ω[i].*(∂X*Izx*x_Ε[i,:,:])
    end
    MPI.Allreduce!(K1,+,comm)
    MPI.Allreduce!(K2,+,comm)
    MPI.Allreduce!(J1,+,comm)
    MPI.Allreduce!(J2,+,comm)
    #apply the measurablility restriction on X variables
    K1[1:nG_,1:nX] .= Gmeasure
    XX′_Ε = [K1 ; K2] \ [J1 ; J2]
    X_Ε = XX′_Ε[1:nX,:]
    X′_Ε = XX′_Ε[nX+1:end,:]

    approx.dX[:Ε] = X_Ε
    approx.dX[:Ω_Ε] = X′_Ε
    approx.dX′[:Ε] = X′_Ε
    #store results
    for i in 1:N
        approx.dx[i][:Ε] = x_Ε[i,:,:] + x_XΕ[i,:,:]*X_Ε + x_X′Ε[i,:,:]*X′_Ε
    end
end


"""
    compute_dy_zz!(approx)

Computes the second of x w.r.t. zz
"""
function compute_dx_zz!(approx)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,Izx,nx,nz = para

    d = Dict{Symbol,GMatrix}()
    F_zz = zeros(Float,nx,nz,nz)
    for i in 1:N
        DF = approx.DF[i]
        HF = approx.HF[i]
        dx = approx.dx[i]

        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x+DF[:x′]*Ix′x*dx[:z]*Izx

        #compute the sum of all the second order terms
        d[:z],d[:x_],d[:x],d[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx[:z]
        F_zz .= 0
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:z,:x_,:x,:x′]
                @doinplace F_zz .+= HF[x1,x2]⋅(d[x1],d[x2])
            end
        end

        approx.d2x[i][:z,:z] = -inv(temp)*F_zz


    end
end


"""
    compute_zΛ!(approx::Approximation)

Computes the cross partial with z and Λ
"""
function compute_zΛ!(approx::Approximation)
    @unpack N,para,dX,dX′ = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nz,nX,nΛ = para

    #Initialize vairables
    d_Λ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    F_zΛ = zeros(Float,nx,nz,nΛ)
    Λ̄_Λ = IΛX*dX[:Λ]
    Λ̄_Λ_values = convert(Vector{Float},eigen(Λ̄_Λ).values)
    Q = convert(Matrix{Float},eigen(Λ̄_Λ).vectors)
    invQ =  inv(Q)
    dx_zΛ = zeros(Float,nx,nz,nΛ)

    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]



        #compute the sum of all the second order terms
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx[:z]
        d_Λ[:Λ],d_Λ[:X],d_Λ[:x_],d_Λ[:x] = I,dX[:Λ],Ix_x*dx[:Λ],dx[:Λ]
        d_Λ[:x′] = Ix′x*(dx[:Λ]*Λ̄_Λ)

        F_zΛ .= 0.
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_zΛ .+= HF[x1,x2]⋅(d_z[x1],d_Λ[x2])
            end
        end
        F_zΛ .= F_zΛ⋅(I,Q)

        dx_zΛ .= 0.
        for j in 1:nΛ
            ρ_Λ = Λ̄_Λ_values[j]#persistence of jth element of Λ
            temp = DF[:x_]*Ix_x + DF[:x] + ρ_Λ*DF[:x′]*Ix′x

            dx_zΛ[:,:,j] .= -inv(temp)*F_zΛ[:,:,j]
        end

        approx.d2x[i][:z,:Λ] = dx_zΛ⋅(I,invQ)
        approx.d2x[i][:Λ,:z] = permutedims(approx.d2x[i][:z,:Λ],[1,3,2])
    end
end


"""
    compute_z∂!(approx::Approximation)

Computes the cross partial with the Frechet derivative.
"""
function compute_z∂!(approx::Approximation)
    @unpack dx,∂x,N,para = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nz,nX = para

    d_∂X = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]

        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x+DF[:x′]*Ix′x*dx[:z]*Izx

        #compute the sum of all the second order terms
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx[:z]
        d_∂X[:X],d_∂X[:x_],d_∂X[:x],d_∂X[:x′] = I,Ix_x*∂x[:X],∂x[:X],Ix′x*∂x[:X]

        F_z∂X = DF[:x′]*Ix′x*d2x[:z,:Λ]⋅(I,IΛX)
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:X,:x_,:x,:x′]
                @doinplace F_z∂X .+= HF[x1,x2]⋅(d_z[x1],d_∂X[x2])
            end
        end

        approx.d2x[i][:z,:∂X] = -inv(temp)*F_z∂X
        approx.d2x[i][:∂X,:z] = permutedims(approx.d2x[i][:z,:∂X],[1,3,2])
    end
end

"""
    compute_z∂!(approx::Approximation)

Computes the cross partial with the Frechet derivative.
"""
function compute_z∂!(approx::Approximation,Δ,sΔ)
    if haskey(approx.dx[1],(:z,sΔ))
        return
    end
    for i in 1:approx.N
        approx.d2x[i][:z,sΔ] = approx.d2x[i][:z,:∂X]⋅(I,approx.dX[sΔ])
        approx.d2x[i][sΔ,:z] = permutedims(approx.d2x[i][:z,sΔ],[1,3,2])
    end
end

"""
    compute_zΘ!(approx::Approximation)

Computes the cross partial with z and Θ
"""
function compute_zΘ!(approx::Approximation)
    @unpack N,para,dX,dX′ = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,nx,nz,nX,nΘ = para

    Λ_Θ = IΛX*dX[:Θ]
    d_Θ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]



        #compute the sum of all the second order terms
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx[:z]
        d_Θ[:Θ],d_Θ[:X],d_Θ[:x_],d_Θ[:x] = I,dX[:Θ],Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*Izx*dx[:Θ]+dx[:Θ]*ρ_Θ+∂x[:X]*dX′[:Θ]+dx[:Λ]*Λ_Θ)

        F_zΘ = DF[:x′]*Ix′x*( d2x[:z,:z]⋅(I,Izx*dx[:Θ]) + d2x[:z,:∂X]⋅(I,dX′[:Θ]) + d2x[:z,:Λ]⋅(I,Λ_Θ) )
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_zΘ .+= HF[x1,x2]⋅(d_z[x1],d_Θ[x2])
            end
        end

        approx.d2x[i][:z,:Θ] = zeros(Float,nx,nz,nΘ)
        for j in 1:nΘ
            ρ_θ = ρ_Θ[j,j]#persistence of jth element of Θ
            temp = DF[:x_]*Ix_x +DF[:x] + ρ_θ*DF[:x′]*Ix′x + DF[:x′]*Ix′x*dx[:z]*Izx

            approx.d2x[i][:z,:Θ][:,:,j] = -inv(temp)*F_zΘ[:,:,j]
        end

        approx.d2x[i][:Θ,:z] = permutedims(approx.d2x[i][:z,:Θ],[1,3,2])
    end
end


"""
    compute_ϵΛ!(approx::Approximation)

Computes the cross partial with ϵ and Λ
"""
function compute_ϵΛ!(approx::Approximation)
    @unpack N,para,dX,dX′ = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nϵ,nX,nΛ,Fmeasure = para
    nF_ = size(Fmeasure,1)
    #Initialize vairables
    d_Λ = Dict{Symbol,GMatrix}()
    d_ϵ = Dict{Symbol,GMatrix}()
    Λ̄_Λ = IΛX*dX[:Λ]
    dx_ϵΛ = zeros(Float,nx,nϵ,nΛ)

    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]

        #compute the sum of all the second order terms
        z_ϵ = Izx*dx[:ϵ]
        d_ϵ[:ϵ],d_ϵ[:x],d_ϵ[:x′] = I,dx[:ϵ],Ix′x*dx[:ϵ]*z_ϵ
        d_Λ[:Λ],d_Λ[:X],d_Λ[:x_],d_Λ[:x] = I,dX[:Λ],Ix_x*dx[:Λ],dx[:Λ]
        d_Λ[:x′] = Ix′x*(dx[:Λ]*Λ̄_Λ)

        F_ϵΛ = DF[:x′]*Ix′x*d2x[:z,:Λ]⋅(z_ϵ,Λ̄_Λ)
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ϵΛ .+= HF[x1,x2]⋅(d_ϵ[x1],d_Λ[x2])
            end
        end

        temp = DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] .= Fmeasure
        dx_ϵΛ = -inv(temp)*F_ϵΛ

        approx.d2x[i][:ϵ,:Λ] = dx_ϵΛ
        approx.d2x[i][:Λ,:ϵ] = permutedims(approx.d2x[i][:ϵ,:Λ],[1,3,2])
    end
end



"""
    compute_ϵΘ!(approx::Approximation)

Computes the cross partial with ϵ
"""
function compute_ϵΘ!(approx::Approximation)
    @unpack N,para,dX,dX′ = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,nX,nx,nϵ,nΘ,nΕ,Fmeasure = para
    #setup some useful variables
    nF_ = size(Fmeasure,1)
    Λ_Θ,Λ_Ε = IΛX*dX[:Θ],IΛX*dX[:Ε]


    d_Θ = Dict{Symbol,GMatrix}()
    d_Ε = Dict{Symbol,GMatrix}()
    d_ϵ = Dict{Symbol,GMatrix}()
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]


        dz_ϵ = Izx*dx[:ϵ]
        #compute the sum of all the second order terms
        d_ϵ[:ϵ],d_ϵ[:x],d_ϵ[:x′] = I,dx[:ϵ],Ix′x*dx[:z]*dz_ϵ
        d_Θ[:Θ],d_Θ[:X],d_Θ[:x_],d_Θ[:x] = I,dX[:Θ],Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*Izx*dx[:Θ]+dx[:Θ]*ρ_Θ+∂x[:X]*dX′[:Θ]+dx[:Λ]*Λ_Θ)
        d_Ε[:Ε],d_Ε[:X],d_Ε[:x] = I,dX[:Ε],dx[:Ε]
        d_Ε[:x′] = Ix′x*(dx[:z]*Izx*dx[:Ε] + dx[:Θ] +∂x[:X]*dX′[:Ε]+dx[:Λ]*Λ_Ε)

        F_ϵΘ = DF[:x′]*Ix′x*( d2x[:z,:z]⋅(dz_ϵ,Izx*dx[:Θ]) + d2x[:z,:Θ]⋅(dz_ϵ,ρ_Θ) + d2x[:z,:∂X]⋅(dz_ϵ,dX′[:Θ]) + d2x[:z,:Λ]⋅(dz_ϵ,Λ_Θ) )
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ϵΘ .+= HF[x1,x2]⋅(d_ϵ[x1],d_Θ[x2])
            end
        end

        F_ϵΕ = DF[:x′]*Ix′x*( d2x[:z,:z]⋅(dz_ϵ,Izx*dx[:Ε]) + d2x[:z,:Θ]⋅(dz_ϵ,I) + d2x[:z,:∂X]⋅(dz_ϵ,dX′[:Ε]) + d2x[:z,:Λ]⋅(dz_ϵ,Λ_Ε))
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ϵΕ .+= HF[x1,x2]⋅(d_ϵ[x1],d_Ε[x2])
            end
        end

        temp = DF[:x] + DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] = Fmeasure #imposes measurablility restriction
        temp = inv(temp)

        approx.d2x[i][:ϵ,:Θ] = -temp*F_ϵΘ
        approx.d2x[i][:ϵ,:Ε] = -temp*F_ϵΕ
        approx.d2x[i][:Θ,:ϵ] = permutedims(approx.d2x[i][:ϵ,:Θ],[1,3,2])
        approx.d2x[i][:Ε,:ϵ] = permutedims(approx.d2x[i][:ϵ,:Ε],[1,3,2])
    end
end


"""
    compute_dy_ϵϵ!(approx)

Computes the second of x w.r.t. ϵϵ
"""
function compute_dx_ϵϵ!(approx::Approximation)
    @unpack dx,N,para = approx
    @unpack Ix′x,Izx,nx,nϵ,Fmeasure = para
    nF_ = size(Fmeasure,1)
    d = Dict{Symbol,GMatrix}()
    d[:ϵ] = I
    for i in 1:N
        DF = approx.DF[i]
        HF = approx.HF[i]
        dx = approx.dx[i]
        d2x = approx.d2x[i]

        temp = DF[:x]+DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] = Fmeasure #impose measurablility

        #compute the sum of all the second order terms
        d[:x],d[:x′] = dx[:ϵ],Ix′x*dx[:z]*Izx*dx[:ϵ]
        F_ϵϵ = (DF[:x′]*Ix′x*d2x[:z,:z])⋅(Izx*dx[:ϵ],Izx*dx[:ϵ])
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:ϵ,:x,:x′]
                @doinplace F_ϵϵ .+= HF[x1,x2]⋅(d[x1],d[x2])
            end
        end

        approx.d2x[i][:ϵ,:ϵ] = -inv(temp)*F_ϵϵ
    end
end

"""
    compute_dΛΛ!(approx::Approximation)

Computes the second order derivative with respect to Λ
"""
function compute_dΛΛ!(approx::Approximation)
    @unpack N,para,ω,dX,comm =approx
    @unpack Ix_x,Ix′x,IX′X,IΛX,nx,nΛ,nX = para

    #Prep work
    Λ̄_Λ = IΛX*dX[:Λ]
    Λ̄_Λ_values = convert(Vector{Float},eigen(Λ̄_Λ).values)
    Q = convert(Matrix{Float},eigen(Λ̄_Λ).vectors)
    invQ = inv(Q)

    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX[:Λ]*Λ̄_Λ

    x_ΛΛ = zeros(Float,nx,nΛ,nΛ,N)
    x_ΛΛX = zeros(Float,nx,nX,nΛ,nΛ,N)
    F_ΛΛ = zeros(Float,nx,nΛ,nΛ)
    R_ΛΛ = zeros(Float,nX,nΛ,nΛ)
    A = zeros(Float,nX,nX,nΛ,nΛ)
    B = zeros(Float,nX,nΛ,nΛ)
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]

        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*dx[:Λ]*Λ̄_Λ

        F_ΛΛ .= 0
        for x1 in [:Λ,:X,:x_,:x,:x′]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ΛΛ .+= HF[x1,x2]⋅(d_Λ[x1],d_Λ[x2])
            end
        end
        #Now R
        R_ΛΛ .= 0
        for x1 in [:Λ,:x,:X,:X′]
            for x2 in [:Λ,:x,:X,:X′]
                @doinplace R_ΛΛ .+= HR[x1,x2]⋅(d_Λ[x1],d_Λ[x2])
            end
        end

        #convert to eigenspace
        F_ΛΛ = F_ΛΛ⋅(Q,Q)
        for j2 in 1:nΛ
            for j1 in 1:nΛ
                ρ_ΛΛ = Λ̄_Λ_values[j1]*Λ̄_Λ_values[j2]
                temp = inv(DF[:x_]*Ix_x + DF[:x] + ρ_ΛΛ*DF[:x′]*Ix′x )
                x_ΛΛ[:,j1,j2,i] .= -temp*F_ΛΛ[:,j1,j2]
                x_ΛΛX[:,:,j1,j2,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)

                A[:,:,j1,j2] .+= ω[i]*(DR[:x]*x_ΛΛX[:,:,j1,j2,i] + DR[:X] + DR[:X′]*IX′X*(ρ_ΛΛ*I + dX[:Λ]*IΛX))
            end
        end
        B .+= ω[i]*(DR[:x]*x_ΛΛ[:,:,:,i]+ R_ΛΛ⋅(Q,Q))
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    #Now solve
    X_ΛΛ = zeros(Float,nX,nΛ,nΛ)
    for j2 in 1:nΛ
        for j1 in 1:nΛ
            X_ΛΛ[:,j1,j2] = -inv(A[:,:,j1,j2])* B[:,j1,j2]
        end
    end

    for i in 1:N
        for j2 in 1:nΛ
            for j1 in 1:nΛ
                x_ΛΛ[:,j1,j2,i] .+= x_ΛΛX[:,:,j1,j2,i]*X_ΛΛ[:,j1,j2]
            end
        end

        approx.d2x[i][:Λ,:Λ] = x_ΛΛ[:,:,:,i]⋅(invQ,invQ)
    end
    approx.d2X[:Λ,:Λ] = X_ΛΛ⋅(invQ,invQ)
end


"""
    compute_∂X_Λ!(approx::Approximation,Δ,sΔ)

Computes the cross frechet derivative and Λ in the direction of Δ
"""
function compute_∂X_Λ!(approx::Approximation,Δ,sΔ)
    if haskey(approx.d2X,(:Λ,sΔ)) #check to see if we've already computed this direction
        return
    end
    @unpack N,para,ω,dX,d2X,comm =approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΛ,nz,nΘ,nX = para

    Λ̄_Λ = IΛX*dX[:Λ]
    Λ̄_Λ_values = convert(Vector{Float},eigen(Λ̄_Λ).values)
    Q = convert(Matrix{Float},eigen(Λ̄_Λ).vectors)
    invQ = inv(Q)

    nΔ = size(Δ(1),2)
    X_Δ = dX[sΔ]
    Λ_Δ = IΛX*X_Δ
    #Now compute the derivatives
    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX[:Λ]*Λ̄_Λ
    d_Δ = Dict{Symbol,GMatrix}()
    d_Δ[:X],d_Δ[:X′] = X_Δ,IX′X*(X_Δ+dX[:Λ]*Λ_Δ)


    R_ΔΛ = zeros(Float,nX,nΔ,nΛ)
    x_ΔΛ = zeros(Float,nx,nΔ,nΛ,N)
    x_ΔΛX = zeros(Float,nx,nX,nΛ,N)
    C = zeros(Float,nX,nX,nΛ) #matrix used to solve for X_ΔΛ
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]

        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*dx[:Λ]*Λ̄_Λ
        d_Δ[:x_],d_Δ[:x] = Ix_x*∂x[:X]*X_Δ,∂x[:X]*X_Δ
        d_Δ[:x′] = Ix′x*(∂x[:X]*X_Δ + dx[:Λ]*Λ_Δ)
        #First do the derivatives w.r.t F
        F_ΔΛ = DF[:x′]*Ix′x*( d2x[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ΔΛ .+= HF[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        #Now R
        d_Δ[:z] = Δ(i)
        R_ΔΛ_temp = DR[:X′]*IX′X*d2X[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) + DR[:x]*d2x[:z,:Λ]⋅(Δ(i),I)

        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Λ,:x,:X,:X′]
                @doinplace R_ΔΛ_temp .+= HR[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        #convert into eigenvector space

        F_ΔΛ = F_ΔΛ⋅(I,Q)
        for j in 1:nΛ
            ρ_Λ = Λ̄_Λ_values[j]
            temp = inv(DF[:x_]*Ix_x + DF[:x] + ρ_Λ*DF[:x′]*Ix′x )
            x_ΔΛ[:,:,j,i] = -temp*F_ΔΛ[:,:,j]
            x_ΔΛX[:,:,j,i] = -temp*(DF[:X]+DF[:x′]*Ix′x*dx[:Λ]*IΛX)

            C[:,:,j] .+= ω[i]*(DR[:x]*x_ΔΛX[:,:,j,i] + DR[:X] + DR[:X′]*IX′X*(ρ_Λ*I + dX[:Λ]*IΛX))
        end
        R_ΔΛ .+= ω[i]*(R_ΔΛ_temp⋅(I,Q) + DR[:x]*x_ΔΛ[:,:,:,i])
    end
    MPI.Allreduce!(R_ΔΛ,+,comm)
    MPI.Allreduce!(C,+,comm)
    #Now solve
    X_ΔΛ = zeros(Float,nX,nΔ,nΛ)
    for j in 1:nΛ
        X_ΔΛ[:,:,j] = -inv(C[:,:,j])*R_ΔΛ[:,:,j]
    end
    #now compute for all agents
    for i in 1:N
        for j in 1:nΛ
            x_ΔΛ[:,:,j,i] .+= x_ΔΛX[:,:,j,i]*X_ΔΛ[:,:,j]
        end
        #convert back into regular space
        approx.d2x[i][sΔ,:Λ] = x_ΔΛ[:,:,:,i]⋅(I,invQ)
        approx.d2x[i][:Λ,sΔ] = permutedims(approx.d2x[i][sΔ,:Λ],[1,3,2])
    end
    approx.d2X[sΔ,:Λ] = X_ΔΛ⋅(I,invQ)
    approx.d2X[:Λ,sΔ] = permutedims(approx.d2X[sΔ,:Λ],[1,3,2])
end


"""
    compute_∂∂!(approx::Approximation,Δ1,sΔ1,Δ2,sΔ2)

Computes the second Frechet derivative in direction Δ1,Δ2
"""
function compute_∂∂!(approx::Approximation,Δ1,sΔ1,Δ2,sΔ2)
    if haskey(approx.d2X,(sΔ1,sΔ2)) #check to see if we've already computed this direction
        return
    end
    @unpack dX,d2X,N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nX = para
    nΔ1 = size(Δ1(1),2)
    nΔ2 = size(Δ2(1),2)
    #First compute cross derivatives w.r.t Λ
    compute_∂X_Λ!(approx,Δ1,sΔ1)
    compute_∂X_Λ!(approx,Δ2,sΔ2)
    X_Δ1,X_Δ2 = dX[sΔ1],dX[sΔ2]
    Λ_Δ1,Λ_Δ2 = IΛX*X_Δ1,IΛX*X_Δ2

    dΔ1 = Dict{Symbol,GMatrix}()
    dΔ1[:X],dΔ1[:X′] = X_Δ1, IX′X*(X_Δ1+dX[:Λ]*Λ_Δ1)
    dΔ2 = Dict{Symbol,GMatrix}()
    dΔ2[:X],dΔ2[:X′] = X_Δ2, IX′X*(X_Δ2+dX[:Λ]*Λ_Δ2)


    R_ΔΔ = zeros(Float,nX,nΔ1,nΔ2)
    x_ΔΔ = zeros(Float,nx,nΔ1,nΔ2,N)
    x_ΔΔX = zeros(Float,nx,nX,N)

    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        temp = inv(DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(dx[:z]*Izx + I))

        dΔ1[:x_],dΔ1[:x],dΔ1[:x′] = Ix_x*∂x[:X]*dΔ1[:X],∂x[:X]*dΔ1[:X],Ix′x*(∂x[:X] + dx[:Λ]*IΛX)*X_Δ1
        dΔ2[:x_],dΔ2[:x],dΔ2[:x′] = Ix_x*∂x[:X]*dΔ2[:X],∂x[:X]*dΔ2[:X],Ix′x*(∂x[:X] + dx[:Λ]*IΛX)*X_Δ2
        F_ΔΔ = DF[:x′]*Ix′x*(d2x[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2x[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2x[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:x_,:x,:x′,:X]
                @doinplace F_ΔΔ .+= HF[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end
        x_ΔΔ[:,:,:,i] = -temp*F_ΔΔ
        x_ΔΔX[:,:,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)

        #Now the terms coming from the R equations
        dΔ1[:z],dΔ2[:z] = Δ1(i),Δ2(i)
        R_ΔΔ_temp = DR[:X′]*IX′X*(d2X[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2X[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2X[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2))
        R_ΔΔ_temp .+= DR[:x]*(d2x[:z,:∂X]⋅(dΔ1[:z],X_Δ2) + d2x[:∂X,:z]⋅(X_Δ1,dΔ2[:z]) + x_ΔΔ[:,:,:,i])
        for x1 in [:z,:x,:X,:X′]
            for x2 in [:z,:x,:X,:X′]
                @doinplace R_ΔΔ_temp .+= HR[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end
        #R_ΔΔ_temp .-= HR[:z,:z]⋅(dΔ1[:z],dΔ1[:z]) #the second derivative of z not included in this term
        R_ΔΔ_temp .+= HR[:x,:z]⋅(dx[:z]*dΔ1[:z],dΔ2[:z]) + HR[:z,:x]⋅(dΔ1[:z],dx[:z]*dΔ2[:z]) + HR[:x,:x]⋅(dx[:z]*dΔ1[:z],dx[:z]*dΔ2[:z]) + DR[:x]*d2x[:z,:z]⋅(dΔ1[:z],dΔ2[:z])
        R_ΔΔ .+= ω[i]*R_ΔΔ_temp
    end
    MPI.Allreduce!(R_ΔΔ,+,comm)
    #Now compute the values
    approx.d2X[sΔ1,sΔ2] = -approx.Dinv*R_ΔΔ
    for i in 1:N
        approx.d2x[i][sΔ1,sΔ2] = x_ΔΔ[:,:,:,i] + x_ΔΔX[:,:,i]*approx.d2X[sΔ1,sΔ2]
    end
end



"""
    compute_ΛΘ!(approx::Approximation)

Computes the cross partial with Λ and Θ
"""
function compute_ΛΘ!(approx::Approximation)
    @unpack N,para,dX,dX′,d2X,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,IX′X,nx,nΘ,nX,nΛ,ρ_Θ = para

    #Initialize vairables
    X_ΩΘ = dX[:Ω_Θ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]
    Λ̄_Λ = IΛX*dX[:Λ]
    Λ̄_Λ_values = convert(Vector{Float},eigen(Λ̄_Λ).values)
    Q = convert(Matrix{Float},eigen(Λ̄_Λ).vectors)
    invQ = inv(Q)

    #Next compute required derivatives
    Ω_Θ(i) = Izx*approx.dx[i][:Θ]
    compute_∂X_Λ!(approx,Ω_Θ,:Ω_Θ)


    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX[:Λ]*Λ̄_Λ
    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] =  I,dX[:Θ],IX′X*(dX[:Θ]*ρ_Θ+X_ΩΘ+dX[:Λ]*Λ_Θ)

    dx_ΛΘ = zeros(Float,nx,nΛ,nΘ,N)
    dx_ΛΘX = zeros(Float,nx,nX,nΛ,nΘ,N)
    dx_ΛΘX′ = zeros(Float,nx,nX,nΛ,nΘ,N)
    A1 = zeros(Float,nX,2*nX,nΛ,nΘ)
    A2 = zeros(Float,nX,2*nX,nΛ,nΘ)
    B1 = zeros(Float,nX,nΛ,nΘ)
    B2 = zeros(Float,nX,nΛ,nΘ)
    for i in 1:N
        DF,HF,dx,d2x,∂x,∂X = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx.∂x[i],approx.∂X[i]
        DR,HR = approx.DR[i],approx.HR[i]
        #compute the sum of all the second order terms
        z_Θ = Izx*dx[:Θ]
        d_Θ[:Θ],d_Θ[:x_],d_Θ[:x] = I,Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*z_Θ+dx[:Θ]*ρ_Θ+∂x[:X]*dX′[:Θ]+ dx[:Λ]*Λ_Θ)
        d_Λ[:x_],d_Λ[:x] = Ix_x*dx[:Λ],dx[:Λ]
        d_Λ[:x′] = Ix′x*(dx[:Λ]*Λ̄_Λ)

        F_ΛΘ = DF[:x′]*Ix′x*(d2x[:Λ,:z]⋅(Λ̄_Λ,z_Θ) + d2x[:Λ,:Ω_Θ]⋅(Λ̄_Λ,I)  + d2x[:Λ,:Λ]⋅(Λ̄_Λ,Λ_Θ))
        for x1 in [:Λ,:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΛΘ .+= HF[x1,x2]⋅(d_Λ[x1],d_Θ[x2])
            end
        end
        #Transform into eigenvalue space
        F_ΛΘ .= F_ΛΘ⋅(Q,I)
        for j2 in 1:nΘ
            for j1 in 1:nΛ
                ρ_Λ = Λ̄_Λ_values[j1]#persistence of jth element of Λ
                ρ_θ = ρ_Θ[j2,j2]
                temp = inv(DF[:x_]*Ix_x + DF[:x] + ρ_Λ*ρ_θ*DF[:x′]*Ix′x + DF[:x′]*Ix′x*dx[:z]*Izx)
                dx_ΛΘ[:,j1,j2,i] .= -temp*F_ΛΘ[:,j1,j2]
                dx_ΛΘX[:,:,j1,j2,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
                dx_ΛΘX′[:,:,j1,j2,i] = -temp*DF[:x′]*Ix′x*∂x[:X]
            end
        end

        #Now R
        R_ΛΘ = DR[:X′]*IX′X*(d2X[:Λ,:Ω_Θ]⋅(Λ̄_Λ,I) + d2X[:Λ,:Λ]⋅(Λ̄_Λ,Λ_Θ))
        for x1 in [:Λ,:x,:X,:X′]
            for x2 in [:Θ,:x,:X,:X′]
                @doinplace R_ΛΘ .+= HR[x1,x2]⋅(d_Λ[x1],d_Θ[x2])
            end
        end
        for j2 in 1:nΘ
            for j1 in 1:nΛ
                ρ_Λ = Λ̄_Λ_values[j1]#persistence of jth element of Λ
                ρ_θ = ρ_Θ[j2,j2]
                #Now do the A terms
                A1[:,1:nX,j1,j2] .+= ω[i].*(DR[:x]*dx_ΛΘX[:,:,j1,j2,i]  + DR[:X] + DR[:X′]*IX′X*(ρ_θ*ρ_Λ*I + dX[:Λ]*IΛX))
                A1[:,nX+1:end,j1,j2] .+= ω[i].*(DR[:x]*dx_ΛΘX′[:,:,j1,j2,i] + DR[:X′]*IX′X)
                A2[:,1:nX,j1,j2] .+=  ω[i].*(-∂X*Izx*dx_ΛΘX[:,:,j1,j2,i])
                A2[:,nX+1:end,j1,j2] .+=  ω[i].*(I-∂X*Izx*dx_ΛΘX′[:,:,j1,j2,i])
            end
        end
        #Transform into eigenvalue space
        B1 += -ω[i]*(DR[:x]*dx_ΛΘ[:,:,:,i] + R_ΛΘ⋅(Q,I))
        B2 += ω[i]*(∂X*Izx*dx_ΛΘ[:,:,:,i])
    end

    MPI.Allreduce!(A1,+,comm)
    MPI.Allreduce!(A2,+,comm)
    MPI.Allreduce!(B1,+,comm)
    MPI.Allreduce!(B2,+,comm)

    #Compute derivative of X
    X_ΛΘ  = zeros(Float,nX,nΛ,nΘ)
    X′_ΛΘ  = zeros(Float,nX,nΛ,nΘ)
    for j2 in 1:nΘ
        for j1 in 1:nΛ
            XX′_ΛΘ = [A1[:,:,j1,j2] ; A2[:,:,j1,j2]] \ [B1[:,j1,j2] ; B2[:,j1,j2]]
            X_ΛΘ[:,j1,j2] .= XX′_ΛΘ[1:nX]
            X′_ΛΘ[:,j1,j2] .= XX′_ΛΘ[nX+1:end]
        end
    end

    #use derivative of X to compute derivative of x
    for i in 1:N
        for j2 in 1:nΘ
            for j1 in 1:nΛ
                dx_ΛΘ[:,j1,j2,i] .+= dx_ΛΘX[:,:,j1,j2,i]*X_ΛΘ[:,j1,j2] + dx_ΛΘX′[:,:,j1,j2,i]*X′_ΛΘ[:,j1,j2]
            end
        end
        #convert back away from eigenvalue space
        approx.d2x[i][:Λ,:Θ] = dx_ΛΘ[:,:,:,i]⋅(invQ,I)
        approx.d2x[i][:Θ,:Λ] = permutedims(approx.d2x[i][:Λ,:Θ],[1,3,2])
    end

    approx.d2X[:Λ,:Θ] = X_ΛΘ⋅(invQ,I)
    approx.d2X[:Θ,:Λ] = permutedims(approx.d2X[:Λ,:Θ],[1,3,2])
end

"""
    compute_∂X_Θ!(approx::Approximation)#

Compute the cross ∂X_Θ⋅Δ
"""
function compute_∂X_Θ!(approx::Approximation,Δ,sΔ)
    if haskey(approx.d2X,(:Θ,sΔ)) #check to see if we've already computed this direction
        return
    end
    @unpack N,para,ω,dX,dX′,d2X,d2X′,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,IX′X,nx,nΘ,nz,nX = para

    #First compute ∂2X_ΘΘ
    nΔ = size(Δ(1),2)
    X_ΩΘ = dX[:Ω_Θ]
    X_Δ,Λ_Δ = dX[sΔ],IΛX*dX[sΔ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]

    #Next compute required derivatives
    Ω_Θ(i) = Izx*approx.dx[i][:Θ]
    compute_∂∂!(approx,Δ,sΔ,Ω_Θ,:Ω_Θ)
    compute_∂X_Λ!(approx,Δ,sΔ)
    compute_∂X_Λ!(approx,Ω_Θ,:Ω_Θ)

    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X] = I,dX[:Θ]
    d_Δ= Dict{Symbol,GMatrix}()
    d_Δ[:X] = X_Δ
    x_ΔΘ = zeros(Float,nx,nΔ,nΘ,N)
    x_ΔΘX = zeros(Float,nx,nX,nΘ,N)
    x_ΔΘX′ = zeros(Float,nx,nX,nΘ,N)
    A1 = zeros(Float,nX,2*nX,nΘ)
    A2 = zeros(Float,nX,2*nX,nΘ)
    B1 = zeros(Float,nX,nΔ,nΘ)
    B2 = zeros(Float,nX,nΔ,nΘ)
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR,∂X = approx.DR[i],approx.HR[i],approx.∂X[i]
        #compute the sum of all the second order terms
        z_Θ = Izx*dx[:Θ]
        d_Θ[:x_],d_Θ[:x] = Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*Izx*dx[:Θ]+dx[:Θ]*ρ_Θ+∂x[:X]*X_ΩΘ +dx[:Λ]*IΛX*dX[:Θ])

        d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = Ix_x*∂x[:X]*X_Δ,∂x[:X]*X_Δ,Ix′x*(∂x[:X] + dx[:Λ]*IΛX)*X_Δ

        F_ΔΘ = DF[:x′]*Ix′x*(d2x[:∂X,:z]⋅(X_Δ,z_Θ) + d2x[sΔ,:Ω_Θ] + d2x[sΔ,:Λ]⋅(I,Λ_Θ)
                             +d2x[:Λ,:z]⋅(Λ_Δ,z_Θ) + d2x[:Λ,:Ω_Θ]⋅(Λ_Δ,I) +
                             +d2x[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ) + d2x[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ))

        for x1 in [:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΔΘ .+= HF[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end

        #place holders for the two derivative terms
        for j in 1:nΘ
            ρ_θ = ρ_Θ[j,j]
            temp = inv(DF[:x_]*Ix_x + DF[:x] + ρ_θ*DF[:x′]*Ix′x + DF[:x′]*Ix′x*dx[:z]*Izx)
            x_ΔΘ[:,:,j,i] = -temp*F_ΔΘ[:,:,j]
            x_ΔΘX[:,:,j,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
            x_ΔΘX′[:,:,j,i] = -temp*DF[:x′]*Ix′x*∂x[:X]
        end

        d_Θ[:X′] = IX′X*(dX[:Θ]*ρ_Θ + X_ΩΘ + dX[:Λ]*IΛX*dX[:Θ])
        d_Δ[:X′] = IX′X*(X_Δ + dX[:Λ]*IΛX*X_Δ)
        d_Δ[:z] = Δ(i)

        R_ΔΘ = DR[:x]*d2x[:z,:Θ]⋅(Δ(i),I) + DR[:X′]*IX′X*(d2X[sΔ,:Ω_Θ] + d2X[sΔ,:Λ]⋅(I,Λ_Θ)
                + d2X[:Λ,:Ω_Θ]⋅(Λ_Δ,I) + d2X[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ) + d2X[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ))

        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Θ,:x,:X,:X′]
                @doinplace R_ΔΘ .+= HR[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end
        x_z = dx[:z]
        ∂X′_zz = -approx.Dinv*(DR[:x]*d2x[:z,:z] + HR[:z,:z] + HR[:z,:x]⋅(I,x_z) + HR[:x,:z]⋅(x_z,I) + HR[:x,:x]⋅(x_z,x_z))
        for j in 1:nΘ
            ρ_θ =ρ_Θ[j,j]
            #Now do the R terms
            A1[:,1:nX,j] .+= ω[i].*(DR[:x]*x_ΔΘX[:,:,j,i] + DR[:X] + DR[:X′]*IX′X*(ρ_θ*I + dX[:Λ]*IΛX))
            A1[:,nX+1:end,j] .+= ω[i].*(DR[:x]*x_ΔΘX′[:,:,j,i] + DR[:X′]*IX′X)
            A2[:,1:nX,j] .+=  ω[i].*(-∂X*Izx*x_ΔΘX[:,:,j,i])
            A2[:,nX+1:end,j] .+=  ω[i].*(I-∂X*Izx*x_ΔΘX′[:,:,j,i])
        end
        B1 .+= -ω[i]*(R_ΔΘ + DR[:x]*x_ΔΘ[:,:,:,i])
        B2 .+= ω[i]*(∂X*Izx*(x_ΔΘ[:,:,:,i] + d2x[:Θ,:z]⋅(I,Δ(i)))
                    +∂X′_zz⋅(z_Θ,Δ(i)))
    end
    MPI.Allreduce!(A1,+,comm)
    MPI.Allreduce!(A2,+,comm)
    MPI.Allreduce!(B1,+,comm)
    MPI.Allreduce!(B2,+,comm)

    #Now compute the aggregate responce
    X_ΔΘ = zeros(Float,nX,nΔ,nΘ)
    X′_ΔΘ = zeros(Float,nX,nΔ,nΘ)
    for j in 1:nΘ
        XX′_ΔΘ = [A1[:,:,j] ; A2[:,:,j]] \ [B1[:,:,j] ; B2[:,:,j]]
        X_ΔΘ[:,:,j] .= XX′_ΔΘ[1:nX,:]
        X′_ΔΘ[:,:,j] .= XX′_ΔΘ[nX+1:end,:]

        for i in 1:N
            x_ΔΘ[:,:,j,i] .+= x_ΔΘX[:,:,j,i]*X_ΔΘ[:,:,j] + x_ΔΘX′[:,:,j,i]*X′_ΔΘ[:,:,j]
        end
    end
    for i in 1:N
        approx.d2x[i][sΔ,:Θ] = x_ΔΘ[:,:,:,i]
        approx.d2x[i][:Θ,sΔ] = permutedims(approx.d2x[i][sΔ,:Θ],[1,3,2])
    end
    approx.d2X[sΔ,:Θ] = X_ΔΘ
    approx.d2X[:Θ,sΔ] = permutedims(X_ΔΘ,[1,3,2])
end


"""
    compute_dx_dΘΘ!(approx::Approximation)

Compute the cross derivatives with respect to Θ
"""
function compute_dx_dΘΘ!(approx::Approximation)
    @unpack N,para,ω,dX,dX′,d2X,d2X′,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΘ,nz,nX = para

    #precompute some of the cross derivatives
    #derivative of omega w.r.t Θ
    Ω_Θ(i) = Izx*approx.dx[i][:Θ]
    #compute the derivative  along Ω_Θ
    compute_∂∂!(approx,Ω_Θ,:Ω_Θ,Ω_Θ,:Ω_Θ)
    compute_∂X_Θ!(approx,Ω_Θ,:Ω_Θ)
    compute_∂X_Λ!(approx,Ω_Θ,:Ω_Θ)
    #Prestore some variables for ease of use
    X_ΩΘ = dX′[:Θ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]

    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] =  I,dX[:Θ],IX′X*(dX[:Θ]*ρ_Θ+X_ΩΘ+dX[:Λ]*Λ_Θ)
    temp = zeros(Float,nx,nx)
    #K and J from dΕ note
    A1 = zeros(Float,nX,2*nX,nΘ,nΘ)
    A2 = zeros(Float,nX,2*nX,nΘ,nΘ)
    B1 = zeros(Float,nX,nΘ,nΘ)
    B2 = zeros(Float,nX,nΘ,nΘ)

    x_ΘΘ = zeros(Float,nx,nΘ,nΘ,N)
    x_ΘΘX = zeros(Float,nx,nX,nΘ,nΘ,N)
    x_ΘΘX′ = zeros(Float,nx,nX,nΘ,nΘ,N)

    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR,∂X = approx.DR[i],approx.HR[i],approx.∂X[i]

        d_Θ[:x_],d_Θ[:x] = Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*Izx*dx[:Θ]+dx[:Θ]*ρ_Θ+∂x[:X]*X_ΩΘ +dx[:Λ]*Λ_Θ)
        z_Θ = Izx*dx[:Θ]

        #Compute second order contribution to F
        F_ΘΘ = DF[:x′]*Ix′x*(0.5*d2x[:z,:z]⋅(z_Θ,z_Θ) + d2x[:z,:∂X]⋅(z_Θ,X_ΩΘ) + d2x[:z,:Θ]⋅(z_Θ,ρ_Θ) + d2x[:z,:Λ]⋅(z_Θ,Λ_Θ)
                                                    + 0.5*d2x[:Ω_Θ,:Ω_Θ]      + d2x[:Ω_Θ,:Θ]⋅(I,ρ_Θ) + d2x[:Ω_Θ,:Λ]⋅(I,Λ_Θ)
                                                                                                    + d2x[:Θ,:Λ]⋅(ρ_Θ,Λ_Θ)
                                                                                                   +0.5*d2x[:Λ,:Λ]⋅(Λ_Θ,Λ_Θ))
        F_ΘΘ += permutedims(F_ΘΘ,[1,3,2])

        for x1 in [:Θ,:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΘΘ .+= HF[x1,x2]⋅(d_Θ[x1],d_Θ[x2])
            end
        end

        #Compute second order contribution to R
        R_ΘΘ = DR[:X′]*IX′X*(0.5*d2X[:Ω_Θ,:Ω_Θ] + d2X[:Ω_Θ,:Θ]⋅(I,ρ_Θ) + d2X[:Ω_Θ,:Λ]⋅(I,Λ_Θ)
                                                        + d2X[:Θ,:Λ]⋅(ρ_Θ,Λ_Θ)
                                                       +0.5*d2X[:Λ,:Λ]⋅(Λ_Θ,Λ_Θ))
        R_ΘΘ += permutedims(R_ΘΘ,[1,3,2])
        for x1 in [:Θ,:X,:x,:X′]
            for x2 in [:Θ,:X,:x,:X′]
                @doinplace R_ΘΘ .+= HR[x1,x2]⋅(d_Θ[x1],d_Θ[x2])
            end
        end

        X_zz = -approx.Dinv*(HR[:z,:z]+HR[:z,:x]⋅(I,dx[:z]) + HR[:x,:z]⋅(dx[:z],I) +HR[:x,:x]⋅(dx[:z],dx[:z]) + DR[:x]*d2x[:z,:z])

        for j1 in 1:nΘ
            for j2 in 1:nΘ
                ρ_θ1 = para.ρ_Θ[j1,j1] #compute for each $ρ_Θ#
                ρ_θ2 = para.ρ_Θ[j2,j2] #compute for each $ρ_Θ#
                temp .= inv(DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(dx[:z]*Izx+ρ_θ1*ρ_θ2.*I))
                #compute coefficients on X
                x_ΘΘ[:,j1,j2,i] .= -temp*F_ΘΘ[:,j1,j2]
                x_ΘΘX[:,:,j1,j2,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
                x_ΘΘX′[:,:,j1,j2,i] .= -temp*DF[:x′]*Ix′x*∂x[:X]

                #Compute K and J
                A1[:,1:nX,j1,j2] .+= ω[i].*(DR[:x]*x_ΘΘX[:,:,j1,j2,i] + DR[:X] + DR[:X′]*IX′X*(ρ_θ1*ρ_θ2*I + dX[:Λ]*IΛX) )
                A1[:,nX+1:end,j1,j2] .+= ω[i].*(DR[:x]*x_ΘΘX′[:,:,j1,j2,i] + DR[:X′]*IX′X)
                A2[:,1:nX,j1,j2] .+=  ω[i].*(-∂X*Izx*x_ΘΘX[:,:,j1,j2,i])
                A2[:,nX+1:end,j1,j2] .+=  ω[i].*(I-∂X*Izx*x_ΘΘX′[:,:,j1,j2,i])
                B1[:,j1,j2] .+= -ω[i].*(DR[:x]*x_ΘΘ[:,j1,j2,i] + R_ΘΘ[:,j1,j2] )
                B2[:,j1,j2] .+=  ω[i].*(∂X*Izx*x_ΘΘ[:,j1,j2,i])# + X_zz⋅(z_Θ[:,j1],z_Θ[:,j2]))
            end
        end
    end
    MPI.Allreduce!(A1,+,comm)
    MPI.Allreduce!(A2,+,comm)
    MPI.Allreduce!(B1,+,comm)
    MPI.Allreduce!(B2,+,comm)
    #solve the linear system
    X_ΘΘ = zeros(Float,nX,nΘ,nΘ)
    X′_ΘΘ = zeros(Float,nX,nΘ,nΘ)
    for j1 in 1:nΘ
        for j2 in 1:nΘ
            XX′_ΘΘ = [A1[:,:,j1,j2] ; A2[:,:,j1,j2]] \ [B1[:,j1,j2] ; B2[:,j1,j2]]
            X_ΘΘ[:,j1,j2] .= XX′_ΘΘ[1:nX]
            X′_ΘΘ[:,j1,j2] .= XX′_ΘΘ[nX+1:end]
        end
    end

    #store results
    approx.d2X[:Θ,:Θ] = X_ΘΘ
    approx.d2X′[:Θ,:Θ] = X′_ΘΘ
    for i in 1:N
        approx.d2x[i][:Θ,:Θ] = zeros(Float,nx,nΘ,nΘ)
        for j1 in 1:nΘ
            for j2 in 1:nΘ
                approx.d2x[i][:Θ,:Θ][:,j1,j2] = x_ΘΘ[:,j1,j2,i] + x_ΘΘX[:,:,j1,j2,i]*X_ΘΘ[:,j1,j2] + x_ΘΘX′[:,:,j1,j2,i]*X′_ΘΘ[:,j1,j2]
            end
        end
    end
end


"""
    compute_dx_d2Ε!(approx::Approximation)

Compute the cross derivatives with respect to E
"""
function compute_dx_d2Ε!(approx::Approximation)
    @unpack N,para,ω,dX,dX′,d2X,d2X′,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΘ,nΕ,nz,nX,Fmeasure,Gmeasure = para
    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    #precompute some of the cross derivatives
    #derivative of omega w.r.t Θ
    Ω_Θ(i) = Izx*approx.dx[i][:Θ]
    Ω_Ε(i) = Izx*approx.dx[i][:Ε]
    #compute the derivative  along Ω_Θ and Ω-Ε
    compute_∂∂!(approx,Ω_Ε,:Ω_Ε,Ω_Ε,:Ω_Ε)
    compute_∂X_Θ!(approx,Ω_Ε,:Ω_Ε)
    compute_∂X_Λ!(approx,Ω_Ε,:Ω_Ε)
    compute_∂∂!(approx,Ω_Ε,:Ω_Ε,Ω_Θ,:Ω_Θ)
    compute_∂X_Θ!(approx,Ω_Θ,:Ω_Θ)
    compute_∂X_Λ!(approx,Ω_Θ,:Ω_Θ)
    #Prestore some variables for ease of use
    X_ΩΘ = dX′[:Θ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]
    X_ΩΕ = dX′[:Ε]
    X_Ε,Λ_Ε = dX[:Ε],IΛX*dX[:Ε]


    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] = I,dX[:Θ], IX′X*(dX[:Θ]*ρ_Θ+X_ΩΘ+dX[:Λ]*Λ_Θ)
    d_Ε = Dict{Symbol,GMatrix}()
    d_Ε[:Ε],d_Ε[:X],d_Ε[:X′] = I,dX[:Ε],IX′X*(dX[:Θ]+X_ΩΕ+dX[:Λ]*Λ_Ε)

    #coefficients for determining X_ΕΘ
    A1 = zeros(Float,nX,2*nX)
    A2 = zeros(Float,nX,2*nX)
    B1 = zeros(Float,nX,nΕ,nΘ)
    B2 = zeros(Float,nX,nΕ,nΘ)
    x_ΕΘ = zeros(Float,nx,nΕ,nΘ,N)
    x_X = zeros(Float,nx,nX,N)
    x_X′ = zeros(Float,nx,nX,N)
    #coefficients for determining X_ΕΕ
    C1 = zeros(Float,nX,nΕ,nΕ)
    C2 = zeros(Float,nX,nΕ,nΕ)
    x_ΕΕ = zeros(Float,nx,nΕ,nΕ,N)



    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR,∂X = approx.DR[i],approx.HR[i],approx.∂X[i]

        #preconstruct derivative terms
        z_Θ = Izx*dx[:Θ]
        d_Θ[:x_],d_Θ[:x] = Ix_x*dx[:Θ],dx[:Θ]
        d_Θ[:x′] = Ix′x*(dx[:z]*z_Θ+dx[:Θ]*ρ_Θ+∂x[:X]*dX′[:Θ]+dx[:Λ]*Λ_Θ)

        z_Ε = Izx*dx[:Ε]
        d_Ε[:x] = dx[:Ε]
        d_Ε[:x′] = Ix′x*(dx[:z]*z_Ε+dx[:Θ]+∂x[:X]*dX′[:Ε]+dx[:Λ]*Λ_Ε)

        #Compute second order contribution to F
        F_ΕΘ = DF[:x′]*Ix′x*(d2x[:z,:z]⋅(z_Ε,z_Θ) + d2x[:z,:∂X]⋅(z_Ε,X_ΩΘ) + d2x[:z,:Θ]⋅(z_Ε,ρ_Θ) + d2x[:z,:Λ]⋅(z_Ε,Λ_Θ)
                            +d2x[:∂X,:z]⋅(X_ΩΕ,z_Θ)+d2x[:Ω_Ε,:Ω_Θ]+ d2x[:Ω_Ε,:Θ]⋅(I,ρ_Θ) + d2x[:Ω_Ε,:Λ]⋅(I,Λ_Θ)
                            +d2x[:Θ,:z]⋅(I,z_Θ)+d2x[:Θ,:Ω_Θ]+d2x[:Θ,:Θ]⋅(I,ρ_Θ)+d2x[:Θ,:Λ]⋅(I,Λ_Θ)
                            +d2x[:Λ,:z]⋅(Λ_Ε,z_Θ)+d2x[:Λ,:Ω_Θ]⋅(Λ_Ε,I)+d2x[:Λ,:Θ]⋅(Λ_Ε,ρ_Θ)+d2x[:Λ,:Λ]⋅(Λ_Ε,Λ_Θ))

        for x1 in [:Ε,:X,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΕΘ .+= HF[x1,x2]⋅(d_Ε[x1],d_Θ[x2])
            end
        end
        F_ΕΕ = DF[:x′]*Ix′x*(.5*d2x[:z,:z]⋅(z_Ε,z_Ε) + d2x[:z,:∂X]⋅(z_Ε,X_ΩΕ) + d2x[:z,:Θ]⋅(z_Ε,I) + d2x[:z,:Λ]⋅(z_Ε,Λ_Ε)
                             +0.5*d2x[:Ω_Ε,:Ω_Ε] + d2x[:Ω_Ε,:Θ] + d2x[:Ω_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2x[:Θ,:Θ]+d2x[:Θ,:Λ]⋅(I,Λ_Ε)+0.5*d2x[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))

        F_ΕΕ += permutedims(F_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ΕΕ .+= HF[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end

        #Compute second order contribution to R
        R_ΕΘ = DR[:X′]*IX′X*(d2X[:Ω_Ε,:Ω_Θ] + d2X[:Ω_Ε,:Θ]⋅(I,ρ_Θ) + d2X[:Ω_Ε,:Λ]⋅(I,Λ_Θ)
                             +d2X[:Θ,:Ω_Θ]  + d2X[:Θ,:Θ]⋅(I,ρ_Θ)  + d2X[:Θ,:Λ]⋅(I,Λ_Θ)
                             +d2X[:Λ,:Ω_Θ]⋅(Λ_Ε,I) +d2X[:Λ,:Θ]⋅(Λ_Ε,ρ_Θ) +d2X[:Λ,:Λ]⋅(Λ_Ε,Λ_Θ))
        for x1 in [:Ε,:X,:x,:X′]
            for x2 in [:Θ,:X,:x,:X′]
                @doinplace R_ΕΘ .+= HR[x1,x2]⋅(d_Ε[x1],d_Θ[x2])
            end
        end
        R_ΕΕ = DR[:X′]*IX′X*(0.5*d2X[:Ω_Ε,:Ω_Ε] + d2X[:Ω_Ε,:Θ] + d2X[:Ω_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2X[:Θ,:Θ]  + d2X[:Θ,:Λ]⋅(I,Λ_Ε) + 0.5*d2X[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))
        R_ΕΕ += permutedims(R_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:X′]
            for x2 in [:Ε,:X,:x,:X′]
                @doinplace R_ΕΕ .+= HR[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end

        X_zz = -approx.Dinv*(HR[:z,:z]+HR[:z,:x]⋅(I,dx[:z]) + HR[:x,:z]⋅(dx[:z],I) +HR[:x,:x]⋅(dx[:z],dx[:z]) + DR[:x]*d2x[:z,:z])

        temp = DF[:x]+DF[:x′]*Ix′x*dx[:z]*Izx
        temp[1:nF_,:] = Fmeasure #apply measurablility restriction
        temp = inv(temp)
        #compute coefficients on X for ΕΘ and ΕΕ
        x_ΕΘ[:,:,:,i] .= -temp*F_ΕΘ
        x_ΕΕ[:,:,:,i] .= -temp*F_ΕΕ
        x_X[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx[:Λ]*IΛX)
        x_X′[:,:,i] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        #Compute K and J
        A1[:,1:nX] .+= ω[i].*(DR[:x]*x_X[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX[:Λ]*IΛX)
        A1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′[:,:,i] + DR[:X′]*IX′X)
        A2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_X[:,:,i])
        A2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′[:,:,i])
        B1 .+= -ω[i].*(DR[:x]*x_ΕΘ[:,:,:,i] + R_ΕΘ )
        B2 .+=  ω[i].*(∂X*Izx*x_ΕΘ[:,:,:,i]+X_zz⋅(z_Ε,z_Θ))
        C1 .+= -ω[i].*(DR[:x]*x_ΕΕ[:,:,:,i] + R_ΕΕ )
        C2 .+=  ω[i].*(∂X*Izx*x_ΕΕ[:,:,:,i])#+X_zz⋅(z_Ε,z_Ε))
    end
    MPI.Allreduce!(A1,+,comm)
    MPI.Allreduce!(A2,+,comm)
    MPI.Allreduce!(B1,+,comm)
    MPI.Allreduce!(B2,+,comm)
    MPI.Allreduce!(C1,+,comm)
    MPI.Allreduce!(C2,+,comm)
    #apply measurablility
    A1[1:nG_,1:nX] = Gmeasure
    #solve the linear system
    XX′_ΕΘ = inv([A1 ; A2]) * [B1 ; B2]
    X_ΕΘ = XX′_ΕΘ[1:nX,:,:]
    X′_ΕΘ = XX′_ΕΘ[nX+1:end,:,:]
    XX′_ΕΕ = inv([A1 ; A2]) * [C1 ; C2]
    X_ΕΕ = XX′_ΕΕ[1:nX,:,:]
    X′_ΕΕ = XX′_ΕΕ[nX+1:end,:,:]

    #store results
    approx.d2X[:Ε,:Θ] = X_ΕΘ
    approx.d2X[:Θ,:Ε] = permutedims(X_ΕΘ,[1,3,2])
    approx.d2X[:Ε,:Ε] = X_ΕΕ
    for i in 1:N
        approx.d2x[i][:Ε,:Θ] = x_ΕΘ[:,:,:,i] + x_X[:,:,i]*X_ΕΘ + x_X′[:,:,i]*X′_ΕΘ
        approx.d2x[i][:Θ,:E] = permutedims(approx.d2x[i][:Ε,:Θ],[1,3,2])
        approx.d2x[i][:Ε,:Ε] = x_ΕΕ[:,:,:,i] + x_X[:,:,i]*X_ΕΕ + x_X′[:,:,i]*X′_ΕΕ
    end
end

"""
    compute_σσ!(approx::Approximation)

Computes the second derivative with respect to σ
"""
function compute_σσ!(approx::Approximation)
    @unpack dx,∂x,N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nϵ,nx,nz,nX,Σ_ϵ,Σ_Ε = para


    x_σσ = zeros(Float,nx,N)
    x_Xσσ = zeros(Float,nx,nX,N)
    x_X′σσ = zeros(Float,nx,nX,N)

    #KK and JJ from appendix
    LL1 = zeros(Float,nX,2*nX)
    LL2 = zeros(Float,nX,2*nX)
    MM1 = zeros(Float,nX,1)
    MM2 = zeros(Float,nX,1)
    R_ϵϵ = zeros(Float,nX,nϵ,nϵ)


    for i in 1:N
        DF,DR,dx,dX,∂x,∂X = approx.DF[i],approx.DR[i],approx.dx[i],approx.dX, approx.∂x[i], approx.∂X[i]
        HF,HR,d2x,d2X = approx.HF[i],approx.HR[i],approx.d2x[i],approx.d2X

        temp = inv( DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(I+dx[:z]*Izx) )

        #prestore some derivatives
        x_ϵϵ = Ix_x*d2x[:ϵ,:ϵ]
        x′ϵϵ = Ix′x*d2x[:ϵ,:ϵ]
        x_ΕΕ = Ix_x*d2x[:Ε,:Ε]
        x′ΕΕ = Ix′x*d2x[:Ε,:Ε]
        X′ΕΕ = IX′X*d2X[:Ε,:Ε]
        zϵϵ = Izx*d2x[:ϵ,:ϵ]
        zϵ = Izx*dx[:ϵ]
        #compute coefficients on X
        x_σσ[:,i] .= -temp*( (DF[:x_]*x_ϵϵ  + DF[:x′]*x′ϵϵ)⋅Σ_ϵ + (DF[:x_]*x_ΕΕ  + DF[:x′]*x′ΕΕ)⋅Σ_Ε )
        x_Xσσ[:,:,i] .= -temp*(DF[:X]+DF[:x′]*Ix′x*dx[:Λ]*IΛX)
        x_X′σσ[:,:,i] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        R_ϵϵ = 0.5*HR[:x,:x]⋅(dx[:ϵ],dx[:ϵ]) + HR[:ϵ,:x]⋅(I,dx[:ϵ]) + 0.5*HR[:ϵ,:ϵ] + 0.5*DR[:x]*d2x[:ϵ,:ϵ]
        R_ϵϵ += permutedims(R_ϵϵ,[1,3,2])


        x_z = dx[:z]
        ∂X′_zz = -approx.Dinv*(DR[:x]*d2x[:z,:z] + HR[:z,:z] + HR[:z,:x]⋅(I,x_z) + HR[:x,:z]⋅(x_z,I) + HR[:x,:x]⋅(x_z,x_z))

        #Compute K and J
        LL1[:,1:nX] .+= ω[i].*(DR[:x]*x_Xσσ[:,:,i] + DR[:X] + DR[:X′]*IX′X*(I+dX[:Λ]*IΛX) )
        LL1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′σσ[:,:,i] + DR[:X′]*IX′X)
        LL2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_Xσσ[:,:,i])
        LL2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′σσ[:,:,i])
        MM1 .+= -ω[i].*(DR[:x]*x_σσ[:,i] + R_ϵϵ⋅Σ_ϵ + DR[:X′]*(X′ΕΕ⋅Σ_Ε))
        MM2 .+=  ω[i].*(∂X*(Izx*x_σσ[:,i]+zϵϵ⋅Σ_ϵ) + (∂X′_zz⋅(zϵ,zϵ))⋅Σ_ϵ)
    end
    MPI.Allreduce!(LL1,+,comm)
    MPI.Allreduce!(LL2,+,comm)
    MPI.Allreduce!(MM1,+,comm)
    MPI.Allreduce!(MM2,+,comm)

    XX′_σσ = [LL1 ; LL2] \ [MM1 ; MM2]
    X_σσ = XX′_σσ[1:nX,:]
    X′_σσ = XX′_σσ[nX+1:end,:]

    approx.d2X[:σ,:σ] = reshape(X_σσ,:,1,1)
    #store results
    for i in 1:N
        approx.d2x[i][:σ,:σ] = reshape(x_σσ[:,i] + x_Xσσ[:,:,i]*X_σσ + x_X′σσ[:,:,i]*X′_σσ,:,1,1)
    end
end






###SOME STUF ONLY FOR THE CARA CASE
function simulate_irf(approx::Approximation,Ω::AgentDist,Ε0,T,s_ϵ=1.)
    para = approx.para
    @unpack nϵ,nx,nX,IΘX = para
    N = length(Ω[2])
    ϵ = ones(N,1)*sqrt(para. Σ_ϵ[1])*s_ϵ

    X2 = zeros(Float,nX,T)
    x2 = zeros(Float,N,nx,T)
    Θ = [0.]
    Ε = [Ε0]
    for t in 1:T
        approximate!(approx,Ω)
        X2[:,t],x2[:,:,t] = simulatepolicies(approx,Ω,Θ,Ε,ϵ)
        Θ = IΘX*X2[:,t]
        Ε = [0.]
    end

    X1 = zeros(Float,nX,T)
    x1 = zeros(Float,N,nx,T)
    Θ = [0.]
    Ε = [0.]
    for t in 1:T
        X1[:,t],x1[:,:,t] = simulatepolicies(approx,Ω,Θ,Ε,ϵ)
        Θ = IΘX*X1[:,t]
        Ε = [0.]
    end

    return X2.-X1,x2.-x1
end



"""
    simulatepolicies_complex(approx::Approximation, Ω::AgentDist,Θ,Ε,ϵ)

Computes the policy rules for X and for x
"""
function simulatepolicies_complex_test(approx::Approximation,Ε,ϵ)
    @unpack z̄,x̄,X̄,para,dX,d2X,comm,Ω,labels,my_i = approx
    @unpack Izx,IΘX,nx,nz = para
    my_Ω,my_ω = Ω
    #compute X
    X = X̄+ dX[:Ε]*Ε + 0.5*(d2X[:Ε,:Ε]⋅(Ε,Ε)+d2X[:σ,:σ][:])

    #compute x
    my_N = length(labels)
    my_x = zeros(Float,nx,my_N)
    for i in 1:my_N
        j = labels[i]
        dx,d2x = approx.dx[j],approx.d2x[j]
        sdϵ = sqrt.(diag(para.Σ_ϵ))
        ϵ̂ = min.(max.(ϵ[:,i],-3 .* sdϵ),3 .* sdϵ)
        ẑ = my_Ω[i,:] - z̄[:,j]
        temp = (0.5*d2x[:ϵ,:ϵ]⋅(ϵ̂,ϵ̂) + d2x[:ϵ,:Ε]⋅(ϵ̂,Ε) +0.5*d2x[:Ε,:Ε]⋅(Ε,Ε)+ 0.5*d2x[:σ,:σ][:] +0.5*d2x[:z,:z]⋅(ẑ,ẑ) + d2x[:z,:ϵ]⋅(ẑ,ϵ̂) + d2x[:z,:Ε]⋅(ẑ,Ε) )
        my_x[:,i] .= x̄[:,j] + dx[:z]*ẑ + dx[:ϵ]*ϵ̂ + dx[:Ε]*Ε + temp
    end
    my_z = Izx*my_x
    my_z[1,:] .= my_x[end-3,:]
    my_z[2,:] .= my_x[end-2,:]


    x = MPI_GatherArray(my_x,comm)
    z = MPI_GatherArray(my_z,comm)
    ω = MPI_GatherArray(my_ω,comm)
    i_sort = sortperm(MPI_GatherArray(my_i,comm))
    if MPI.Comm_rank(comm) == 0
        return convert(Vector{Float},X),convert(Matrix{Float},x[:,i_sort]),(convert(Matrix{Float},z'[i_sort,:]),convert(Vector{Float},ω[i_sort]))
    else
        return convert(Vector{Float},X),convert(Matrix{Float},my_x),(convert(Matrix{Float},my_z),convert(Vector{Float},my_ω))
    end
end
