include("approximationMPI.jl")
using Parameters
using LinearAlgebra
using TensorOperations
using NLsolve

"""
Struct holding all of the data for the approximationalong a path
"""
@with_kw mutable struct ApproximationPath{Para}
    #Hold the parameter values
    para::Para
    approx::Vector{Approximation{Para}} = Vector{Approximation{Para}}()
end


"""
    approximate(approx::Approximation,Ω::AgentDist)

Approximates policy rules around current distribution of agents, Θ, Λ
"""
function approximate!(approxPath::ApproximationPath,Ω::AgentDist,Θ,Λ,K=0,K_shock=0,k_cluster=0)
    comm = MPI.COMM_WORLD
    s,r = MPI.Comm_size(comm),MPI.Comm_rank(comm)
    if k_cluster > 0
        Ω̄,my_Ω,my_labels,my_i = approximateDistribution!(comm,Ω,k_cluster)
    else
        N = length(Ω[2])
        Ω̄,my_Ω,my_labels,my_i = Ω,(zeros(Float,0,0),zeros(Float,0)),zeros(Int,0),zeros(Int,0)
    end
    N = length(Ω̄[2])
    my_Ω̄ = (Ω̄[1][MPI_slice(r,s,N),:],Ω̄[2][MPI_slice(r,s,N)])
    temp = copy.(my_Ω̄)
    para = approxPath.para
    #Solve for steady state
    z̄,Θ̄,Λ̄,x̄,X̄ = para.computePath(my_Ω̄,Θ,Λ)
    if K == 0
        K = size(X̄,2)
    end
    if K_shock == 0
        K_shock =K
    end

    approxPath.approx = [Approximation(para=para) for k in 1:K]
    for k in 1:K
        approxPath.approx[k].N = length(my_Ω̄[2])
        if k == size(X̄,2)
            setw̄!(approxPath.approx[k],x̄[:,:,k],X̄[:,k])
        else
            setw̄!(approxPath.approx[k],z̄[:,:,k],Θ̄[:,k],Λ̄[:,k],x̄[:,:,k],X̄[:,k],para.Ix′x*x̄[:,:,k+1],para.IX′X*X̄[:,k+1])
        end
        setupApproximation!(approxPath.approx[k],my_Ω̄[2])
        if k <= K_shock
            approxPath.approx[k].K = K_shock-k+1
        end
    end

    approxPath.approx[1].x̄,approxPath.approx[1].z̄,approxPath.approx[1].X̄ = x̄[:,:,1],z̄[:,:,1],X̄[:,1]
    approxPath.approx[1].Ω= my_Ω
    approxPath.approx[1].labels = my_labels
    approxPath.approx[1].my_i = my_i

    if K == K_shock
        approximate_firstorder!(approxPath.approx[K])
        for k in reverse(1:K-1)
            approximate_firstorder!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
        end

        approximate_secondorder!(approxPath.approx[K])
        for k in reverse(1:K-1)
            approximate_secondorder!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
        end
    else
        approximate_firstorder!(approxPath.approx[K])
        for k in reverse(1:K-1)
            if k < K_shock
                approximate_firstorder!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
            else
                approximate_firstorder_final!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
            end
        end
        approximate_secondorder!(approxPath.approx[K])
        for k in reverse(1:K-1)
            if k < K_shock
                approximate_secondorder!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
            else
                approximate_secondorder_final!(approxPath.approx[k],@view approxPath.approx[k+1:end]) #use view so it does not copy
            end
        end
    end
    return Ω̄,my_Ω̄,temp
end

"""
    getw̄(approx::Approximation,x̄,X̄)

Solves for the vector of w̄  given x̄ and X̄
"""
function setw̄!(approx::Approximation,z̄,Θ̄,Λ̄,x̄,X̄,x̄′,X̄′)
    @unpack para,N = approx
    @unpack Izx,Ix′x,Ix_x,IΘX,IΛX,IX′X,nϵ,nΕ = para

    X̄vec = X̄.*ones(N)'

    approx.w̄ = [z̄;x̄;Ix_x*x̄;x̄′;Θ̄.*ones(N)';Λ̄.*ones(N)';X̄vec;X̄′.*ones(N)';zeros(Float,nϵ,N);zeros(Float,nΕ,N)]
end

"""
    approximate_firstorder!(approx::Approximation)

Do the first order approximation
"""
function approximate_firstorder!(approx::Approximation,approxPath′)
    compute_dx_z!(approx,approxPath′[1])
    compute_dx_ϵ!(approx,approxPath′[1])
    compute_dΛ!(approx,approxPath′[1])
    compute_∂!(approx,approxPath′[1])
    compute_dΘ!(approx,approxPath′[1])
    compute_dΕ!(approx,approxPath′)
end

"""
    approximate_firstorder!(approx::Approximation)

Do the first order approximation
"""
function approximate_firstorder_final!(approx::Approximation,approxPath′)
    compute_dx_z!(approx,approxPath′[1])
    compute_dx_ϵ!(approx,approxPath′[1])
    compute_dΛ!(approx,approxPath′[1])
    compute_dΘ!(approx,approxPath′[1])
    compute_∂_final!(approx,approxPath′[1])
    compute_dΕ_final!(approx,approxPath′[1])
end


"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder!(approx::Approximation,approxPath)
    compute_dx_zz!(approx,approxPath[1])
    compute_zΛ!(approx,approxPath[1])
    compute_dx_ϵϵ!(approx,approxPath[1])
    compute_dx_zϵ!(approx,approxPath[1])
    compute_zΘ!(approx,approxPath[1])
    compute_dΛΛ!(approx,approxPath[1])
    compute_ΛΘ!(approx,approxPath[1])
    compute_dx_dΘΘ!(approx,approxPath[1])
    compute_dx_d2Ε!(approx,approxPath)
    compute_ϵΕ!(approx,approxPath)
    compute_σσ!(approx,approxPath)
end

"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder_final!(approx::Approximation,approxPath)
    compute_dx_zz!(approx,approxPath[1])
    compute_zΛ!(approx,approxPath[1])
    compute_dx_ϵϵ!(approx,approxPath[1])
    compute_dx_zϵ!(approx,approxPath[1])
    compute_zΘ!(approx,approxPath[1])
    compute_dΛΛ!(approx,approxPath[1])
    compute_ΛΘ!(approx,approxPath[1])
    compute_dx_dΘΘ!(approx,approxPath[1])
    compute_dx_d2Ε_final!(approx,approxPath[1])
    compute_ϵΕ_final!(approx,approxPath[1])
    compute_σσ_final!(approx,approxPath[1])
end

"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder_noshock!(approx::Approximation,approx′)
    compute_dx_zz!(approx,approx′)
    compute_zΛ!(approx,approx′)
    compute_dx_ϵϵ!(approx,approx′)
    compute_zΘ!(approx,approx′)
    compute_dΛΛ!(approx,approx′)
    compute_ΛΘ!(approx,approx′)
    compute_dx_dΘΘ!(approx,approx′)
end

"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder_shock_final!(approx::Approximation,approx′)
    compute_dx_d2Ε_final!(approx,approx′)
    compute_ϵΕ_final!(approx,approx′)
    compute_σσ_final!(approx,approx′)
end


"""
    approximate_secondorder!(approx::Approximation)

Do the second order approximation
"""
function approximate_secondorder_shock!(approx::Approximation,approxPath)
    compute_dx_d2Ε!(approx,approxPath)
    compute_ϵΕ!(approx,approxPath)
    compute_σσ!(approx,approxPath)
end


"""
    compute_d_Λ!(approx)

Computes the derivative with respect to Lambda by solving a non-linear equation
in X̄
"""
function compute_dΛ!(approx::Approximation,approx′::Approximation)
    @unpack dx,N,para,ω,comm = approx
    @unpack Ix_x,IX′X,Izx,Ix′x,IΛX,nx,nX,nΛ = para
    dX′ = approx′.dX

    x_Λ = zeros(Float,nx,nΛ,N)
    x_ΛX = zeros(Float,nx,nX,N)
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΛ)
    for i in 1:N
        DF,DR,dx′ = approx.DF[i],approx.DR[i],approx′.dx[i]
        temp = inv(DF[:x_]*Ix_x + DF[:x] )

        x_Λ[:,:,i] .= -temp*DF[:Λ]
        x_ΛX[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        #Now G
        A .+= -ω[i]*(DR[:X] + DR[:x]*x_ΛX[:,:,i] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        B .+= ω[i]*(DR[:x]*x_Λ[:,:,i] + DR[:Λ])

    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    approx.dX[:Λ] = A\B
    for i in 1:N
        approx.dx[i][:Λ] = x_Λ[:,:,i] + x_ΛX[:,:,i]*approx.dX[:Λ]
    end
end # function

"""
    compute_dy_z!(approx)

Computes the derivatives of x w.r.t. z
"""
function compute_dx_z!(approx::Approximation,approx′)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,Izx,nz = para

    for i in 1:N
        DF,dx′ = approx.DF[i],approx′.dx[i]
        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx

        approx.dx[i][:z] = -temp\DF[:z]
    end
end

"""
    compute_dy_ϵ!(approx)

Computes the derivatives of x w.r.t. ϵ
"""
function compute_dx_ϵ!(approx::Approximation,approx′)
    @unpack N,para = approx
    @unpack Ix_x,Ix′x,Izx,Fmeasure = para
    nF_ = size(Fmeasure,1) #number of variables
    for i in 1:N
        DF = approx.DF[i]
        dx′ = approx′.dx[i]
        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] .= Fmeasure

        approx.dx[i][:ϵ] = -temp\DF[:ϵ] #first nF_ equations won't depend on epsilon
    end
end



"""
    compute_∂!(approx)

Compute the Frechet derivative of x and X
"""
function compute_∂!(approx::Approximation,approx′)
    @unpack N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nX = para

    #Compute C(z) from the appendix
    for i in 1:N
        DF = approx.DF[i]
        dx = approx.dx[i]
        dx′ = approx′.dx[i]
        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx

        approx.C[i] = -temp\[(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX) DF[:x′]*Ix′x*approx′.C[i]]
    end

    #compute D from the appendix
    D = zeros(Float,para.nX,para.nX)
    approx.E = zeros(Float,para.nX,(approx.K-1)*para.nX)
    for i in 1:N
        DR = approx.DR[i]
        C0 = approx.C[i][:,1:nX]
        D .+= ω[i]*(DR[:x]*C0 + DR[:X′]*IX′X*approx′.dX[:Λ]*IΛX + DR[:X])
        approx.E .+= ω[i]*(DR[:x]*approx.C[i][:,nX+1:end])
        approx.E[:,1:nX] .+= ω[i]*DR[:X′]*IX′X
    end
    MPI.Allreduce!(D,+,comm)
    MPI.Allreduce!(approx.E,+,comm)
    approx.Dinv = inv(D)
end



"""
    compute_∂_final!(approx,approx′)

Compute the Frechet derivative of x and X
"""
function compute_∂_final!(approx::Approximation,approx′)
    @unpack N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nX = para

    #Compute C(z) from the appendix
    for i in 1:N
        DF = approx.DF[i]
        dx = approx.dx[i]
        dx′ = approx′.dx[i]
        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(dx′[:z]*Izx +I)

        approx.C[i] = -temp\(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        approx.∂x[i][:X] = approx.C[i]
    end

    #compute D from the appendix
    D = zeros(Float,para.nX,para.nX)
    for i in 1:N
        DR = approx.DR[i]
        D .+= ω[i]*(DR[:x]*approx.C[i] + DR[:X′]*IX′X*(approx′.dX[:Λ]*IΛX+I) + DR[:X])
    end
    MPI.Allreduce!(D,+,comm)
    approx.Dinv = inv(D)
    for i in 1:N
        DR,dx = approx.DR[i],approx.dx[i]
        approx.∂X[i] = -approx.Dinv*(DR[:z]  + DR[:x]*dx[:z])
    end
end

function compute_∂!(approxPath::ApproximationPath,Δ,sΔ)
    @unpack N,para,ω,comm = approxPath.approx[end]
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΘ,nx,nX,ρ_Θ = para

    nΔ = size(Δ(1),2)
    X_Δ = zeros(Float,nX,nΔ)
    approx = approxPath.approx[end]
    for i in 1:N
        DR,dx = approx.DR[i],approx.dx[i]
        X_Δ .+= -ω[i]*approx.Dinv*(DR[:z]  + DR[:x]*dx[:z])*Δ(i)
    end
    approx.dX[sΔ] = X_Δ
    for i in 1:N
        approx.dx[i][sΔ] = approx.∂x[i][:X]*X_Δ
    end

    for k in reverse(1:length(approxPath.approx)-1)
        approx = approxPath.approx[k]
        X_Δtemp = zeros(Float,nX,nΔ)
        for i in 1:N
            DR,dx = approx.DR[i],approx.dx[i]
            X_Δtemp .+= -ω[i]*approx.Dinv*((DR[:z]  + DR[:x]*dx[:z])*Δ(i) + approx.E*X_Δ)
        end
        approx.dX[sΔ] = X_Δtemp
        X_Δ = [X_Δtemp; X_Δ]
        for i in 1:N
            approx.dx[i][sΔ] = approx.C[i] * X_Δ
        end
    end
end

"""
    compute_dΘ!(approx::Approximation)

Computes the derivative w.r.t the aggregate state Θ, exploits that
Θ does not effect the distribution
"""
function compute_dΘ!(approx::Approximation,approx′)
    @unpack N,para,ω,dX,comm,K = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΘ,nx,nX,ρ_Θ = para

    x_Θ = zeros(Float,nx,nΘ,N)
    x_XΘ = zeros(Float,nx,nX,N)
    #K and J from dΕ note
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΘ)

    for i in 1:N
        DF,DR,dx,C = approx.DF[i],approx.DR[i],approx.dx[i],approx.C[i]
        dx′,dX′ = approx′.dx[i],approx′.dX
        temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx)
        #compute coefficients on X
        x_Θ[:,:,i] .= -temp*(DF[:Θ] + DF[:x′]*Ix′x*dx′[:Θ]*ρ_Θ)
        x_XΘ[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        #Now compute K and J
        A .+= ω[i].*(DR[:x]*x_XΘ[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        B .+= -ω[i]*(DR[:x]*x_Θ[:,:,i] + DR[:X′]*IX′X*dX′[:Θ]*ρ_Θ + DR[:Θ])
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)

    X_Θ = A\B
    approx.dX[:Θ] = X_Θ
    #store results
    for i in 1:N
        approx.dx[i][:Θ] = x_Θ[:,:,i] + x_XΘ[:,:,i]*X_Θ
    end
end

"""
    compute_dΕ!(approx::Approximation)

Computes the derivative w.r.t the aggregate state Ε
"""
function compute_dΕ!(approx::Approximation,approxPath′)
    @unpack N,para,ω,dX,comm,K = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΕ,nx,nX,ρ_Θ,Fmeasure,Gmeasure = para
    approx′ = approxPath′[1] #next periods approxiation object

    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    x_Ε = zeros(Float,nx,nΕ,N)
    x_XΕ = zeros(Float,nx,nX,N)
    x_C′ = zeros(Float,nx,nx,N)

    #K and J from dΕ note
    A = zeros(Float,K*nX,K*nX)
    B = zeros(Float,K*nX,nΕ)

    for i in 1:N
        DF,DR,dx = approx.DF[i],approx.DR[i],approx.dx[i]
        dx′,C′,dX′ = approx′.dx[i],approx′.C[i],approx′.dX

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] .= Fmeasure #apply the measurablility restriction on subset of the variables
        temp = inv(temp)
        #compute coefficients on X
        x_Ε[:,:,i] .= -temp*(DF[:Ε] + DF[:x′]*Ix′x*dx′[:Θ])
        x_XΕ[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_C′[:,:,i] .= -temp*DF[:x′]*Ix′x

        #Now compute K and J
        A[1:nX,1:nX] .+= ω[i].*(DR[:x]*x_XΕ[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        A[1:nX,nX+1:2*nX] .+= ω[i]*DR[:X′]*IX′X
        A[1:nX,nX+1:end] .+= ω[i]*(DR[:x]*x_C′[:,:,i]*C′)
        B[1:nX,:] .+= -ω[i]*(DR[:x]*x_Ε[:,:,i] + DR[:X′]*IX′X*dX′[:Θ] + DR[:Ε])

        for k in 1:K-1
            approx_k = approxPath′[k]
            Dkinv,Ek,DRk,dxk = approx_k.Dinv,approx_k.E,approx_k.DR[i],approx_k.dx[i]

            A[k*nX+1:(k+1)*nX,1:nX] .+= ω[i]*Dkinv*(DRk[:z]+DRk[:x]*dxk[:z])*Izx*x_XΕ[:,:,i]
            A[k*nX+1:(k+1)*nX,nX+1:end] .+= ω[i]*Dkinv*(DRk[:z]+DRk[:x]*dxk[:z])*Izx*x_C′[:,:,i]*C′
            A[k*nX+1:(k+1)*nX,k*nX+1:(k+1)*nX] .+= ω[i]*eye(nX)
            if k < K-1
                A[k*nX+1:(k+1)*nX,(k+1)*nX+1:end] .+= ω[i]*Dkinv*Ek
            end
            B[k*nX+1:(k+1)*nX,:] .+= -ω[i]*Dkinv*(DRk[:z] + DRk[:x]*dxk[:z])*Izx*x_Ε[:,:,i]
        end
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    #apply the measurablility restriction on X variables
    A[1:nG_,1:nX] .= Gmeasure
    XX′_Ε = A \ B
    X_Ε = XX′_Ε[1:nX,:]
    X′_Ε = XX′_Ε[nX+1:end,:]

    approx.dX[:Ε] = X_Ε
    #store the derivatives in the direction of Ω_Ε
    sΩ_Ε = Symbol(:Ω_Ε,K)
    for k in 1:K-1
        approxPath′[k].dX[sΩ_Ε] = X′_Ε[(k-1)*nX+1:k*nX,:]
    end
    #store results
    for i in 1:N
        approx.dx[i][:Ε] = x_Ε[:,:,i] + x_XΕ[:,:,i]*X_Ε + x_C′[:,:,i]*approx′.C[i]*X′_Ε
        for k in 1:K-1
            approxPath′[k].dx[i][sΩ_Ε] = approxPath′[k].C[i]*X′_Ε[(k-1)*nX+1:end,:]
        end
    end
end

"""
    compute_dΕ_final!(approx::Approximation,approx′)

Computes the derivative w.r.t the aggregate state Ε
"""
function compute_dΕ_final!(approx::Approximation,approx′)
    @unpack N,para,ω,dX,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nΕ,nx,nX,ρ_Θ,Fmeasure,Gmeasure = para
    dX′ = approx′.dX
    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    x_Ε = zeros(Float,N,nx,nΕ)
    x_XΕ = zeros(Float,N,nx,nX)
    x_X′Ε = zeros(Float,N,nx,nX)

    #K and J from dΕ note
    K1 = zeros(Float,nX,2*nX)
    K2 = zeros(Float,nX,2*nX)
    J1 = zeros(Float,nX,nΕ)
    J2 = zeros(Float,nX,nΕ)

    for i in 1:N
        DF,DR,dx,dX,∂x,∂X = approx.DF[i],approx.DR[i],approx.dx[i],approx.dX, approx.∂x[i], approx.∂X[i]
        dx′ = approx′.dx[i]

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] .= Fmeasure #apply the measurablility restriction on subset of the variables
        temp = inv(temp)
        #compute coefficients on X
        x_Ε[i,:,:] .= -temp*(DF[:Ε]+DF[:x′]*Ix′x*dx′[:Θ])
        x_XΕ[i,:,:] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_X′Ε[i,:,:] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        #Compute K and J
        K1[:,1:nX] .+= ω[i].*(DR[:x]*x_XΕ[i,:,:] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        K1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′Ε[i,:,:] + DR[:X′]*IX′X)
        K2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_XΕ[i,:,:])
        K2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′Ε[i,:,:])
        J1 .+= -ω[i].*(DR[:x]*x_Ε[i,:,:] + DR[:Ε] + DR[:X′]*IX′X*dX′[:Θ])
        J2 .+= ω[i].*(∂X*Izx*x_Ε[i,:,:])
    end
    MPI.Allreduce!(K1,+,comm)
    MPI.Allreduce!(K2,+,comm)
    MPI.Allreduce!(J1,+,comm)
    MPI.Allreduce!(J2,+,comm)
    #apply the measurablility restriction on X variables
    K1[1:nG_,1:nX] .= Gmeasure
    XX′_Ε = [K1 ; K2] \ [J1 ; J2]
    X_Ε = XX′_Ε[1:nX,:]
    X′_Ε = XX′_Ε[nX+1:end,:]

    approx.dX[:Ε] = X_Ε
    approx.dX[:Ω_Ε] = X′_Ε
    approx.dX′[:Ε] = X′_Ε
    #store results
    for i in 1:N
        approx.dx[i][:Ε] = x_Ε[i,:,:] + x_XΕ[i,:,:]*X_Ε + x_X′Ε[i,:,:]*X′_Ε
    end
end


"""
    compute_dy_zz!(approx::Approximation,approx′)

Computes the second of x w.r.t. zz
"""
function compute_dx_zz!(approx::Approximation,approx′)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,Izx,nx,nz = para

    d = Dict{Symbol,GMatrix}()
    F_zz = zeros(Float,nx,nz,nz)
    for i in 1:N
        DF,HF,dx,dx′ = approx.DF[i],approx.HF[i],approx.dx[i],approx′.dx[i]
        d2x′ = approx′.d2x[i]

        temp = DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx

        #compute the sum of all the second order terms
        z_z = Izx*dx[:z]
        d[:z],d[:x_],d[:x],d[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]
        F_zz .= DF[:x′]*Ix′x*d2x′[:z,:z]⋅(z_z,z_z)
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:z,:x_,:x,:x′]
                @doinplace F_zz .+= HF[x1,x2]⋅(d[x1],d[x2])
            end
        end

        approx.d2x[i][:z,:z] = -inv(temp)*F_zz
    end
end


"""
    compute_zΛ!(approx::Approximation,approx′)

Computes the cross partial with z and Λ
"""
function compute_zΛ!(approx::Approximation,approx′)
    @unpack N,para,dX = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nz,nX,nΛ = para

    #Initialize vairables
    d_Λ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    Λ̄_Λ = IΛX*dX[:Λ]

    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]



        #compute the sum of all the second order terms
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]
        d_Λ[:Λ],d_Λ[:X],d_Λ[:x_],d_Λ[:x] = I,dX[:Λ],Ix_x*dx[:Λ],dx[:Λ]
        d_Λ[:x′] = Ix′x*dx′[:Λ]*Λ̄_Λ

        F_zΛ = DF[:x′]*Ix′x*d2x′[:z,:Λ]⋅(I,Λ̄_Λ)
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_zΛ .+= HF[x1,x2]⋅(d_z[x1],d_Λ[x2])
            end
        end

        temp = DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx

        approx.d2x[i][:z,:Λ] = -inv(temp)*F_zΛ
        approx.d2x[i][:Λ,:z] = permutedims(approx.d2x[i][:z,:Λ],[1,3,2])
    end
end


"""
    compute_z∂!(approx::Approximation,approxPath,Δ,sΔ)

Computes the cross partial with the Frechet derivative.
"""
function compute_z∂!(approx::Approximation,approxPath,Δ,sΔ)
    if haskey(approx.dx[1],(:z,sΔ))
        return
    end
    @unpack dX,N,para = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nz,nX = para
    approx′ = approxPath[1]
    approxPath′ = @view approxPath[2:end]
    if length(approxPath) == 1
        compute_z∂!(approx′,Δ,sΔ)
    elseif approx′.K==1
        compute_z∂_final!(approx′,approxPath′[1],Δ,sΔ)
    else
        compute_z∂!(approx′,approxPath′,Δ,sΔ)
    end

    d_Δ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    Λ_Δ = IΛX*dX[sΔ]
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx

        #compute the sum of all the second order terms
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]
        d_Δ[:X],d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = dX[sΔ],Ix_x*dx[sΔ],dx[sΔ],Ix′x*dx′[sΔ]

        F_zΔ = DF[:x′]*Ix′x*(d2x′[:z,:Λ]⋅(I,Λ_Δ) + d2x′[:z,sΔ])
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:X,:x_,:x,:x′]
                @doinplace F_zΔ .+= HF[x1,x2]⋅(d_z[x1],d_Δ[x2])
            end
        end
        approx.d2x[i][:z,sΔ] = -inv(temp)*F_zΔ
        approx.d2x[i][sΔ,:z] = permutedims(approx.d2x[i][:z,sΔ],[1,3,2])
    end

end

"""
    compute_z∂_final!(approx::Approximation,approx′,Δ,sΔ)

Computes the cross partial with the Frechet derivative.
"""
function compute_z∂_final!(approx::Approximation,approx′,Δ,sΔ)
    if haskey(approx.dx[1],(:z,sΔ))
        return
    end
    @unpack dX,N,para = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nx,nz,nX = para

    d_Δ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    Λ_Δ = IΛX*dX[sΔ]
    for i in 1:N
        DF,HF,dx,d2x,∂x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx.∂x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        temp = DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(dx′[:z]*Izx + I)

        #compute the sum of all the second order terms
        x_Δ = ∂x[:X]*dX[sΔ]
        d_z[:z],d_z[:x_],d_z[:x],d_z[:x′] = I,Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]
        d_Δ[:X],d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = dX[sΔ],Ix_x*x_Δ,x_Δ,Ix′x*x_Δ

        F_zΔ = DF[:x′]*Ix′x*(d2x′[:z,:Λ]⋅(I,Λ_Δ))
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:X,:x_,:x,:x′]
                @doinplace F_zΔ .+= HF[x1,x2]⋅(d_z[x1],d_Δ[x2])
            end
        end
        approx.d2x[i][:z,sΔ] = -inv(temp)*F_zΔ
        approx.d2x[i][sΔ,:z] = permutedims(approx.d2x[i][:z,sΔ],[1,3,2])
    end

end

"""
    compute_zΘ!(approx::Approximation,approx′)

Computes the cross partial with z and Θ.  Going to exploit that Θ only affects
Λ
"""
function compute_zΘ!(approx::Approximation,approx′)
    @unpack N,para,dX = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,nx,nz,nX,nΘ = para
    Λ_Θ = IΛX*dX[:Θ]
    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X] = I,dX[:Θ]
    d_z = Dict{Symbol,GMatrix}()
    d_z[:z] = I
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        z_Θ = Izx*dx[:Θ]
        #compute the sum of all the second order terms
        d_z[:x_],d_z[:x],d_z[:x′] = Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]
        d_Θ[:x_],d_Θ[:x],d_Θ[:x′] = Ix_x*dx[:Θ],dx[:Θ],Ix′x*(dx′[:z]*z_Θ+dx′[:Θ]*ρ_Θ+dx′[:Λ]*Λ_Θ)

        F_zΘ = DF[:x′]*Ix′x*( d2x′[:z,:z]⋅(I,z_Θ) + d2x′[:z,:Θ]⋅(I,ρ_Θ) + d2x′[:z,:Λ]⋅(I,Λ_Θ) )
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_zΘ .+= HF[x1,x2]⋅(d_z[x1],d_Θ[x2])
            end
        end

        temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx)
        approx.d2x[i][:z,:Θ] = -temp*F_zΘ
        approx.d2x[i][:Θ,:z] = permutedims(approx.d2x[i][:z,:Θ],[1,3,2])
    end
end

"""
    compute_ϵΕ!(approx::Approximation,approxPath)

Computes the cross partial with ϵ
"""
function compute_ϵΕ!(approx::Approximation,approxPath)
    @unpack N,para,dX,K = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nX,nx,nϵ,nΕ,Fmeasure = para
    approx′ = approxPath[1]
    sΩ_Ε = Symbol(:Ω_Ε,K)
    Ω_Ε(i) = Izx*approx.dx[i][:Ε]
    if length(approxPath) == 1
        compute_z∂!(approx′,Ω_Ε,sΩ_Ε)
    elseif approx′.K==1
        compute_z∂_final!(approx′,approxPath[2],Ω_Ε,sΩ_Ε)
    else
        approxPath′ = @view approxPath[2:end]
        compute_z∂!(approx′,approxPath′,Ω_Ε,sΩ_Ε)
    end
    #setup some useful variables
    nF_ = size(Fmeasure,1)
    Λ_Ε = IΛX*dX[:Ε]
    d_Ε = Dict{Symbol,GMatrix}()
    d_Ε[:Ε],d_Ε[:X] = I,dX[:Ε]
    d_ϵ = Dict{Symbol,GMatrix}()
    d_ϵ[:ϵ] = I
    d_z = Dict{Symbol,GMatrix}()
    d_z[:z] = I
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        dz_ϵ = Izx*dx[:ϵ]
        z_z = Izx*dx[:z]
        #compute the sum of all the second order terms
        d_ϵ[:x],d_ϵ[:x′] = dx[:ϵ],Ix′x*dx′[:z]*dz_ϵ
        d_Ε[:x],d_Ε[:x′] = dx[:Ε],Ix′x*(dx′[:z]*Izx*dx[:Ε] + dx′[:Θ] +dx′[sΩ_Ε]+dx′[:Λ]*Λ_Ε)
        d_z[:x_],d_z[:x],d_z[:x′] = Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]*z_z
        F_ϵΕ = DF[:x′]*Ix′x*( d2x′[:z,:z]⋅(dz_ϵ,Izx*dx[:Ε]) + d2x′[:z,:Θ]⋅(dz_ϵ,I) + d2x′[:z,sΩ_Ε]⋅(dz_ϵ,I) + d2x′[:z,:Λ]⋅(dz_ϵ,Λ_Ε))
        F_zΕ = DF[:x′]*Ix′x*( d2x′[:z,:z]⋅(z_z,Izx*dx[:Ε]) + d2x′[:z,:Θ]⋅(z_z,I) + d2x′[:z,sΩ_Ε]⋅(z_z,I) + d2x′[:z,:Λ]⋅(z_z,Λ_Ε))
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ϵΕ .+= HF[x1,x2]⋅(d_ϵ[x1],d_Ε[x2])
            end
        end
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_zΕ .+= HF[x1,x2]⋅(d_z[x1],d_Ε[x2])
            end
        end

        temp = DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #imposes measurablility restriction
        temp = inv(temp)
        approx.d2x[i][:ϵ,:Ε] = -temp*F_ϵΕ
        approx.d2x[i][:Ε,:ϵ] = permutedims(approx.d2x[i][:ϵ,:Ε],[1,3,2])
        approx.d2x[i][:z,:Ε] = -temp*F_zΕ
    end
end


"""
    compute_ϵΕ!(approx::Approximation,approxPath)

Computes the cross partial with ϵ
"""
function compute_ϵΕ_final!(approx::Approximation,approx′)
    @unpack N,para,dX,K = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,nX,nx,nϵ,nΕ,Fmeasure = para
    sΩ_Ε = :Ω_Ε
    Ω_Ε(i) = Izx*approx.dx[i][:Ε]
    compute_z∂_final!(approx,approx′,Ω_Ε,sΩ_Ε)
    #setup some useful variables
    nF_ = size(Fmeasure,1)
    Λ_Ε = IΛX*dX[:Ε]
    d_Ε = Dict{Symbol,GMatrix}()
    d_Ε[:Ε],d_Ε[:X] = I,dX[:Ε]
    d_ϵ = Dict{Symbol,GMatrix}()
    d_ϵ[:ϵ] = I
    d_z = Dict{Symbol,GMatrix}()
    d_z[:z] = I
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        dz_z = Izx*dx[:z]
        dz_ϵ = Izx*dx[:ϵ]
        x_ΩΕ = approx.∂x[i][:X] * approx.dX[sΩ_Ε]
        #compute the sum of all the second order terms
        d_ϵ[:x],d_ϵ[:x′] = dx[:ϵ],Ix′x*dx′[:z]*dz_ϵ
        d_z[:x_],d_z[:x],d_z[:x′] = Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]*dz_z
        d_Ε[:x],d_Ε[:x′] = dx[:Ε],Ix′x*(dx′[:z]*Izx*dx[:Ε] + dx′[:Θ] +x_ΩΕ+dx′[:Λ]*Λ_Ε)

        F_ϵΕ = DF[:x′]*Ix′x*( d2x′[:z,:z]⋅(dz_ϵ,Izx*dx[:Ε]) + d2x′[:z,:Θ]⋅(dz_ϵ,I) + d2x[:z,sΩ_Ε]⋅(dz_ϵ,I) + d2x′[:z,:Λ]⋅(dz_ϵ,Λ_Ε))
        F_zΕ = DF[:x′]*Ix′x*( d2x′[:z,:z]⋅(dz_z,Izx*dx[:Ε]) + d2x′[:z,:Θ]⋅(dz_z,I) + d2x[:z,sΩ_Ε]⋅(dz_z,I) + d2x′[:z,:Λ]⋅(dz_z,Λ_Ε))
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ϵΕ .+= HF[x1,x2]⋅(d_ϵ[x1],d_Ε[x2])
            end
        end
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_zΕ .+= HF[x1,x2]⋅(d_z[x1],d_Ε[x2])
            end
        end

        temp = DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #imposes measurablility restriction
        temp = inv(temp)
        approx.d2x[i][:ϵ,:Ε] = -temp*F_ϵΕ
        approx.d2x[i][:Ε,:ϵ] = permutedims(approx.d2x[i][:ϵ,:Ε],[1,3,2])
        approx.d2x[i][:z,:Ε] = -temp*F_zΕ
    end
end
#

"""
    compute_dy_ϵϵ!(approx,approx′)

Computes the second of x w.r.t. ϵϵ
"""
function compute_dx_ϵϵ!(approx::Approximation,approx′::Approximation)
    @unpack dx,N,para = approx
    @unpack Ix′x,Izx,nx,nϵ,Fmeasure = para
    nF_ = size(Fmeasure,1)
    d = Dict{Symbol,GMatrix}()
    d[:ϵ] = I
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #impose measurablility

        #compute the sum of all the second order terms
        z_ϵ = Izx*dx[:ϵ]
        d[:x],d[:x′] = dx[:ϵ],Ix′x*dx′[:z]*z_ϵ
        F_ϵϵ = DF[:x′]*Ix′x*d2x′[:z,:z]⋅(z_ϵ,z_ϵ)
        for x1 in [:ϵ,:x,:x′]
            for x2 in [:ϵ,:x,:x′]
                @doinplace F_ϵϵ .+= HF[x1,x2]⋅(d[x1],d[x2])
            end
        end

        approx.d2x[i][:ϵ,:ϵ] = -inv(temp)*F_ϵϵ
    end
end

"""
    compute_dy_zϵ!(approx,approx′)

Computes the second of x w.r.t. zϵ
"""
function compute_dx_zϵ!(approx::Approximation,approx′::Approximation)
    @unpack dx,N,para = approx
    @unpack Ix_x,Ix′x,Izx,nx,nϵ,Fmeasure = para
    nF_ = size(Fmeasure,1)
    d_ϵ = Dict{Symbol,GMatrix}()
    d_z = Dict{Symbol,GMatrix}()
    d_ϵ[:ϵ],d_z[:z] = I,I
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #impose measurablility

        #compute the sum of all the second order terms
        z_z = Izx*dx[:z]
        z_ϵ = Izx*dx[:ϵ]
        d_z[:x_],d_z[:x],d_z[:x′] = Ix_x*dx[:z],dx[:z],Ix′x*dx′[:z]*z_z
        d_ϵ[:x],d_ϵ[:x′] = dx[:ϵ],Ix′x*dx′[:z]*z_ϵ
        F_zϵ = DF[:x′]*Ix′x*d2x′[:z,:z]⋅(z_z,z_ϵ)
        for x1 in [:z,:x_,:x,:x′]
            for x2 in [:ϵ,:x,:x′]
                @doinplace F_zϵ .+= HF[x1,x2]⋅(d_z[x1],d_ϵ[x2])
            end
        end

        approx.d2x[i][:z,:ϵ] = -inv(temp)*F_zϵ
    end
end

"""
    compute_dΛΛ!(approx::Approximation,approx′::Approximation)

Computes the second order derivative with respect to Λ
"""
function compute_dΛΛ!(approx::Approximation,approx′::Approximation)
    @unpack N,para,ω,dX,comm =approx
    @unpack Ix_x,Ix′x,IX′X,IΛX,nx,nΛ,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    #Prep work
    Λ̄_Λ = IΛX*dX[:Λ]


    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX′[:Λ]*Λ̄_Λ

    x_ΛΛ = zeros(Float,nx,nΛ,nΛ,N)
    x_ΛΛX = zeros(Float,nx,nX,N)
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΛ,nΛ)
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*dx′[:Λ]*Λ̄_Λ

        F_ΛΛ = DF[:x′]*Ix′x*d2x′[:Λ,:Λ]⋅(Λ̄_Λ,Λ̄_Λ)
        for x1 in [:Λ,:X,:x_,:x,:x′]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ΛΛ .+= HF[x1,x2]⋅(d_Λ[x1],d_Λ[x2])
            end
        end
        #Now R
        R_ΛΛ = DR[:X′]*IX′X*d2X′[:Λ,:Λ]⋅(Λ̄_Λ,Λ̄_Λ)
        for x1 in [:Λ,:x,:X,:X′]
            for x2 in [:Λ,:x,:X,:X′]
                @doinplace R_ΛΛ .+= HR[x1,x2]⋅(d_Λ[x1],d_Λ[x2])
            end
        end

        temp = inv(DF[:x_]*Ix_x + DF[:x])
        x_ΛΛ[:,:,:,i] .= -temp*F_ΛΛ
        x_ΛΛX[:,:,i]  .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        A .+= ω[i]*(DR[:x]*x_ΛΛX[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        B .+= ω[i]*(DR[:x]*x_ΛΛ[:,:,:,i]+ R_ΛΛ)
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    #Now solve
    X_ΛΛ = -inv(A)*B

    for i in 1:N
        approx.d2x[i][:Λ,:Λ] = x_ΛΛ[:,:,:,i] + x_ΛΛX[:,:,i] * X_ΛΛ
    end
    approx.d2X[:Λ,:Λ] = X_ΛΛ
end


"""
    compute_∂X_Λ!(approx::Approximation,approx′::Approximation,Δ,sΔ)

Computes the cross frechet derivative and Λ in the direction of Δ
"""
function compute_∂X_Λ!(approx::Approximation,approxPath,Δ,sΔ)
    if haskey(approx.d2X,(:Λ,sΔ)) #check to see if we've already computed this direction
        return
    end
    #compute future derivative
    approx′ = approxPath[1]
    approxPath′ = @view approxPath[2:end]
    if length(approxPath) == 1
        compute_∂X_Λ!(approx′,Δ,sΔ)
    elseif approx′.K==1
        compute_∂X_Λ_final!(approx′,approxPath′[1],Δ,sΔ)
    else
        compute_∂X_Λ!(approx′,approxPath′,Δ,sΔ)
    end

    @unpack N,para,ω,dX,d2X,comm =approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΛ,nz,nΘ,nX = para

    dX′,d2X′ = approx′.dX,approx′.d2X


    Λ̄_Λ = IΛX*dX[:Λ]

    nΔ = size(Δ(1),2)
    X_Δ = dX[sΔ]
    Λ_Δ = IΛX*X_Δ
    #Now compute the derivatives
    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX′[:Λ]*Λ̄_Λ
    d_Δ = Dict{Symbol,GMatrix}()
    d_Δ[:X],d_Δ[:X′] = X_Δ,IX′X*(dX′[sΔ]+dX′[:Λ]*Λ_Δ)


    R_ΔΛ = zeros(Float,nX,nΔ,nΛ)
    x_ΔΛ = zeros(Float,nx,nΔ,nΛ,N)
    x_ΔΛX = zeros(Float,nx,nX,N)
    A = zeros(Float,nX,nX) #matrix used to solve for X_ΔΛ
    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*dx′[:Λ]*Λ̄_Λ
        d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = Ix_x*dx[sΔ],dx[sΔ],Ix′x*(dx′[sΔ] + dx′[:Λ]*Λ_Δ)
        #First do the derivatives w.r.t F
        F_ΔΛ = DF[:x′]*Ix′x*( d2x′[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) + d2x′[sΔ,:Λ]⋅(I,Λ̄_Λ) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ΔΛ .+= HF[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        temp = inv(DF[:x_]*Ix_x + DF[:x] )
        x_ΔΛ[:,:,:,i] = -temp*F_ΔΛ
        x_ΔΛX[:,:,i] = -temp*(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        #Now R
        d_Δ[:z] = Δ(i) #captures derivative w.r.t Δ using integration by parts
        R_ΔΛ_temp = DR[:X′]*IX′X*(d2X′[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) + d2X′[sΔ,:Λ]⋅(I,Λ̄_Λ)) + DR[:x]*d2x[:z,:Λ]⋅(Δ(i),I)
        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Λ,:x,:X,:X′]
                @doinplace R_ΔΛ_temp .+= HR[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        R_ΔΛ .+= ω[i]*(R_ΔΛ_temp + DR[:x]*x_ΔΛ[:,:,:,i])
        A .+= ω[i]*(DR[:x]*x_ΔΛX[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
    end
    MPI.Allreduce!(R_ΔΛ,+,comm)
    MPI.Allreduce!(A,+,comm)
    #Now solve for X_ΔΛ
    X_ΔΛ = -inv(A)*R_ΔΛ
    #now compute for all agents
    for i in 1:N
        approx.d2x[i][sΔ,:Λ] = x_ΔΛ[:,:,:,i] + x_ΔΛX[:,:,i]*X_ΔΛ
        approx.d2x[i][:Λ,sΔ] = permutedims(approx.d2x[i][sΔ,:Λ],[1,3,2])
    end
    approx.d2X[sΔ,:Λ] = X_ΔΛ
    approx.d2X[:Λ,sΔ] = permutedims(approx.d2X[sΔ,:Λ],[1,3,2])
end


"""
    compute_∂X_Λ!(approx::Approximation,approx′,Δ,sΔ)

Computes the cross frechet derivative and Λ in the direction of Δ
"""
function compute_∂X_Λ_final!(approx::Approximation,approx′,Δ,sΔ)
    if haskey(approx.d2X,(:Λ,sΔ)) #check to see if we've already computed this direction
        return
    end
    @unpack N,para,ω,dX,d2X,comm =approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΛ,nz,nΘ,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X

    Λ̄_Λ = IΛX*dX[:Λ]
    Λ̄_Λ_values = convert(Vector{Float},eigen(Λ̄_Λ).values)
    Q = convert(Matrix{Float},eigen(Λ̄_Λ).vectors)
    invQ = inv(Q)

    nΔ = size(Δ(1),2)
    X_Δ = dX[sΔ]
    Λ_Δ = IΛX*X_Δ
    #Now compute the derivatives
    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX′[:Λ]*Λ̄_Λ
    d_Δ = Dict{Symbol,GMatrix}()
    d_Δ[:X],d_Δ[:X′] = X_Δ,IX′X*(X_Δ+dX′[:Λ]*Λ_Δ)


    R_ΔΛ = zeros(Float,nX,nΔ,nΛ)
    x_ΔΛ = zeros(Float,nx,nΔ,nΛ,N)
    x_ΔΛX = zeros(Float,nx,nX,nΛ,N)
    C = zeros(Float,nX,nX,nΛ) #matrix used to solve for X_ΔΛ
    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*dx′[:Λ]*Λ̄_Λ
        d_Δ[:x_],d_Δ[:x] = Ix_x*∂x[:X]*X_Δ,∂x[:X]*X_Δ
        d_Δ[:x′] = Ix′x*(∂x[:X]*X_Δ + dx′[:Λ]*Λ_Δ)
        #First do the derivatives w.r.t F
        F_ΔΛ = DF[:x′]*Ix′x*( d2x′[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:Λ,:X,:x_,:x,:x′]
                @doinplace F_ΔΛ .+= HF[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        #Now R
        d_Δ[:z] = Δ(i)
        R_ΔΛ_temp = DR[:X′]*IX′X*d2X′[:Λ,:Λ]⋅(Λ_Δ,Λ̄_Λ) + DR[:x]*d2x[:z,:Λ]⋅(Δ(i),I)

        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Λ,:x,:X,:X′]
                @doinplace R_ΔΛ_temp .+= HR[x1,x2]⋅(d_Δ[x1],d_Λ[x2])
            end
        end
        #convert into eigenvector space

        F_ΔΛ = F_ΔΛ⋅(I,Q)
        for j in 1:nΛ
            ρ_Λ = Λ̄_Λ_values[j]
            temp = inv(DF[:x_]*Ix_x + DF[:x] + ρ_Λ*DF[:x′]*Ix′x )
            x_ΔΛ[:,:,j,i] = -temp*F_ΔΛ[:,:,j]
            x_ΔΛX[:,:,j,i] = -temp*(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

            C[:,:,j] .+= ω[i]*(DR[:x]*x_ΔΛX[:,:,j,i] + DR[:X] + DR[:X′]*IX′X*(ρ_Λ*I + dX′[:Λ]*IΛX))
        end
        R_ΔΛ .+= ω[i]*(R_ΔΛ_temp⋅(I,Q) + DR[:x]*x_ΔΛ[:,:,:,i])
    end
    MPI.Allreduce!(R_ΔΛ,+,comm)
    MPI.Allreduce!(C,+,comm)
    #Now solve
    X_ΔΛ = zeros(Float,nX,nΔ,nΛ)
    for j in 1:nΛ
        X_ΔΛ[:,:,j] = -inv(C[:,:,j])*R_ΔΛ[:,:,j]
    end
    #now compute for all agents
    for i in 1:N
        for j in 1:nΘ
            x_ΔΛ[:,:,j,i] .+= x_ΔΛX[:,:,j,i]*X_ΔΛ[:,:,j]
        end
        #convert back into regular space
        approx.d2x[i][sΔ,:Λ] = x_ΔΛ[:,:,:,i]⋅(I,invQ)
        approx.d2x[i][:Λ,sΔ] = permutedims(approx.d2x[i][sΔ,:Λ],[1,3,2])
    end
    approx.d2X[sΔ,:Λ] = X_ΔΛ⋅(I,invQ)
    approx.d2X[:Λ,sΔ] = permutedims(approx.d2X[sΔ,:Λ],[1,3,2])
end


"""
    compute_∂∂!(approx::Approximation,approxPath,Δ1,sΔ1,Δ2,sΔ2)

Computes the second Frechet derivative in direction Δ1,Δ2
"""
function compute_∂∂!(approx::Approximation,approxPath,Δ1,sΔ1,Δ2,sΔ2)
    if haskey(approx.d2X,(sΔ1,sΔ2)) #check to see if we've already computed this direction
        return
    end
    approx′ = approxPath[1]
    approxPath′ = @view approxPath[2:end]
    if length(approxPath) == 1
        compute_∂X_Λ!(approx′,Δ1,sΔ1)
        compute_∂X_Λ!(approx′,Δ2,sΔ2)
        compute_∂∂!(approx′,Δ1,sΔ1,Δ2,sΔ2)
    elseif approx′.K ==1
        compute_∂X_Λ_final!(approx′,approxPath′[1],Δ1,sΔ1)
        compute_∂X_Λ_final!(approx′,approxPath′[1],Δ2,sΔ2)
        compute_∂∂_final!(approx′,approxPath′[1],Δ1,sΔ1,Δ2,sΔ2)
    else
        compute_∂X_Λ!(approx′,approxPath′,Δ1,sΔ1)
        compute_∂X_Λ!(approx′,approxPath′,Δ2,sΔ2)
        compute_∂∂!(approx′,approxPath′,Δ1,sΔ1,Δ2,sΔ2)
    end
    compute_z∂!(approx,approxPath,Δ1,sΔ1)
    compute_z∂!(approx,approxPath,Δ2,sΔ2)

    @unpack dX,d2X,N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    nΔ1 = size(Δ1(1),2)
    nΔ2 = size(Δ2(1),2)

    X_Δ1,X_Δ2 = dX[sΔ1],dX[sΔ2]
    Λ_Δ1,Λ_Δ2 = IΛX*X_Δ1,IΛX*X_Δ2

    dΔ1 = Dict{Symbol,GMatrix}()
    dΔ1[:X],dΔ1[:X′] = X_Δ1, IX′X*(dX′[sΔ1]+dX′[:Λ]*Λ_Δ1)
    dΔ2 = Dict{Symbol,GMatrix}()
    dΔ2[:X],dΔ2[:X′] = X_Δ2, IX′X*(dX′[sΔ2]+dX′[:Λ]*Λ_Δ2)


    R_ΔΔ = zeros(Float,nX,nΔ1,nΔ2)
    x_ΔΔ = zeros(Float,nx,nΔ1,nΔ2,N)
    x_ΔΔX = zeros(Float,nx,nX,N)

    for i in 1:N
        DF,HF,dx,d2x,dx′,d2x′ = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx′.dx[i],approx′.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        temp = inv(DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx )

        dΔ1[:x_],dΔ1[:x],dΔ1[:x′] = Ix_x*dx[sΔ1],dx[sΔ1],Ix′x*(dx′[sΔ1] + dx′[:Λ]*Λ_Δ1)
        dΔ2[:x_],dΔ2[:x],dΔ2[:x′] = Ix_x*dx[sΔ2],dx[sΔ2],Ix′x*(dx′[sΔ2] + dx′[:Λ]*Λ_Δ2)
        F_ΔΔ = DF[:x′]*Ix′x*(d2x′[sΔ1,sΔ2] + d2x′[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2x′[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2x′[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:x_,:x,:x′,:X]
                @doinplace F_ΔΔ .+= HF[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end

        x_ΔΔ[:,:,:,i] = -temp*F_ΔΔ
        x_ΔΔX[:,:,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        #Now the terms coming from the R equations
        dΔ1[:z],dΔ2[:z] = Δ1(i),Δ2(i)
        R_ΔΔ_temp = DR[:X′]*IX′X*(d2X′[sΔ1,sΔ2] + d2X′[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2X′[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2X′[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2))
        R_ΔΔ_temp .+= DR[:x]*(d2x[:z,sΔ2]⋅(dΔ1[:z],I) + d2x[sΔ1,:z]⋅(I,dΔ2[:z])+ x_ΔΔ[:,:,:,i])
        for x1 in [:z,:x,:X,:X′]
            for x2 in [:z,:x,:X,:X′]
                @doinplace R_ΔΔ_temp .+= HR[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end
        #R_ΔΔ_temp .-= HR[:z,:z]⋅(dΔ1[:z],dΔ1[:z]) #the second derivative of z not included in this term
        R_ΔΔ_temp .+=HR[:x,:z]⋅(dx[:z]*dΔ1[:z],dΔ2[:z])+ HR[:z,:x]⋅(dΔ1[:z],dx[:z]*dΔ2[:z]) + HR[:x,:x]⋅(dx[:z]*dΔ1[:z],dx[:z]*dΔ2[:z]) + DR[:x]*d2x[:z,:z]⋅(dΔ1[:z],dΔ2[:z])
        R_ΔΔ .+= ω[i]*R_ΔΔ_temp
    end
    MPI.Allreduce!(R_ΔΔ,+,comm)
    #Now compute the values
    approx.d2X[sΔ1,sΔ2] = -approx.Dinv*R_ΔΔ
    for i in 1:N
        approx.d2x[i][sΔ1,sΔ2] = x_ΔΔ[:,:,:,i] + x_ΔΔX[:,:,i]*approx.d2X[sΔ1,sΔ2]
    end
end


"""
    compute_∂∂_final!(approx::Approximation,approxPath,Δ1,sΔ1,Δ2,sΔ2)

Computes the second Frechet derivative in direction Δ1,Δ2
"""
function compute_∂∂_final!(approx::Approximation,approx′,Δ1,sΔ1,Δ2,sΔ2)
    if haskey(approx.d2X,(sΔ1,sΔ2)) #check to see if we've already computed this direction
        return
    end
    compute_∂X_Λ_final!(approx,approx′,Δ1,sΔ1)
    compute_∂X_Λ_final!(approx,approx′,Δ2,sΔ2)
    compute_z∂_final!(approx,approx′,Δ1,sΔ1)
    compute_z∂_final!(approx,approx′,Δ2,sΔ2)

    @unpack dX,d2X,N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    nΔ1 = size(Δ1(1),2)
    nΔ2 = size(Δ2(1),2)

    X_Δ1,X_Δ2 = dX[sΔ1],dX[sΔ2]
    Λ_Δ1,Λ_Δ2 = IΛX*X_Δ1,IΛX*X_Δ2

    dΔ1 = Dict{Symbol,GMatrix}()
    dΔ1[:X],dΔ1[:X′] = X_Δ1, IX′X*(X_Δ1+dX′[:Λ]*Λ_Δ1)
    dΔ2 = Dict{Symbol,GMatrix}()
    dΔ2[:X],dΔ2[:X′] = X_Δ2, IX′X*(X_Δ2+dX′[:Λ]*Λ_Δ2)


    R_ΔΔ = zeros(Float,nX,nΔ1,nΔ2)
    x_ΔΔ = zeros(Float,nx,nΔ1,nΔ2,N)
    x_ΔΔX = zeros(Float,nx,nX,N)

    for i in 1:N
        DF,HF,dx,d2x,dx′,d2x′ = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx′.dx[i],approx′.d2x[i]
        DR,HR,∂x = approx.DR[i],approx.HR[i],approx.∂x[i]
        temp = inv(DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*(dx′[:z]*Izx+I) )

        x_Δ1 = ∂x[:X]*X_Δ1
        x_Δ2 = ∂x[:X]*X_Δ2
        dΔ1[:x_],dΔ1[:x],dΔ1[:x′] = Ix_x*x_Δ1,x_Δ1,Ix′x*(x_Δ1 + dx′[:Λ]*Λ_Δ1)
        dΔ2[:x_],dΔ2[:x],dΔ2[:x′] = Ix_x*x_Δ2,x_Δ2,Ix′x*(x_Δ2 + dx′[:Λ]*Λ_Δ2)
        F_ΔΔ = DF[:x′]*Ix′x*(d2x[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2x[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2x′[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2) )
        for x1 in [:x_,:x,:x′,:X]
            for x2 in [:x_,:x,:x′,:X]
                @doinplace F_ΔΔ .+= HF[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end

        x_ΔΔ[:,:,:,i] = -temp*F_ΔΔ
        x_ΔΔX[:,:,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        #Now the terms coming from the R equations
        dΔ1[:z],dΔ2[:z] = Δ1(i),Δ2(i)
        R_ΔΔ_temp = DR[:X′]*IX′X*(d2X[sΔ1,:Λ]⋅(I,Λ_Δ2) + d2X[:Λ,sΔ2]⋅(Λ_Δ1,I) + d2X′[:Λ,:Λ]⋅(Λ_Δ1,Λ_Δ2))
        R_ΔΔ_temp .+= DR[:x]*(d2x[:z,sΔ2]⋅(dΔ1[:z],I) + d2x[sΔ1,:z]⋅(I,dΔ2[:z])+ x_ΔΔ[:,:,:,i])
        for x1 in [:z,:x,:X,:X′]
            for x2 in [:z,:x,:X,:X′]
                @doinplace R_ΔΔ_temp .+= HR[x1,x2]⋅(dΔ1[x1],dΔ2[x2])
            end
        end
        #R_ΔΔ_temp .-= HR[:z,:z]⋅(dΔ1[:z],dΔ1[:z]) #the second derivative of z not included in this term
        R_ΔΔ_temp .+=HR[:x,:z]⋅(dx[:z]*dΔ1[:z],dΔ2[:z])+ HR[:z,:x]⋅(dΔ1[:z],dx[:z]*dΔ2[:z]) + HR[:x,:x]⋅(dx[:z]*dΔ1[:z],dx[:z]*dΔ2[:z]) + DR[:x]*d2x[:z,:z]⋅(dΔ1[:z],dΔ2[:z])
        R_ΔΔ .+= ω[i]*R_ΔΔ_temp
    end
    MPI.Allreduce!(R_ΔΔ,+,comm)
    #Now compute the values
    approx.d2X[sΔ1,sΔ2] = -approx.Dinv*R_ΔΔ
    for i in 1:N
        approx.d2x[i][sΔ1,sΔ2] = x_ΔΔ[:,:,:,i] + x_ΔΔX[:,:,i]*approx.d2X[sΔ1,sΔ2]
    end
end


"""
    compute_ΛΘ!(approx::Approximation,approx′::Approximation)

Computes the cross partial with Λ and Θ, exploiting that neither affects the
distribution
"""
function compute_ΛΘ!(approx::Approximation,approx′::Approximation)
    @unpack N,para,dX,d2X,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IΛX,IX′X,nx,nΘ,nX,nΛ,ρ_Θ = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    #Initialize vairables
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]
    Λ̄_Λ = IΛX*dX[:Λ]

    d_Λ = Dict{Symbol,GMatrix}()
    d_Λ[:Λ],d_Λ[:X],d_Λ[:X′] = I,dX[:Λ],IX′X*dX′[:Λ]*Λ̄_Λ
    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] =  I,dX[:Θ],IX′X*(dX′[:Θ]*ρ_Θ+dX′[:Λ]*Λ_Θ)

    dx_ΛΘ = zeros(Float,nx,nΛ,nΘ,N)
    dx_ΛΘX = zeros(Float,nx,nX,N)
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΛ,nΘ)
    for i in 1:N
        DF,HF,dx,d2x,dx′,d2x′ = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx′.dx[i],approx′.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        z_Θ = Izx*dx[:Θ]
        #compute the sum of all the second order terms
        d_Θ[:x_],d_Θ[:x],d_Θ[:x′] = Ix_x*dx[:Θ],dx[:Θ],Ix′x*(dx′[:z]*z_Θ + dx′[:Θ]*ρ_Θ+dx′[:Λ]*Λ_Θ)
        d_Λ[:x_],d_Λ[:x],d_Λ[:x′] = Ix_x*dx[:Λ],dx[:Λ],Ix′x*(dx′[:Λ]*Λ̄_Λ)

        F_ΛΘ = DF[:x′]*Ix′x*(d2x′[:Λ,:z]⋅(Λ̄_Λ,z_Θ) + d2x′[:Λ,:Θ]⋅(Λ̄_Λ,ρ_Θ)  + d2x′[:Λ,:Λ]⋅(Λ̄_Λ,Λ_Θ))
        for x1 in [:Λ,:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΛΘ .+= HF[x1,x2]⋅(d_Λ[x1],d_Θ[x2])
            end
        end

        temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx)
        dx_ΛΘ[:,:,:,i] .= -temp*F_ΛΘ
        dx_ΛΘX[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        #Now R
        R_ΛΘ = DR[:X′]*IX′X*(d2X′[:Λ,:Θ]⋅(Λ̄_Λ,ρ_Θ) + d2X′[:Λ,:Λ]⋅(Λ̄_Λ,Λ_Θ))
        for x1 in [:Λ,:x,:X,:X′]
            for x2 in [:Θ,:x,:X,:X′]
                @doinplace R_ΛΘ .+= HR[x1,x2]⋅(d_Λ[x1],d_Θ[x2])
            end
        end
        A .+= ω[i].*(DR[:x]*dx_ΛΘX[:,:,i]  + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        B .+= ω[i].*(DR[:x]*dx_ΛΘ[:,:,:,i] + R_ΛΘ)
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)

    #Compute derivative of X
    X_ΛΘ  = -inv(A)*B

    #use derivative of X to compute derivative of x
    for i in 1:N
        approx.d2x[i][:Λ,:Θ] = dx_ΛΘ[:,:,:,i] + dx_ΛΘX[:,:,i]*X_ΛΘ
        approx.d2x[i][:Θ,:Λ] = permutedims(approx.d2x[i][:Λ,:Θ],[1,3,2])
    end

    approx.d2X[:Λ,:Θ] = X_ΛΘ
    approx.d2X[:Θ,:Λ] = permutedims(approx.d2X[:Λ,:Θ],[1,3,2])
end

"""
    compute_∂X_Θ!(approx::Approximation,approxPath,Δ,sΔ)

Compute the cross ∂X_Θ⋅Δ
"""
function compute_∂X_Θ!(approx::Approximation,approxPath,Δ,sΔ)
    if haskey(approx.d2X,(:Θ,sΔ)) #check to see if we've already computed this direction
        return
    end
    approx′ = approxPath[1]
    approxPath′ = @view approxPath[2:end]
    #compute future derivatives
    if length(approxPath) == 1
        compute_z∂!(approx′,Δ,sΔ)
        compute_∂X_Λ!(approx′,Δ,sΔ)
        compute_∂X_Θ!(approx′,Δ,sΔ)
    elseif approx′.K ==1
        compute_z∂_final!(approx′,approxPath′[1],Δ,sΔ)
        compute_∂X_Λ_final!(approx′,approxPath′[1],Δ,sΔ)
        compute_∂X_Θ_final!(approx′,approxPath′[1],Δ,sΔ)
    else
        compute_z∂!(approx′,approxPath′,Δ,sΔ)
        compute_∂X_Λ!(approx′,approxPath′,Δ,sΔ)
        compute_∂X_Θ!(approx′,approxPath′,Δ,sΔ)
    end
    @unpack N,para,ω,dX,d2X,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,IX′X,nx,nΘ,nz,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X #Future derivatives
    #First compute ∂2X_ΘΘ
    nΔ = size(Δ(1),2)
    X_Δ,Λ_Δ = dX[sΔ],IΛX*dX[sΔ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]


    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] = I,dX[:Θ],IX′X*(dX′[:Θ]*ρ_Θ + dX′[:Λ]*IΛX*dX[:Θ])
    d_Δ= Dict{Symbol,GMatrix}()
    d_Δ[:X],d_Δ[:X′] = X_Δ,IX′X*(dX′[sΔ] + dX′[:Λ]*IΛX*X_Δ)
    x_ΔΘ = zeros(Float,nx,nΔ,nΘ,N)
    x_ΔΘX = zeros(Float,nx,nX,N)
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΔ,nΘ)

    for i in 1:N
        DF,HF,dx,d2x,dx′,d2x′ = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx′.dx[i],approx′.d2x[i]
        DR,HR = approx.DR[i],approx.HR[i]
        z_Θ = Izx*dx[:Θ]
        #compute the sum of all the second order terms
        d_Θ[:x_],d_Θ[:x],d_Θ[:x′] = Ix_x*dx[:Θ],dx[:Θ],Ix′x*(dx′[:z]*z_Θ + dx′[:Θ]*ρ_Θ + dx′[:Λ]*Λ_Θ)
        d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = Ix_x*dx[sΔ],dx[sΔ],Ix′x*(dx′[sΔ] + dx′[:Λ]*Λ_Δ)

        F_ΔΘ = DF[:x′]*Ix′x*(d2x′[sΔ,:z]⋅(I,z_Θ)+d2x′[sΔ,:Θ]⋅(I,ρ_Θ) + d2x′[sΔ,:Λ]⋅(I,Λ_Θ)
                            +d2x′[:Λ,:z]⋅(Λ_Δ,z_Θ)+d2x′[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ) + d2x′[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ))
        for x1 in [:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΔΘ .+= HF[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end

        temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*dx′[:z]*Izx)
        x_ΔΘ[:,:,:,i] = -temp*F_ΔΘ
        x_ΔΘX[:,:,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)

        d_Δ[:z] = Δ(i)

        R_ΔΘ = DR[:x]*d2x[:z,:Θ]⋅(Δ(i),I) + DR[:X′]*IX′X*(d2X′[sΔ,:Θ]⋅(I,ρ_Θ) + d2X′[sΔ,:Λ]⋅(I,Λ_Θ)
                                            +d2X′[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ) + d2X′[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ))

        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Θ,:x,:X,:X′]
                @doinplace R_ΔΘ .+= HR[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end

        A .+= ω[i].*(DR[:x]*x_ΔΘX[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)

        B .+= -ω[i]*(R_ΔΘ + DR[:x]*x_ΔΘ[:,:,:,i])
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)

    #Now compute the aggregate responce
    X_ΔΘ = inv(A)* B
    for i in 1:N
        approx.d2x[i][sΔ,:Θ] = x_ΔΘ[:,:,:,i] + x_ΔΘX[:,:,i] * X_ΔΘ
        approx.d2x[i][:Θ,sΔ] = permutedims(approx.d2x[i][sΔ,:Θ],[1,3,2])
    end
    approx.d2X[sΔ,:Θ] = X_ΔΘ
    approx.d2X[:Θ,sΔ] = permutedims(X_ΔΘ,[1,3,2])
end

"""
    compute_∂X_Θ_final!(approx::Approximation,approx′,Δ,sΔ)

Compute the cross ∂X_Θ⋅Δ
"""
function compute_∂X_Θ_final!(approx::Approximation,approx′,Δ,sΔ)
    if haskey(approx.d2X,(:Θ,sΔ)) #check to see if we've already computed this direction
        return
    end
    compute_∂X_Λ_final!(approx,approx′,Δ,sΔ)
    compute_z∂_final!(approx,approx′,Δ,sΔ)
    @unpack N,para,ω,dX,d2X,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IΛX,IX′X,nx,nΘ,nz,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X #Future derivatives
    #First compute ∂2X_ΘΘ
    nΔ = size(Δ(1),2)
    X_Δ,Λ_Δ = dX[sΔ],IΛX*dX[sΔ]
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]


    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] = I,dX[:Θ],IX′X*(dX′[:Θ]*ρ_Θ + dX′[:Λ]*IΛX*dX[:Θ])
    d_Δ= Dict{Symbol,GMatrix}()
    d_Δ[:X],d_Δ[:X′] = X_Δ,IX′X*(X_Δ + dX′[:Λ]*IΛX*X_Δ)
    x_ΔΘ = zeros(Float,nx,nΔ,nΘ,N)
    x_ΔΘX = zeros(Float,nx,nX,nΘ,N)
    A = zeros(Float,nX,nX,nΘ)
    B = zeros(Float,nX,nΔ,nΘ)

    for i in 1:N
        DF,HF,dx,d2x,dx′,d2x′ = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i],approx′.dx[i],approx′.d2x[i]
        DR,HR,∂x = approx.DR[i],approx.HR[i],approx.∂x[i]
        z_Θ = Izx*dx[:Θ]
        #compute the sum of all the second order terms
        d_Θ[:x_],d_Θ[:x],d_Θ[:x′] = Ix_x*dx[:Θ],dx[:Θ],Ix′x*(dx′[:z]*z_Θ + dx′[:Θ]*ρ_Θ + dx′[:Λ]*Λ_Θ)
        d_Δ[:x_],d_Δ[:x],d_Δ[:x′] = Ix_x*∂x[:X]*X_Δ,∂x[:X]*X_Δ,Ix′x*(∂x[:X]*X_Δ + dx′[:Λ]*Λ_Δ)

        F_ΔΘ = DF[:x′]*Ix′x*(d2x[sΔ,:z]⋅(I,z_Θ)+d2x[sΔ,:Λ]⋅(I,Λ_Θ)+d2x′[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ)+d2x′[:Λ,:z]⋅(Λ_Δ,z_Θ) + d2x′[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ))
        for x1 in [:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΔΘ .+= HF[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end
        for j in 1:nΘ
            ρ_θ = ρ_Θ[j,j]
            temp = inv(DF[:x_]*Ix_x + DF[:x] + DF[:x′]*Ix′x*(dx′[:z]*Izx+ρ_θ*I))
            x_ΔΘ[:,:,j,i] = -temp*F_ΔΘ[:,:,j]
            x_ΔΘX[:,:,j,i] = -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        end

        d_Δ[:z] = Δ(i)

        R_ΔΘ = DR[:x]*d2x[:z,:Θ]⋅(Δ(i),I) + DR[:X′]*IX′X*(d2X[sΔ,:Λ]⋅(I,Λ_Θ)
                                            +d2X′[:Λ,:Θ]⋅(Λ_Δ,ρ_Θ) + d2X′[:Λ,:Λ]⋅(Λ_Δ,Λ_Θ))

        for x1 in [:z,:x,:X,:X′]
            for x2 in [:Θ,:x,:X,:X′]
                @doinplace R_ΔΘ .+= HR[x1,x2]⋅(d_Δ[x1],d_Θ[x2])
            end
        end
        for j in 1:nΘ
            ρ_θ = ρ_Θ[j,j]
            A[:,:,j] .+= ω[i].*(DR[:x]*x_ΔΘX[:,:,j,i] + DR[:X] + DR[:X′]*IX′X*(dX′[:Λ]*IΛX+ρ_θ*I))
        end

        B .+= -ω[i]*(R_ΔΘ + DR[:x]*x_ΔΘ[:,:,:,i])
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)

    #Now compute the aggregate responce
    X_ΔΘ = zeros(Float,nX,nΔ,nΘ)
    for j in 1:nΘ
        X_ΔΘ[:,:,j] .= A[:,:,j]\B[:,:,j]
    end
    for i in 1:N
        for j in 1:nΘ
            x_ΔΘ[:,:,j,i] .+= x_ΔΘX[:,:,j,i]* X_ΔΘ[:,:,j]
        end
        approx.d2x[i][sΔ,:Θ] = x_ΔΘ[:,:,:,i]
        approx.d2x[i][:Θ,sΔ] = permutedims(approx.d2x[i][sΔ,:Θ],[1,3,2])
    end
    approx.d2X[sΔ,:Θ] = X_ΔΘ
    approx.d2X[:Θ,sΔ] = permutedims(X_ΔΘ,[1,3,2])
end


"""
    compute_dx_dΘΘ!(approx::Approximation,approx′::Approximation)

Compute the cross derivatives with respect to Θ, exploiting that it won't affect
the distribution
"""
function compute_dx_dΘΘ!(approx::Approximation,approx′::Approximation)
    @unpack N,para,ω,dX,d2X,comm = approx
    @unpack ρ_Θ,Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΘ,nz,nX = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    #Prestore some variables for ease of use
    X_Θ,Λ_Θ = dX[:Θ],IΛX*dX[:Θ]

    d_Θ = Dict{Symbol,GMatrix}()
    d_Θ[:Θ],d_Θ[:X],d_Θ[:X′] =  I,dX[:Θ],IX′X*(dX′[:Θ]*ρ_Θ+dX′[:Λ]*Λ_Θ)
    #K and J from dΕ note
    A = zeros(Float,nX,nX)
    B = zeros(Float,nX,nΘ,nΘ)

    x_ΘΘ = zeros(Float,nx,nΘ,nΘ,N)
    x_ΘΘX = zeros(Float,nx,nX,N)

    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        DR,HR,dx′,d2x′ = approx.DR[i],approx.HR[i],approx′.dx[i],approx′.d2x[i]

        z_Θ = Izx*dx[:Θ]
        d_Θ[:x_],d_Θ[:x],d_Θ[:x′] = Ix_x*dx[:Θ],dx[:Θ],Ix′x*(dx′[:z]*z_Θ + dx′[:Θ]*ρ_Θ + dx′[:Λ]*Λ_Θ)

        #Compute second order contribution to F
        F_ΘΘ = DF[:x′]*Ix′x*(0.5*d2x′[:z,:z]⋅(z_Θ,z_Θ) + d2x′[:z,:Θ]⋅(z_Θ,ρ_Θ)+ d2x′[:z,:Λ]⋅(z_Θ,Λ_Θ)
                                                     +0.5*d2x′[:Θ,:Θ]⋅(ρ_Θ,ρ_Θ)+ d2x′[:Θ,:Λ]⋅(ρ_Θ,Λ_Θ)
                                                                                +0.5*d2x′[:Λ,:Λ]⋅(Λ_Θ,Λ_Θ))
        F_ΘΘ += permutedims(F_ΘΘ,[1,3,2])

        for x1 in [:Θ,:X,:x_,:x,:x′]
            for x2 in [:Θ,:X,:x_,:x,:x′]
                @doinplace F_ΘΘ .+= HF[x1,x2]⋅(d_Θ[x1],d_Θ[x2])
            end
        end
        temp = inv(DF[:x_]*Ix_x + DF[:x] +DF[:x′]*Ix′x*dx′[:z]*Izx)
        #compute coefficients on X
        x_ΘΘ[:,:,:,i] .= -temp*F_ΘΘ
        x_ΘΘX[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)


        #Compute second order contribution to R
        R_ΘΘ = DR[:X′]*IX′X*(0.5*d2X′[:Θ,:Θ]⋅(ρ_Θ,ρ_Θ)+ d2X′[:Θ,:Λ]⋅(ρ_Θ,Λ_Θ)
                                                     +0.5*d2X′[:Λ,:Λ]⋅(Λ_Θ,Λ_Θ))

        R_ΘΘ += permutedims(R_ΘΘ,[1,3,2])
        for x1 in [:Θ,:X,:x,:X′]
            for x2 in [:Θ,:X,:x,:X′]
                @doinplace R_ΘΘ .+= HR[x1,x2]⋅(d_Θ[x1],d_Θ[x2])
            end
        end
                #Compute K and J
        A .+= ω[i].*(DR[:x]*x_ΘΘX[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        B .+= -ω[i].*(DR[:x]*x_ΘΘ[:,:,:,i] + R_ΘΘ )
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    #solve the linear system
    X_ΘΘ = inv(A)*B
    #store results
    approx.d2X[:Θ,:Θ] = X_ΘΘ
    for i in 1:N
        approx.d2x[i][:Θ,:Θ] = x_ΘΘ[:,:,:,i] + x_ΘΘX[:,:,i]*X_ΘΘ
    end
end


"""
    compute_dx_d2Ε!(approx::Approximation,approxPath′)

Compute the cross derivatives with respect to E
"""
function compute_dx_d2Ε!(approx::Approximation,approxPath)
    @unpack N,para,ω,dX,d2X,comm,K = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΕ,nz,nX,Fmeasure,Gmeasure = para
    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    approx′ = approxPath[1]
    dX′,d2X′ = approx′.dX,approx′.d2X
    #compute distributional derivatives
    sΩ_Ε = Symbol(:Ω_Ε,K)
    Ω_Ε(i) = Izx*approx.dx[i][:Ε]
    if length(approxPath) == 1
        compute_z∂!(approx′,Ω_Ε,sΩ_Ε)
        compute_∂∂!(approx′,Ω_Ε,sΩ_Ε,Ω_Ε,sΩ_Ε)
        compute_∂X_Θ!(approx′,Ω_Ε,sΩ_Ε)
        compute_∂X_Λ!(approx′,Ω_Ε,sΩ_Ε)
    elseif approx′.K==1
        compute_z∂_final!(approx′,approxPath[2],Ω_Ε,sΩ_Ε)
        compute_∂∂_final!(approx′,approxPath[2],Ω_Ε,sΩ_Ε,Ω_Ε,sΩ_Ε)
        compute_∂X_Θ_final!(approx′,approxPath[2],Ω_Ε,sΩ_Ε)
        compute_∂X_Λ_final!(approx′,approxPath[2],Ω_Ε,sΩ_Ε)
    else
        approxPath′ = @view approxPath[2:end]
        compute_z∂!(approx′,approxPath′,Ω_Ε,sΩ_Ε)
        compute_∂∂!(approx′,approxPath′,Ω_Ε,sΩ_Ε,Ω_Ε,sΩ_Ε)
        compute_∂X_Θ!(approx′,approxPath′,Ω_Ε,sΩ_Ε)
        compute_∂X_Λ!(approx′,approxPath′,Ω_Ε,sΩ_Ε)
    end
    X_Ε,Λ_Ε = dX[:Ε],IΛX*dX[:Ε]
    d_Ε = Dict{Symbol,GMatrix}()
    d_Ε[:Ε],d_Ε[:X],d_Ε[:X′] = I,dX[:Ε],IX′X*(dX′[:Θ]+dX′[sΩ_Ε]+dX′[:Λ]*Λ_Ε)

    #coefficients for determining X_ΕΕ
    A = zeros(Float,K*nX,K*nX)
    B = zeros(Float,K*nX,nΕ,nΕ)
    x_ΕΕ = zeros(Float,nx,nΕ,nΕ,N)
    x_X = zeros(Float,nx,nX,N)
    x_C′ = zeros(Float,nx,nx,N)


    for i in 1:N
        DF,HF,dx,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.d2x[i]
        DR,HR,dx′,d2x′ = approx.DR[i],approx.HR[i],approx′.dx[i],approx′.d2x[i]
        C′ = approx′.C[i]
        #preconstruct derivative terms
        z_Ε = Izx*dx[:Ε]
        d_Ε[:x] = dx[:Ε]
        d_Ε[:x′] = Ix′x*(dx′[:z]*z_Ε+dx′[:Θ]+dx′[sΩ_Ε]+dx′[:Λ]*Λ_Ε)

        #Compute second order contribution to F
        F_ΕΕ = DF[:x′]*Ix′x*(.5*d2x′[:z,:z]⋅(z_Ε,z_Ε) + d2x′[:z,sΩ_Ε]⋅(z_Ε,I) + d2x′[:z,:Θ]⋅(z_Ε,I) + d2x′[:z,:Λ]⋅(z_Ε,Λ_Ε)
                             +0.5*d2x′[sΩ_Ε,sΩ_Ε] + d2x′[sΩ_Ε,:Θ] + d2x′[sΩ_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2x′[:Θ,:Θ]+d2x′[:Θ,:Λ]⋅(I,Λ_Ε)+0.5*d2x′[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))

        F_ΕΕ += permutedims(F_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ΕΕ .+= HF[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end

        R_ΕΕ = DR[:X′]*IX′X*(0.5*d2X′[sΩ_Ε,sΩ_Ε] + d2X′[sΩ_Ε,:Θ] + d2X′[sΩ_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2X′[:Θ,:Θ]  + d2X′[:Θ,:Λ]⋅(I,Λ_Ε) + 0.5*d2X′[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))
        R_ΕΕ += permutedims(R_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:X′]
            for x2 in [:Ε,:X,:x,:X′]
                @doinplace R_ΕΕ .+= HR[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #apply measurablility restriction
        temp = inv(temp)
        #compute coefficients on X for ΕΕ
        x_ΕΕ[:,:,:,i] .= -temp*F_ΕΕ
        x_X[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_C′[:,:,i] .= -temp*DF[:x′]*Ix′x

        #Compute K and J
        A[1:nX,1:nX] .+= ω[i].*(DR[:x]*x_X[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        A[1:nX,nX+1:2*nX] .+= ω[i]*DR[:X′]*IX′X
        A[1:nX,nX+1:end] .+= ω[i]*(DR[:x]*x_C′[:,:,i]*C′)
        B[1:nX,:,:] .+= -ω[i]*(DR[:x]*x_ΕΕ[:,:,:,i] + R_ΕΕ)
        for k in 1:K-1
            approx_k = approxPath[k]
            Dkinv,Ek,DRk,dxk = approx_k.Dinv,approx_k.E,approx_k.DR[i],approx_k.dx[i]
            HRk,d2xk,d2Xk = approx_k.HR[i],approx_k.d2x[i],approx_k.d2X
            #R_zz = HRk[:z,:z]+HRk[:z,:x]⋅(I,dxk[:z]) + HRk[:x,:z]⋅(dxk[:z],I) +HRk[:x,:x]⋅(dxk[:z],dxk[:z]) + DRk[:x]*d2xk[:z,:z]
            R_z = DRk[:z] + DRk[:x]*dxk[:z]

            A[k*nX+1:(k+1)*nX,1:nX] .+= ω[i]*Dkinv*R_z*Izx*x_X[:,:,i]
            A[k*nX+1:(k+1)*nX,nX+1:end] .+= ω[i]*Dkinv*R_z*Izx*x_C′[:,:,i]*C′
            A[k*nX+1:(k+1)*nX,k*nX+1:(k+1)*nX] .+= ω[i]*eye(nX)
            if k < K-1
                A[k*nX+1:(k+1)*nX,(k+1)*nX+1:end] .+= ω[i]*Dkinv*Ek
            end
            B[k*nX+1:(k+1)*nX,:,:] .+= -ω[i]*Dkinv*(R_z*Izx*x_ΕΕ[:,:,:,i])# + R_zz⋅(z_Ε,z_Ε)
        end
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)
    #apply measurablility
    A[1:nG_,1:nX] .= Gmeasure
    #solve the linear system
    XX′_ΕΕ = inv(A)*B
    X_ΕΕ = XX′_ΕΕ[1:nX,:,:]
    X′_ΕΕ = XX′_ΕΕ[nX+1:end,:,:]

    #store results
    approx.d2X[:Ε,:Ε] = X_ΕΕ
    for i in 1:N
        approx.d2x[i][:Ε,:Ε] = x_ΕΕ[:,:,:,i] + x_X[:,:,i]*X_ΕΕ + x_C′[:,:,i]*approx′.C[i]*X′_ΕΕ
    end
end


"""
    compute_dx_d2Ε!(approx::Approximation)

Compute the cross derivatives with respect to E
"""
function compute_dx_d2Ε_final!(approx::Approximation,approx′)
    @unpack N,para,ω,dX,d2X,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nx,nΕ,nz,nX,Fmeasure,Gmeasure = para
    dX′,d2X′ = approx′.dX,approx′.d2X
    nF_,nG_ = size(Fmeasure,1),size(Gmeasure,1)
    #precompute some of the cross derivatives
    Ω_Ε(i) = Izx*approx.dx[i][:Ε]
    #compute the derivative  along Ω_Θ and Ω-Ε
    compute_∂∂_final!(approx,approx′,Ω_Ε,:Ω_Ε,Ω_Ε,:Ω_Ε)
    compute_∂X_Θ_final!(approx,approx′,Ω_Ε,:Ω_Ε)
    compute_∂X_Λ_final!(approx,approx′,Ω_Ε,:Ω_Ε)
    compute_z∂_final!(approx,approx′,Ω_Ε,:Ω_Ε)
    #Prestore some variables for ease of use
    X_ΩΕ = dX[:Ω_Ε]
    X_Ε,Λ_Ε = dX[:Ε],IΛX*dX[:Ε]

    d_Ε = Dict{Symbol,GMatrix}()
    d_Ε[:Ε],d_Ε[:X],d_Ε[:X′] = I,dX[:Ε],IX′X*(dX′[:Θ]+X_ΩΕ+dX′[:Λ]*Λ_Ε)

    #coefficients for determining X_ΕΕ
    x_ΕΕ = zeros(Float,nx,nΕ,nΕ,N)
    x_X = zeros(Float,nx,nX,N)
    x_X′ = zeros(Float,nx,nX,N)
    A1 = zeros(Float,nX,2*nX)
    A2 = zeros(Float,nX,2*nX)
    C1 = zeros(Float,nX,nΕ,nΕ)
    C2 = zeros(Float,nX,nΕ,nΕ)

    for i in 1:N
        DF,HF,dx,∂x,d2x = approx.DF[i],approx.HF[i],approx.dx[i],approx.∂x[i],approx.d2x[i]
        DR,HR,dx′,d2x′ = approx.DR[i],approx.HR[i],approx′.dx[i],approx′.d2x[i]

        #preconstruct derivative terms

        z_Ε = Izx*dx[:Ε]
        d_Ε[:x] = dx[:Ε]
        d_Ε[:x′] = Ix′x*(dx′[:z]*z_Ε+dx′[:Θ]+∂x[:X]*X_ΩΕ+dx′[:Λ]*Λ_Ε)

        F_ΕΕ = DF[:x′]*Ix′x*(.5*d2x′[:z,:z]⋅(z_Ε,z_Ε) + d2x[:z,:Ω_Ε]⋅(z_Ε,I) + d2x′[:z,:Θ]⋅(z_Ε,I) + d2x′[:z,:Λ]⋅(z_Ε,Λ_Ε)
                             +0.5*d2x[:Ω_Ε,:Ω_Ε] + d2x[:Ω_Ε,:Θ] + d2x[:Ω_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2x′[:Θ,:Θ]+d2x′[:Θ,:Λ]⋅(I,Λ_Ε)+0.5*d2x′[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))

        F_ΕΕ += permutedims(F_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:x′]
            for x2 in [:Ε,:X,:x,:x′]
                @doinplace F_ΕΕ .+= HF[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end

        R_ΕΕ = DR[:X′]*IX′X*(0.5*d2X[:Ω_Ε,:Ω_Ε] + d2X[:Ω_Ε,:Θ] + d2X[:Ω_Ε,:Λ]⋅(I,Λ_Ε)
                             +0.5*d2X′[:Θ,:Θ]  + d2X′[:Θ,:Λ]⋅(I,Λ_Ε) + 0.5*d2X′[:Λ,:Λ]⋅(Λ_Ε,Λ_Ε))
        R_ΕΕ += permutedims(R_ΕΕ,[1,3,2])
        for x1 in [:Ε,:X,:x,:X′]
            for x2 in [:Ε,:X,:x,:X′]
                @doinplace R_ΕΕ .+= HR[x1,x2]⋅(d_Ε[x1],d_Ε[x2])
            end
        end


        ∂X = -approx.Dinv*(DR[:z] + DR[:x]*dx[:z])

        temp = DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx
        temp[1:nF_,:] = Fmeasure #apply measurablility restriction
        temp = inv(temp)
        #compute coefficients on X for ΕΘ and ΕΕ
        x_ΕΕ[:,:,:,i] .= -temp*F_ΕΕ
        x_X[:,:,i] .= -temp*(DF[:X] + DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_X′[:,:,i] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        #Compute K and J
        A1[:,1:nX] .+= ω[i].*(DR[:x]*x_X[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        A1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′[:,:,i] + DR[:X′]*IX′X)
        A2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_X[:,:,i])
        A2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′[:,:,i])
        C1 .+= -ω[i].*(DR[:x]*x_ΕΕ[:,:,:,i] + R_ΕΕ )
        C2 .+=  ω[i].*(∂X*Izx*x_ΕΕ[:,:,:,i])#+X_zz⋅(z_Ε,z_Ε))
    end
    MPI.Allreduce!(A1,+,comm)
    MPI.Allreduce!(A2,+,comm)
    MPI.Allreduce!(C1,+,comm)
    MPI.Allreduce!(C2,+,comm)
    #apply measurablility
    A1[1:nG_,1:nX] = Gmeasure
    #solve the linear system
    XX′_ΕΕ = inv([A1 ; A2]) * [C1 ; C2]
    X_ΕΕ = XX′_ΕΕ[1:nX,:,:]
    X′_ΕΕ = XX′_ΕΕ[nX+1:end,:,:]

    #store results
    approx.d2X[:Ε,:Ε] = X_ΕΕ
    for i in 1:N
        approx.d2x[i][:Ε,:Ε] = x_ΕΕ[:,:,:,i] + x_X[:,:,i]*X_ΕΕ + x_X′[:,:,i]*X′_ΕΕ
    end
end

"""
    compute_σσ!(approx::Approximation,approxPath)

Computes the second derivative with respect to σ
"""
function compute_σσ!(approx::Approximation,approxPath)
    @unpack N,para,ω,comm,K = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nϵ,nx,nz,nX,Σ_ϵ,Σ_Ε = para
    approx′ = approxPath[1]
    dX′,d2X′ = approx′.dX,approx′.d2X

    x_σσ = zeros(Float,nx,N)
    x_X = zeros(Float,nx,nX,N)
    x_C′ = zeros(Float,nx,nx,N)

    #KK and JJ from appendix
    A = zeros(Float,K*nX,K*nX)
    B = zeros(Float,K*nX)


    for i in 1:N
        DF,DR,dx,d2x = approx.DF[i],approx.DR[i],approx.dx[i],approx.d2x[i]
        HF,HR,dx′,d2x′ = approx.HF[i],approx.HR[i],approx′.dx[i],approx′.d2x[i]
        C′ = approx′.C[i]

        temp = inv( DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx)

        #prestore some derivatives
        x_ϵϵ = Ix_x*d2x[:ϵ,:ϵ]
        x′ϵϵ = Ix′x*d2x′[:ϵ,:ϵ]
        x_ΕΕ = Ix_x*d2x[:Ε,:Ε]
        x′ΕΕ = Ix′x*d2x′[:Ε,:Ε]
        X′ΕΕ = IX′X*d2X′[:Ε,:Ε]
        X′σσ = IX′X*d2X′[:σ,:σ][:]
        zϵϵ = Izx*d2x[:ϵ,:ϵ]
        zϵ = Izx*dx[:ϵ]
        #compute coefficients on X
        x_σσ[:,i] .= -temp*( (DF[:x_]*x_ϵϵ  + DF[:x′]*x′ϵϵ)⋅Σ_ϵ + (DF[:x_]*x_ΕΕ  + DF[:x′]*x′ΕΕ)⋅Σ_Ε + DF[:x′]*Ix′x*d2x′[:σ,:σ][:] )
        x_X[:,:,i] .= -temp*(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_C′[:,:,i] .= -temp*DF[:x′]*Ix′x

        R_ϵϵ = 0.5*HR[:x,:x]⋅(dx[:ϵ],dx[:ϵ]) + HR[:ϵ,:x]⋅(I,dx[:ϵ]) + 0.5*HR[:ϵ,:ϵ] + 0.5*DR[:x]*d2x[:ϵ,:ϵ]
        R_ϵϵ += permutedims(R_ϵϵ,[1,3,2])
        #Now compute K and J
        A[1:nX,1:nX] .+= ω[i].*(DR[:x]*x_X[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX)
        A[1:nX,nX+1:2*nX] .+= ω[i]*DR[:X′]*IX′X
        A[1:nX,nX+1:end] .+= ω[i]*(DR[:x]*x_C′[:,:,i]*C′)
        B[1:nX] .+= -ω[i]*(DR[:x]*x_σσ[:,i] + R_ϵϵ⋅Σ_ϵ + DR[:X′]*(X′ΕΕ⋅Σ_Ε + X′σσ) )

        for k in 1:K-1
            approx_k = approxPath[k]
            Dkinv,Ek,DRk,dxk = approx_k.Dinv,approx_k.E,approx_k.DR[i],approx_k.dx[i]
            HRk,d2xk = approx_k.HR[i],approx_k.d2x[i]
            xk_z = dxk[:z]
            Rk_zz = DRk[:x]*d2xk[:z,:z] + HRk[:z,:z] + HRk[:z,:x]⋅(I,xk_z) + HRk[:x,:z]⋅(xk_z,I) + HRk[:x,:x]⋅(xk_z,xk_z)
            Rk_z = DRk[:z]+DRk[:x]*dxk[:z]

            A[k*nX+1:(k+1)*nX,1:nX] .+= ω[i]*Dkinv*Rk_z*Izx*x_X[:,:,i]
            A[k*nX+1:(k+1)*nX,nX+1:end] .+= ω[i]*Dkinv*Rk_z*Izx*x_C′[:,:,i]*C′
            A[k*nX+1:(k+1)*nX,k*nX+1:(k+1)*nX] .+= ω[i]*eye(nX)
            if k < K-1
                A[k*nX+1:(k+1)*nX,(k+1)*nX+1:end] .+= ω[i]*Dkinv*Ek
            end
            B[k*nX+1:(k+1)*nX] .+= -ω[i]*Dkinv*(Rk_z*(Izx*x_σσ[:,i] + zϵϵ⋅Σ_ϵ) + (Rk_zz⋅(zϵ,zϵ))⋅Σ_ϵ)
        end
    end
    MPI.Allreduce!(A,+,comm)
    MPI.Allreduce!(B,+,comm)

    XX′_σσ = A \ B
    X_σσ = XX′_σσ[1:nX]
    X′_σσ = XX′_σσ[nX+1:end]

    approx.d2X[:σ,:σ] = reshape(X_σσ,:,1,1)
    #store results
    for i in 1:N
        approx.d2x[i][:σ,:σ] = reshape(x_σσ[:,i] + x_X[:,:,i]*X_σσ + x_C′[:,:,i]*approx′.C[i]*X′_σσ,:,1,1)
    end
end



"""
    compute_σσ_final!(approx::Approximation,approx′)

Computes the second derivative with respect to σ
"""
function compute_σσ_final!(approx::Approximation,approx′)
    @unpack N,para,ω,comm = approx
    @unpack Ix_x,Ix′x,Izx,IX′X,IΛX,nϵ,nx,nz,nX,Σ_ϵ,Σ_Ε = para
    dX′,d2X′ = approx′.dX,approx′.d2X

    x_σσ = zeros(Float,nx,N)
    x_Xσσ = zeros(Float,nx,nX,N)
    x_X′σσ = zeros(Float,nx,nX,N)

    #KK and JJ from appendix
    LL1 = zeros(Float,nX,2*nX)
    LL2 = zeros(Float,nX,2*nX)
    MM1 = zeros(Float,nX,1)
    MM2 = zeros(Float,nX,1)
    R_ϵϵ = zeros(Float,nX,nϵ,nϵ)


    for i in 1:N
        DF,DR,dx,dX,∂x,∂X = approx.DF[i],approx.DR[i],approx.dx[i],approx.dX, approx.∂x[i],approx.∂X[i]
        HF,HR,d2x,d2X = approx.HF[i],approx.HR[i],approx.d2x[i],approx.d2X
        dx′,d2x′ = approx′.dx[i],approx′.d2x[i]

        temp = inv( DF[:x_]*Ix_x+DF[:x]+DF[:x′]*Ix′x*dx′[:z]*Izx )

        #prestore some derivatives
        x_ϵϵ = Ix_x*d2x[:ϵ,:ϵ]
        x′ϵϵ = Ix′x*d2x′[:ϵ,:ϵ]
        x_ΕΕ = Ix_x*d2x[:Ε,:Ε]
        x′ΕΕ = Ix′x*d2x[:Ε,:Ε]
        X′ΕΕ = IX′X*d2X[:Ε,:Ε]
        X′σσ = IX′X*d2X′[:σ,:σ][:]
        zϵϵ = Izx*d2x[:ϵ,:ϵ]
        zϵ = Izx*dx[:ϵ]
        #compute coefficients on X
        x_σσ[:,i] .= -temp*( (DF[:x_]*x_ϵϵ  + DF[:x′]*x′ϵϵ)⋅Σ_ϵ + (DF[:x_]*x_ΕΕ  + DF[:x′]*x′ΕΕ)⋅Σ_Ε + DF[:x′]*Ix′x*d2x′[:σ,:σ][:] )
        x_Xσσ[:,:,i] .= -temp*(DF[:X]+DF[:x′]*Ix′x*dx′[:Λ]*IΛX)
        x_X′σσ[:,:,i] .= -temp*DF[:x′]*Ix′x*∂x[:X]

        R_ϵϵ = 0.5*HR[:x,:x]⋅(dx[:ϵ],dx[:ϵ]) + HR[:ϵ,:x]⋅(I,dx[:ϵ]) + 0.5*HR[:ϵ,:ϵ] + 0.5*DR[:x]*d2x[:ϵ,:ϵ]
        R_ϵϵ += permutedims(R_ϵϵ,[1,3,2])


        x_z = dx[:z]
        ∂X′_zz = -approx.Dinv*(DR[:x]*d2x[:z,:z] + HR[:z,:z] + HR[:z,:x]⋅(I,x_z) + HR[:x,:z]⋅(x_z,I) + HR[:x,:x]⋅(x_z,x_z))

        #Compute K and J
        LL1[:,1:nX] .+= ω[i].*(DR[:x]*x_Xσσ[:,:,i] + DR[:X] + DR[:X′]*IX′X*dX′[:Λ]*IΛX )
        LL1[:,nX+1:end] .+= ω[i].*(DR[:x]*x_X′σσ[:,:,i] + DR[:X′]*IX′X)
        LL2[:,1:nX] .+=  ω[i].*(-∂X*Izx*x_Xσσ[:,:,i])
        LL2[:,nX+1:end] .+=  ω[i].*(I-∂X*Izx*x_X′σσ[:,:,i])
        MM1 .+= -ω[i].*(DR[:x]*x_σσ[:,i] + R_ϵϵ⋅Σ_ϵ + DR[:X′]*(X′ΕΕ⋅Σ_Ε + X′σσ))
        MM2 .+=  ω[i].*(∂X*(Izx*x_σσ[:,i]+zϵϵ⋅Σ_ϵ) + (∂X′_zz⋅(zϵ,zϵ))⋅Σ_ϵ)
    end
    MPI.Allreduce!(LL1,+,comm)
    MPI.Allreduce!(LL2,+,comm)
    MPI.Allreduce!(MM1,+,comm)
    MPI.Allreduce!(MM2,+,comm)

    XX′_σσ = [LL1 ; LL2] \ [MM1 ; MM2]
    X_σσ = XX′_σσ[1:nX,:]
    X′_σσ = XX′_σσ[nX+1:end,:]

    approx.d2X[:σ,:σ] = reshape(X_σσ,:,1,1)
    #store results
    for i in 1:N
        approx.d2x[i][:σ,:σ] = reshape(x_σσ[:,i] + x_Xσσ[:,:,i]*X_σσ + x_X′σσ[:,:,i]*X′_σσ,:,1,1)
    end
end

"""
    simulatepolicies(approx::Approximation, Ω::AgentDist,Θ,Ε,ϵ)

Computes the policy rules for X and for x
"""
function simulatepolicies(approxPath::ApproximationPath,Ε,ϵ)
    approx = approxPath.approx[1]
    @unpack x̄,X̄,para,dX,d2X,comm = approx
    @unpack IΘX,nx = para

    #compute X
    X = X̄+ dX[:Ε]*Ε + 0.5*(d2X[:Ε,:Ε]⋅(Ε,Ε)+d2X[:σ,:σ][:])

    #compute x
    my_N = size(x̄,2)
    my_x = similar(x̄)
    for i in 1:my_N
        dx,d2x = approx.dx[i],approx.d2x[i]
        ϵ̂ = ϵ[:,i]
        temp = 0.5*d2x[:ϵ,:ϵ]⋅(ϵ̂,ϵ̂) + d2x[:ϵ,:Ε]⋅(ϵ̂,Ε) +0.5*d2x[:Ε,:Ε]⋅(Ε,Ε)+ 0.5*d2x[:σ,:σ][:]
        my_x[:,i] .= x̄[:,i] + dx[:ϵ]*ϵ̂ + dx[:Ε]*Ε + temp
    end

    x = MPI.Gather(my_x,0,comm)
    if MPI.Comm_rank(comm) == 0
        return X,reshape(x,nx,:)
    else
        return X,my_x
    end
end


"""
    simulatepolicies_complex(approx::Approximation, Ω::AgentDist,Θ,Ε,ϵ)

Computes the policy rules for X and for x
"""
function simulatepolicies_complex(approxPath::ApproximationPath,Ε,ϵ)
    approx = approxPath.approx[1]
    @unpack z̄,x̄,X̄,para,dX,d2X,comm,Ω,labels,my_i,ω = approx
    @unpack Izx,IΘX,nx,nz = para
    my_Ω,my_ω = Ω
    #compute X
    X = X̄+ dX[:Ε]*Ε + 0.5*(d2X[:Ε,:Ε]⋅(Ε,Ε)+d2X[:σ,:σ][:])

    #compute x
    my_N = length(labels)
    my_x = zeros(Float,nx,my_N)
    for i in 1:my_N
        j = labels[i]
        dx,d2x = approx.dx[j],approx.d2x[j]
        sdϵ = sqrt.(diag(para.Σ_ϵ))
        ϵ̂ = min.(max.(ϵ[:,i],-3 .* sdϵ),3 .* sdϵ)
        ẑ = my_Ω[i,:] - z̄[:,j]
        temp = (0.5*d2x[:ϵ,:ϵ]⋅(ϵ̂,ϵ̂) + d2x[:ϵ,:Ε]⋅(ϵ̂,Ε) +0.5*d2x[:Ε,:Ε]⋅(Ε,Ε)+ 0.5*d2x[:σ,:σ][:] +0.5*d2x[:z,:z]⋅(ẑ,ẑ) + d2x[:z,:ϵ]⋅(ẑ,ϵ̂) + d2x[:z,:Ε]⋅(ẑ,Ε) )
        my_x[:,i] .= x̄[:,j] + dx[:z]*ẑ + dx[:ϵ]*ϵ̂ + dx[:Ε]*Ε + temp
    end
    my_z = Izx*my_x


    x = MPI_GatherArray(my_x,comm)
    z = MPI_GatherArray(my_z,comm)
    ω = MPI_GatherArray(my_ω,comm)
    i_sort = sortperm(MPI_GatherArray(my_i,comm))
    if MPI.Comm_rank(comm) == 0
        return convert(Vector{Float},X),convert(Matrix{Float},x[:,i_sort]),(convert(Matrix{Float},z'[i_sort,:]),convert(Vector{Float},ω[i_sort]))
    else
        return convert(Vector{Float},X),convert(Matrix{Float},my_x),(convert(Matrix{Float},my_z),convert(Vector{Float},my_ω))
    end
end

#
#
#
#
#
# ###SOME STUF ONLY FOR THE CARA CASE
# function simulate_irf(approx::Approximation,Ω::AgentDist,Ε0,T,s_ϵ=1.)
#     para = approx.para
#     @unpack nϵ,nx,nX,IΘX = para
#     N = length(Ω[2])
#     ϵ = ones(N,1)*sqrt(para. Σ_ϵ[1])*s_ϵ
#
#     X2 = zeros(Float,nX,T)
#     x2 = zeros(Float,N,nx,T)
#     Θ = [0.]
#     Ε = [Ε0]
#     for t in 1:T
#         approximate!(approx,Ω)
#         X2[:,t],x2[:,:,t] = simulatepolicies(approx,Ω,Θ,Ε,ϵ)
#         Θ = IΘX*X2[:,t]
#         Ε = [0.]
#     end
#
#     X1 = zeros(nX,T)
#     x1 = zeros(N,nx,T)
#     Θ = [0.]
#     Ε = [0.]
#     for t in 1:T
#         X1[:,t],x1[:,:,t] = simulatepolicies(approx,Ω,Θ,Ε,ϵ)
#         Θ = IΘX*X1[:,t]
#         Ε = [0.]
#     end
#
#     return X2.-X1,x2.-x1
# end
