module CalibrationCARA_NK
  using NLsolve
  using Parameters
  using SymEngine
  using LinearAlgebra

  @with_kw mutable struct CARAmodelNK

    #number of dimensions of each variable
    nz::Int  = 0
    nx::Int  = 0
    nx_::Int = 0
    nx′::Int = 0
    nX::Int  = 0
    nΘ::Int  = 0
    nΛ::Int = 0
    nX′::Int = 0
    nϵ::Int = 0
    nΕ::Int = 0
    nw::Int = 0

    #Parameters
    β::Float64 = 0.96
    α::Float64 = 0.6
    ϕ::Float64 = 1.5
    γ::Float64 = 3.
    Φ::Float64 = 6.
    τ::Float64 = 0.25
    δ::Float64 = 0.
    ψ::Float64 = 41.
    B̄::Float64 = 1.
    Gov::Float64 = 0.05
    θ̄::Float64 = log(2.)


    ρ_θ::Float64 = 0.73
    ρ_Θ::Matrix{Float64} = ones(1,1)*ρ_θ
    Σ_ϵ::Matrix{Float64} = .25^2 * ones(1,1)
    Σ_Ε::Matrix{Float64} = .03^2 * ones(1,1)
    Npath::Int=200
    X̄_Λ0::Matrix{Float64} = zeros(0,0)

    #Compiled Functions
    DF::Function = x->0
    DR::Function = x->0
    HF::Function = x->0
    HR::Function = x->0
    steadyState::Function = x->0
    getw̄::Function = x->0
    computePath::Function = x->0


    #Projection matrices
    Izx::Matrix{Float64} = zeros(0,0)
    Ix′x::Matrix{Float64} = zeros(0,0)
    Ix_x::Matrix{Float64} = zeros(0,0)
    IΘX::Matrix{Float64} = zeros(0,0)
    IΛX::Matrix{Float64} = zeros(0,0)
    IX′X::Matrix{Float64} = zeros(0,0)
    Fmeasure::Matrix{Float64} = zeros(0,0)
    Gmeasure::Matrix{Float64} = zeros(0,0)

    #indexdictionary
    ind::Dict{Symbol,Vector{Int}} = Dict{Symbol,Vector{Int}}()

    #variables holding the symbols
    zlab::Vector{Symbol}
    xlab::Vector{Symbol}
    x_lab::Vector{Symbol}
    x′lab::Vector{Symbol}
    Xlab::Vector{Symbol}
    Θlab::Vector{Symbol}
    Λlab::Vector{Symbol}
    X′lab::Vector{Symbol}
  end

  """
    stripsymbol(s)

  Strips the symbol `s` of any _ or \` following it.
  """
  function stripsymbol(sym)
    s = String(sym)
    if endswith(s,"′") || endswith(s,"_")
      return Symbol(s[1:length(s)-1])
    end
    return sym
  end

  """
      getMeasurabilityMatrix(xsym)

  Constrct a measurablility matrix given a vector of sybols
  """
  function getMeasurabilityMatrix(xsym)
    strippedx = stripsymbol.(xsym)
    xmeas = strippedx[strippedx .!= xsym]
    return Int.(xmeas .== reshape(strippedx,1,:))
  end


  """
  Constructs the CARAmodelNK environment.
  """
  function CARAmodelNK()
    para = CARAmodelNK(zlab = zlab,xlab = xlab,x_lab = x_lab,
    x′lab=x′lab,Xlab=Xlab,Θlab=Θlab,X′lab=X′lab,Λlab=Λlab)

    para.DF = (ret,wbar) -> DF(ret,wbar,para)
    para.HF = (ret,wbar) -> HF(ret,wbar,para)
    para.DR = (ret,wbar) -> DR(ret,wbar,para)
    para.HR = (ret,wbar) -> HR(ret,wbar,para)


    para.steadyState = (Ω,ω) -> findSteadyState(para,Ω,ω)
    para.getw̄ = (x̄,X̄) -> getw̄(para,x̄,X̄)
    para.computePath = (Ω,Θ,Λ) -> solvePath(para,Ω,Θ[1])

    #get length of all the variables
    para.nz,para.nx,para.nx_,para.nx′,para.nX,para.nΘ,para.nΛ,para.nX′,para.nϵ,para.nΕ = map(length,[z,x,x_,x′,X,Θ,Λ,X,ϵ,Ε])
    para.nw = length(w)
    #projection matrixes
    para.Izx = Int.(stripsymbol.(zlab) .== reshape(xlab,1,:))
    para.Ix′x = Int.(stripsymbol.(x′lab) .== reshape(xlab,1,:))+ Int.(x′lab .== reshape(stripsymbol.(xlab),1,:))
    para.Ix_x = Int.(stripsymbol.(x_lab) .== reshape(xlab,1,:))
    para.IΘX = Int.(stripsymbol.(Θlab) .== reshape(Xlab,1,:))
    para.IΛX = Int.(stripsymbol.(Λlab) .== reshape(Xlab,1,:))
    para.IX′X = Int.(stripsymbol.(X′lab) .== reshape(Xlab,1,:))+ Int.(X′lab .== reshape(stripsymbol.(Xlab),1,:))
    #measurablility constraints
    para.Fmeasure = getMeasurabilityMatrix(xlab)
    para.Gmeasure = getMeasurabilityMatrix(Xlab)
    #construct index dictionary
    para.ind[:z] = (zlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:x] = (xlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:x_] = (x_lab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:x′] = (x′lab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:X] = (Xlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:X′] = (X′lab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:Θ] = (Θlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:Λ] = (Λlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:ϵ] = (ϵlab .== reshape(wlab,1,:))*(1:para.nw)
    para.ind[:Ε] = (Εlab .== reshape(wlab,1,:))*(1:para.nw)
    para.X̄_Λ0 = zeros(para.nX,para.nΛ)
    para.X̄_Λ0[2,1] = 0.1

    return para
  end

  """
  Constructs function based on Array of symbols
  """
  function constructFunction(name::Symbol,A::Array{Basic},vars::Array{Basic},parameters::Array{Basic})
    sizeA = size(A)
    nA = length(A)
    vars = map(Symbol,vars)
    parameters = map(Symbol,parameters)

    #meta programming
    #unpack parameters line
    unpack = :@unpack
    push!(unpack.args, Expr(:(=),Expr(:tuple,parameters...),:para))

    body = Expr(:block,unpack)
    push!(body.args, Expr(:(=),Expr(:tuple,vars...),:w))
    push!(body.args, :(@assert size(ret) == $sizeA) ) #make sure size is right
    push!(body.args, :(ret[:] .= 0))
    for i in 1:nA
      Aexpr = SymEngine.walk_expression(A[i])
      if A[i] != 0
        push!(body.args, :(ret[$i] = $Aexpr) )
      end
    end

    #Now lambdify
    return Expr(:function,
    Expr(:call,name,:ret,:w,:(para::CARAmodelNK)),
    body)
  end

  module CARA_Environment
    using SymEngine

    #Define symbols used in this environment
    z = [symbols("b_")...]
    x = [symbols("b,c,y,uc,ucπ")...]
    x_= [symbols("uc_")...]
    x′= [symbols("ucπ′")...]
    X = [symbols("θ,λ,Π,Z,Y,Q,W,D,Tr,ZΠ")...]
    Θ = [symbols("θ_")...]
    Λ = [symbols("λ_")]
    X′= [symbols("ZΠ′")...]
    ϵ = [symbols("ε")...]
    Ε = [symbols("ε_θ")...]
    w = vcat(z,x,x_,x′,Θ,Λ,X,X′,ϵ,Ε)

    zlab = map(Symbol,z)
    xlab = map(Symbol,x)
    x_lab = map(Symbol,x_)
    x′lab = map(Symbol,x′)
    Xlab = map(Symbol,X)
    Θlab = map(Symbol,Θ)
    Λlab = map(Symbol,Λ)
    X′lab = map(Symbol,X′)
    ϵlab = map(Symbol,ϵ)
    Εlab = map(Symbol,Ε)
    wlab = map(Symbol,w)


    parameters = [ symbols("β,α,ϕ,γ,Φ,τ,δ,ψ,B̄,ρ_θ,Gov,θ̄")...]

    #unpack variables
    wtuple = Expr(:tuple,map(Symbol,w)...)
    eval(Expr(:(=),wtuple,:w))
    paratuple = Expr(:tuple,map(Symbol,parameters)...)
    eval(Expr(:(=),paratuple,:parameters))

    Fhat =[uc - γ*exp(-γ*c),
           y - ((1-τ)*W*(1+ε) + Tr + (1+δ*ε)*D),
           c + Q*b - b_/(1+Π) - y,
           ucπ - uc/(1+Π),
           Q*uc - β*ucπ′]

    Rhat = [θ - ρ_θ*θ_ - ε_θ,
            1/Q - (1/β + ϕ*Π), #i=1/β+ϕ*Π
            W - (1-α)/α * (Z/exp(θ+θ̄))^(1/α),
            Y - (Z-(Z/exp(θ+θ̄))^(1/α)),
            D - (Y - W - 0.5*ψ*Z*Π^2),
            B̄/(1+Π) + Tr + Gov - τ*W - Q*B̄,
            c + Gov - (Y - 0.5*ψ*Z*Π^2),
            ZΠ - Z*(1+Π)*Π,
            (1-Φ*(1-α^(-1)*exp(-(θ+θ̄)/α)*Z^((1-α)/α)))*Z + β*ψ*ZΠ′ - ψ*ZΠ,
            λ - 0.1*λ_] #Q*(1+Π′)

    export z,x,x_,x′,X,X′,Θ,Λ,ϵ,Ε,w,Fhat,Rhat,zlab,xlab,x_lab,x′lab,Xlab,X′lab,Θlab,Λlab,ϵlab,Εlab,wlab,parameters
  end

  using .CARA_Environment

  DFhat = Array{Basic,2}(undef,length(x),length(w))
  HFhat = Array{Basic,3}(undef,length(x),length(w),length(w))
  #DHFhat = Array{Basic,4}(undef,length(y),length(w),length(w),length(w))
  for i in 1:length(x)
    for j in 1:length(w)
      DFhat[i,j] = diff(Fhat[i],w[j])
      for k in 1:length(w)
        HFhat[i,j,k] = diff(DFhat[i,j],w[k])
      end
    end
  end
  eval(constructFunction(:F,Fhat,w,parameters))
  eval(constructFunction(:DF,DFhat,w,parameters))
  eval(constructFunction(:HF,HFhat,w,parameters))

  DRhat = Array{Basic,2}(undef,length(X),length(w))
  HRhat = Array{Basic,3}(undef,length(X),length(w),length(w))
  #DHFhat = Array{Basic,4}(undef,length(y),length(w),length(w),length(w))
  for i in 1:length(X)
    for j in 1:length(w)
      DRhat[i,j] = diff(Rhat[i],w[j])
      for k in 1:length(w)
        HRhat[i,j,k] = diff(DRhat[i,j],w[k])
      end
    end
  end
  eval(constructFunction(:R,Rhat,w,parameters))
  eval(constructFunction(:DR,DRhat,w,parameters))
  eval(constructFunction(:HR,HRhat,w,parameters))


  """
    findSteadyState(para,Ω,ω)

  Find the steady state associated with a distribution Ω with weights
  ω.
  """
  function findSteadyState(para::CARAmodelNK,Ω,ω)
    @unpack β,α,γ,τ,δ,ψ,Φ,B̄,Gov,θ̄ = para
    N = length(ω)
    θ̄ = exp(θ̄)
    #need to compute aggregate first "θ,Π,Z,Y,Q,W,D,Tr"
    Z̄ = (α*θ̄^(1/α)*(Φ-1)/Φ)^(α/(1-α))
    Ȳ = Z̄-(Z̄/θ̄)^(1/α)
    Π̄ = 0.
    W̄ = (1-α)/α * (Z̄/θ̄)^(1/α)
    D̄ = Ȳ-W̄
    Q̄ = β
    T̄r = τ*W̄- Gov +(Q̄-1)*B̄
    ZΠ̄ = 0.

    #individual variables
    b̄ = Ω[:,1]
    ȳ  = ((1-τ)*W̄ .+ T̄r .+ D̄).*ones(length(b̄))
    c̄ = (1-β).*b̄ .+ ȳ
    ūc = γ.*exp.(-γ.*c̄)
    ūcπ = ūc


    #stack into vectors
    x = [b̄';c̄'; ȳ'; ūc'; ūcπ']
    X = [0.,0.,Π̄,Z̄,Ȳ,Q̄,W̄,D̄,T̄r,ZΠ̄]

    return (x=x,X=X)
  end


  """
      solvePath(para,θ0,Ω0,T=para.Npath)

  Solves for the transition path given initial conditions
  """
  function solvePath(para,Ω0,θ0,T=para.Npath)
    @unpack β,α,γ,τ,δ,ψ,Φ,B̄,Gov,ρ_θ,ϕ,θ̄ = para
    b0 = reshape(Ω0[1][:,1],1,:)
    ω = Ω0[2]
    N = length(ω)
    θ = θ0.*ρ_θ.^(1:T)
    x̄,X̄ = findSteadyState(para,Ω0[1],Ω0[2])
    Z̄,Π̄,Ȳ,T̄r,W̄,D̄ = X̄[para.Xlab.==:Z][1],X̄[para.Xlab.==:Π][1],X̄[para.Xlab.==:Y][1],X̄[para.Xlab.==:Tr][1],X̄[para.Xlab.==:W][1],X̄[para.Xlab.==:D][1]
    C̄ = Ȳ - Gov - 0.5*ψ*Z̄*Π̄^2
    ȳ = ((1-τ)*W̄ + T̄r + D̄).*ones(1,N)
    W,Y,Q,C,Π,Z =  zeros(T),zeros(T),zeros(T),zeros(T+1),zeros(T+1),zeros(T+1)
    Π[T+1],Z[T+1],C[T+1] = Π̄,Z̄,C̄
    for t in reverse(1:T)
      Π[t],Z[t] = solveΠZ(para,θ[t],Z[t+1],Π[t+1],C[t+1])
      Y[t] = Z[t]-(Z[t]/exp(θ[t]+θ̄))^(1/α)
      C[t] = Y[t] - Gov - 0.5*ψ*Z[t]*Π[t]^2
      i    = 1/β + ϕ * Π[t]
      Q[t] = 1/i
      W[t] = (1-α)/α * (Z[t]/exp(θ[t]+θ̄))^(1/α)
    end
    D  = Y .- W .- 0.5*ψ*Z[1:T].*Π[1:T].^2
    Tr = τ.*W .+ B̄.*Q .- Gov .- B̄./(1 .+ Π[1:T])
    ZΠ = Z[1:T].*(1 .+ Π[1:T]).*Π[1:T]
    rinv = Q.*(1 .+ Π[2:end])
    Rinv= [1;cumprod(rinv)]

    y = [((1-τ).*W .+ Tr .+ D).*ones(1,N);ȳ/(1-β)]
    Cfactor = 1 ./(Rinv'*[ones(T);1/(1-β)])
    wealth_hat = (b0 .- para.B̄)./(1+Π[1])
    c̄ = C̄ .+ wealth_hat.*Cfactor
    c = [C[1:T] .+ wealth_hat.*Cfactor;c̄/(1-β)]
    surplus = c-y
    Qb = vcat([cumprod(rinv[t:end])'*surplus[t+1:end,:] for t in 1:T]...)::Matrix{Float64}
    b = Qb./Q
    uc = γ.*exp.(-γ.*c[1:T,:])
    ucπ = uc./(1 .+ Π[1:T])

    x̄ = zeros(para.nx,N,T)
    z̄ = zeros(para.nz,N,T)
    z̄[1,:,:] .= [b0;b[1:T-1,:]]'
    x̄[1,:,:] .= b'
    x̄[2,:,:] .= c[1:T,:]'
    x̄[3,:,:] .= y[1:T,:]'
    x̄[4,:,:] .= uc'
    x̄[5,:,:] .= ucπ'

    X̄ = [θ zeros(T) Π[1:T] Z[1:T] Y Q W D Tr ZΠ[1:T]]
    Θ̄ = [θ0 θ[1:end-1]']
    Λ̄ = zeros(1,T)

    return (z̄=z̄,Θ̄=Θ̄,Λ̄=Λ̄,x̄=x̄,X̄=X̄')
  end

  """
      solveΠ(para,MC,Z)

  Solves for Π given marginal cost term, Z, and expectation term of Phillips curve
  """
  function solveΠ(para,MC,Z,ZΠ′)
    ψ = para.ψ
    β = para.β
    c = -(MC*Z+β*ψ*ZΠ′)/(ψ*Z)
    a = 1.
    b = 1.

    return (-b + sqrt(b^2-4*a*c))/2a
  end

  """
      solveΠZ(para,MC,Z)

  Solves for Π given marginal cost term, Z, and expectation term of Phillips curve
  """
  function solveΠZ(para,θ,Z′,Π′,C′)
    @unpack β,α,γ,τ,δ,ψ,Φ,B̄,Gov,ρ_θ,ϕ,θ̄ = para

    function ΠZres!(res,ΠlZ)
      Π,lZ = ΠlZ
      Z = exp(lZ)
      MC = (1 - Φ*(1 - α^(-1)*exp.(-(θ+θ̄)/α)*Z^((1-α)/α)))
      Y = Z-(Z/exp(θ+θ̄))^(1/α)
      C = Y - Gov - 0.5*ψ*Z*Π^2
      i = 1/β + ϕ * Π-1
      R = (1+i)/(1+Π′)

      res[1] = log(1/(β*R)) + γ*(C′-C)
      res[2] = MC + β*(Z′/Z)*ψ*Π′*(1+Π′) - ψ*Π*(1+Π)
    end
    ret = nlsolve(ΠZres!,[Π′,log(Z′)],ftol=1e-10)
    if (!ret.f_converged && !ret.x_converged)
      error("Could not find root")
    end
    return [ret.zero[1],exp(ret.zero[2])]
  end

  """
      getw̄(approx::Approximation,x̄,X̄)

  Computes for the vector of w̄  given x̄ and X̄
  """
  function getw̄(para,x̄,X̄)
    @unpack Izx,Ix′x,Ix_x,IΘX,IΛX,IX′X,nϵ,nΕ = para

    N = size(x̄,2)

    X̄vec = X̄.*ones(N)'

    return [Izx*x̄;x̄;Ix_x*x̄;Ix′x*x̄;IΘX*X̄vec;IΛX*X̄vec;X̄vec;IX′X*X̄vec;zeros(nϵ,N);zeros(nΕ,N)]
  end

  """
      getw̄(approx::Approximation,x̄,X̄)

  Solves for the vector of w̄  given x̄ and X̄
  """
  function getw̄(para,z̄,Θ̄,Λ̄,x̄,X̄,x̄′,X̄′)
      @unpack Izx,Ix′x,Ix_x,IΘX,IΛX,IX′X,nϵ,nΕ = para
      N = size(x̄,2)

      X̄vec = X̄.*ones(N)'

      return [z̄;x̄;Ix_x*x̄;x̄′;Θ̄.*ones(N)';Λ̄.*ones(N)';X̄vec;X̄′.*ones(N)';zeros(nϵ,N);zeros(nΕ,N)]
  end

  function testSS()
    para = CARAmodelNK()
    Ω = (reshape(para.B̄.+[-1;1],:,1),0.5*ones(2))
    x̄,X̄ = para.steadyState(Ω[1],Ω[2])
    w̄ = para.getw̄(x̄,X̄)
    ret = zeros(size(x̄,1))
    F(ret,w̄[:,1],para)
    println(ret)

    temp = zeros(size(X̄,1))
    ret = similar(temp)
    for i in 1:size(x̄,2)
      R(ret,w̄[:,i],para)
      temp .+=0.5*ret
    end
    println(temp)

    (z̄,Θ̄,Λ̄,x̄,X̄) = para.computePath(Ω,[-0.01],0.)
    k = 1
    w̄ = getw̄(para,z̄[:,:,k],Θ̄[:,k],Λ̄[:,k],x̄[:,:,k],X̄[:,k],para.Ix′x*x̄[:,:,k+1],para.IX′X*X̄[:,k+1])

    ret = zeros(size(x̄,1))
    F(ret,w̄[:,1],para)
    println(ret)

    temp = zeros(size(X̄,1))
    ret = similar(temp)
    for i in 1:size(x̄,2)
      R(ret,w̄[:,i],para)
      temp .+=0.5*ret
    end
    println(temp)
  end

  testSS()

  export CARAmodelNK
end  # module CalibrationCARA
