using Roots
using Parameters
using NLsolve

"""
    solvehyperbolic(A)

Solves the equation x/(1-x) = A
"""
function solvehyperbolic(A)
  f(x) = (1-x)*A-x

  return find_zero(f,0.5,xatol=1e-10,atol=1e-10)
end

"""
    findCARASteadyState(para)

Find the steady state associated with no aggregate shocks
"""
function findCARASteadyState(para)
  @unpack β,α,γ,τ,δ,ψ,Φ,B̄,Gov,ρ_θ,Σ_ϵ,θ̄ = para
  θ̄ = exp(θ̄)
  σ = sqrt(Σ_ϵ[1])
  z̄ = (α*θ̄^(1/α)*(Φ-1)/Φ)^(α/(1-α))
  Ȳ = z̄-(z̄/θ̄)^(1/α)
  π̄ = 0
  W̄ = (1-α)/α * (z̄/θ̄)^(1/α)
  D̄ = Ȳ-W̄
  σ2y = ((1-τ)^2*W̄^2 +δ^2*D̄^2)*σ^2

  f(R) = log(1/(R*β)) - 0.5*γ^2*((R-1)/R)^2*σ2y
  R̄ = find_zero(f,(1,1/β),xatol=1e-10,atol=1e-10)
  μ̄  = (R̄-1)/R̄

  T̄ = τ*W̄- Gov +(1/R̄-1)*B̄
  ȳ = (1-τ)*W̄ + T̄ + D̄
  C̄ = (1/R̄)*ȳ

  return (z̄,Ȳ,π̄,W̄,D̄,R̄,μ̄,T̄,ȳ,C̄)
end

"""
    findCARASteadyState(para)

Find the steady state associated with no aggregate shocks
"""
function findCARASteadyStateAlt(para)
  @unpack β,α,γ,τ,δ,ψ,ϕ,Φ,B̄,Gov,ρ_θ,Σ_ϵ,θ̄ = para
  σ = sqrt(Σ_ϵ[1])
  θ̄ = exp(θ̄)
  function residual!(res,x)
    z̄,π̄ = x
    ī = 1/β-1 + ϕ*π̄
    R̄ = (1+ī)/(1+π̄)
    Ȳ = z̄-(z̄/θ̄)^(1/α)
    W̄ = (1-α)/α * (z̄/θ̄)^(1/α)
    D̄ = Ȳ-W̄ - 0.5*ψ*π̄^2*z̄
    T̄ = τ*W̄ + 1/(1+ī)*B̄ - Gov - B̄/(1+π̄)
    Eȳ = (1-τ)*W̄ + T̄ + D̄
    σ2y = ((1-τ)^2*W̄^2 +δ^2*D̄^2)*σ^2

    μ̄ = (R̄-1)/R̄

    res[1] = log(1/(R̄*β)) - 0.5*γ^2*μ̄^2*σ2y
    res[2] = (1-Φ*( 1-α^(-1)*θ̄^(-1/α)*z̄^((1-α)/α))) + β*ψ*π̄*(1+π̄) - ψ*π̄*(1+π̄)
  end
  w̄ = findCARASteadyState(para)
  ret = nlsolve(residual!,[w̄[1],w̄[3]],ftol=1e-12,xtol=1e-12)
  if any(isnan.(ret.zero))
    throw(Exception("Could not find root"))
  end
  z̄,π̄ = ret.zero

  ī = 1/β-1 + ϕ*π̄
  R̄ = (1+ī)/(1+π̄)
  Ȳ = z̄-(z̄/θ̄)^(1/α)
  W̄ = (1-α)/α * (z̄/θ̄)^(1/α)
  D̄ = Ȳ-W̄ - 0.5*ψ*π̄^2*z̄
  T̄ = τ*W̄ + 1/(1+ī)*B̄ - Gov - B̄/(1+π̄)
  Eȳ = (1-τ)*W̄ + T̄ + D̄
  σ2y = ((1-τ)^2*W̄^2 +δ^2*D̄^2)*σ^2
  C̄ = (1/R̄)*Eȳ

  μ̄ = (R̄-1)/R̄

  return (z̄,Ȳ,π̄,W̄,D̄,R̄,μ̄,T̄,Eȳ,C̄)
end

"""
  closedFormSolution(para,Ω,Θ)

Given parameters para, distribution Ω and current aggregate productivity θ, compute
the current consumption and price policies.
"""
function closedFormSolution(para,Ω,θ_0,s_ϵ=1.,T=500)
  @unpack β,α,ϕ,γ,τ,δ,ψ,Φ,B̄,Gov,ρ_θ,Σ_ϵ,θ̄= para
  b = Ω[1]
  N = size(b,1)
  σ = sqrt(Σ_ϵ[1])
  ϵ = ones(N)*σ*s_ϵ

  θ = θ_0.*[ρ_θ^(t-1) for t in 1:T]
  z = zeros(T)
  μ = zeros(T)
  C = zeros(T)
  R = zeros(T)
  W = zeros(T)
  Tr = zeros(T)
  Y = zeros(T)
  πt = zeros(T)
  D = zeros(T)
  Ey = zeros(T)
  i = zeros(T)

  #find steady state
  z̄,Ȳ,π̄,W̄,D̄,R̄,μ̄,T̄r,ȳ,C̄ = findCARASteadyStateAlt(para)
  r̄ = 1/β-1
  #Work backwards from steady state
  z[end] = z̄
  μ[end] = μ̄
  C[end] = C̄
  R[end] = R̄
  W[end] = W̄
  Tr[end] = T̄r
  Y[end] = Ȳ
  πt[end] = π̄
  D[end] = D̄
  Ey[end] = ȳ
  i[end] = R̄*(1+π̄)-1
  for t in reverse(1:T-1)
    πt[t],z[t],i[t],R[t],μ[t],W[t],Y[t],D[t],Tr[t],Ey[t],C[t] = findcurrentperiodvalues(para,exp(θ[t]+θ̄),r̄,π̄,z̄,
                                                        μ[t+1],C[t+1],πt[t+1],z[t+1],W[t+1],D[t+1],Ey[t+1])
  end

  y = (1-τ).*W'.*(1 .+ ϵ) .+ Tr' .+ (1 .+ δ*ϵ)*D'
  c = μ'.* (y .+ b./(1 .+ πt')) .+ C'
  Q = 1 ./(1 .+ i)
  return πt,z,i,R,μ,W,Y,D,Tr,Ey,C,y,c
end


"""
  closedFormSolution(para,Ω,Θ)

Given parameters para, distribution Ω and current aggregate productivity θ, compute
the current consumption and price policies.
"""
function closedFormSolution2(para,Ω,θ_0,s_ϵ=1.,T=500)
  @unpack β,α,ϕ,γ,τ,δ,ψ,Φ,B̄,Gov,ρ_θ,Σ_ϵ,θ̄ = para
  b = Ω[1]
  N = size(b,1)
  σ = sqrt(Σ_ϵ[1])
  ϵ = ones(N)*σ*s_ϵ

  θ = θ_0.*[ρ_θ^(t-1) for t in 1:T]
  z = zeros(T)
  μ = zeros(T)
  C = zeros(T)
  R = zeros(T)
  W = zeros(T)
  Tr = zeros(T)
  Y = zeros(T)
  πt = zeros(T)
  D = zeros(T)
  Ey = zeros(T)
  i = zeros(T)

  #find steady state
  z̄,Ȳ,π̄,W̄,D̄,R̄,μ̄,T̄r,ȳ,C̄ = findCARASteadyStateAlt(para)
  r̄ = 1/β-1
  #Work backwards from steady state
  z[end] = z̄
  μ[end] = μ̄
  C[end] = C̄
  R[end] = R̄
  W[end] = W̄
  Tr[end] = T̄r
  Y[end] = Ȳ
  πt[end] = π̄
  D[end] = D̄
  Ey[end] = ȳ
  i[end] = R̄-1
  for t in reverse(1:T-1)
    πt[t],z[t],i[t],R[t],μ[t],W[t],Y[t],D[t],Tr[t],Ey[t],C[t] = findcurrentperiodvalues(para,exp(θ[t]+θ̄),r̄,π̄,z̄,
                                                        μ[t+1],C[t+1],πt[t+1],z[t+1],W[t+1],D[t+1],Ey[t+1])
  end

  y = (1-τ).*W'.*(1 .+ ϵ) .+ Tr' .+ (1 .+ δ*ϵ)*D'
  c = μ'.* (y .+ b./(1 .+ πt')) .+ C'
  Q = 1 ./(1 .+ i)

  μB = (1 .+ i).*( (1 .- μ).*((1-τ).*W + Tr + D) -  C)
  μBb = (1 .+ i).* (1 .- μ)./(1 .+ πt)
  μBϵ = (1 .+ i).* (1 .- μ).*((1-τ).*W .+ δ.*D)
  return πt,z,i,R,μ,W,Y,D,Tr,Ey,C,y,c,μB,μBb,μBϵ
end




"""
  closedFormIRF(para,Ω,Θ)

Given parameters para, distribution Ω and current aggregate productivity θ, compute
the current consumption and price policies.
"""
function closedFormIRF(para,Ω,θ_0,T,s_ϵ=1.)
  @unpack τ,δ = para

  πt,z,i,R,μ,W,Y,D,Tr,Ey,C,y,c = closedFormSolution(para,Ω,θ_0,s_ϵ)
  πt0,z0,i0,R0,μ0,W0,Y0,D0,Tr0,Ey0,C0,y0,c0 = closedFormSolution(para,Ω,0.,s_ϵ)


  return (πt[1:T]-πt0[1:T],z[1:T]-z0[1:T],i[1:T] -i0[1:T],R[1:T]-R0[1:T],μ[1:T]-μ0[1:T],W[1:T]-W0[1:T],Y[1:T]-Y0[1:T],
          D[1:T]-D0[1:T],Tr[1:T]-Tr0[1:T],Ey[1:T]-Ey0[1:T],C[1:T]-C0[1:T],y[:,1:T]-y0[:,1:T],c[:,1:T]-c0[:,1:T])
end


"""
  closedFormLinearize(para,Ω,Θ)

Given parameters para, distribution Ω and current aggregate productivity θ, compute
the current consumption and price policies.
"""
function closedFormLinearize(para,Ω,θ)
  dθ = 0.00001
  X = closedFormSolution2(para,Ω,dθ,0)
  X0 = closedFormSolution2(para,Ω,0.,0)

  X = map(x->(x[1][1]-x[2][1])/dθ*θ+x[2][1],zip(X,X0))

  return X
end


function findcurrentperiodvalues(para,θ,r̄,π̄,z̄,μ′,C′,π′,z′,W′,D′,Ey′)
  @unpack β,α,γ,τ,δ,ψ,ϕ,Φ,B̄,Gov,ρ_θ,Σ_ϵ = para
  σ = sqrt(Σ_ϵ[1])
  σ2y′ = (1-τ)^2*W′^2*σ^2 +δ^2*D′^2*σ^2
  function residual!(res,x)
    π̂,lnz = x
    z = exp(lnz)
    i = r̄ + ϕ * π̂
    R = (1+i)/(1+π′)
    μ = solvehyperbolic(R*μ′)
    W = (1-α)/α*(z/θ)^(1/α)
    Y = z-(z/θ)^(1/α)
    D = Y-W-0.5*ψ*π̂^2*z
    T = τ*W + 1/(1+i)*B̄ - Gov - B̄/(1+π̂)
    Ey = (1-τ)*W + T + D
    C = Y-0.5*ψ*π̂^2*z - Gov - μ*(Ey+B̄/(1+π̂))


    res[1] = log(1/(β*R)) + γ*(μ′*Ey′-μ′*R*C+ C′-C) - 0.5*γ^2*μ′^2*σ2y′
    res[2] = (1-Φ*( 1-α^(-1)*θ^(-1/α)*z^((1-α)/α))) + β*(z′/z)*ψ*π′*(1+π′) - ψ*π̂*(1+π̂)
  end

  ret = nlsolve(residual!,[π̄,log(z̄)],ftol=1e-12,xtol=1e-12)
  if (!ret.f_converged  && !ret.x_converged)
    error("Could not find root")
  end
  π,lnz = ret.zero
  z = exp(lnz)
  i = r̄ + ϕ * π
  R = (1+i)/(1+π′)
  μ = solvehyperbolic(R*μ′)
  W = (1-α)/α*(z/θ)^(1/α)
  Y = z-(z/θ)^(1/α)
  D = Y-W-0.5*ψ*π^2*z
  T = τ*W + 1/(1+i)*B̄ - Gov - B̄/(1+π)
  Ey = (1-τ)*W + T + D
  C = Y-0.5*ψ*π^2*z - Gov - μ*(Ey+B̄/(1+π))

  return π,z,i,R,μ,W,Y,D,T,Ey,C
end
