include("calibrationCARA_NK.jl")
include("approximationMPI.jl")
include("approximationPathMPI.jl")
include("closedform_NK.jl")
BLAS.set_num_threads(1)
MPI.Init()
using .CalibrationCARA_NK
using FastGaussQuadrature
using DataFrames
using Random
using Distributions
using Plots
using CSV
using DelimitedFiles

function getBounds(b′,ω,p,N)
  P = (ω .* reshape(p,1,:))[:]
  B = b′[:]

  I = sortperm(B)

  cumP = cumsum(P[I])
  i̲ = findfirst(cumP.>=1/N/50)
  i̅ = findlast(cumP .<= 1-1/N/50)

  b̲ = B[I][i̲]
  b̅ = B[I][i̅]
  bmin = sum(B[B.<b̲].*P[B.<b̲])/sum(P[B.<b̲])
  bmax = sum(B[B.>b̅].*P[B.>b̅])/sum(P[B.>b̅])

  return b̲,b̅,bmin,bmax
end

function getNextDistribution(Ω::AgentDist,b′,p)
  Ω,ω = Ω
  N = size(Ω,1)
  S = length(p)

  b̲,b̅,bmin,bmax = getBounds(b′,ω,p,N)

  Ω′ = reshape(collect(LinRange(b̲,b̅,N)),:,1)
  Ω′[1] = bmin
  Ω′[end] = bmax
  ω′ = zeros(N)

  for i in 1:N
    for s in 1:S
      if b′[i,s] < b̲
        ω′[1] += ω[i]*p[s]
      elseif b′[i,s] > b̅
        ω′[end] += ω[i]*p[s]
      else
        i′ = findfirst(x->x>=b′[i,s],Ω′[:])
        α = (b′[i,s] - Ω′[i′])/(Ω′[i′-1]-Ω′[i′])
        ω′[i′-1] += α*ω[i]*p[s]
        ω′[i′] += (1-α)*ω[i]*p[s]
      end
    end
  end

  return (Ω′,ω′)
end




function getPoliciesGlobal(para,Ω,θ,S=5)
  @unpack τ,δ = para
  x,w = gausshermite(S)
  σ = sqrt(para.Σ_ϵ[1])
  ϵ = sqrt(2)*σ.*x
  ϵ = reshape(ϵ,1,:)
  p = w/sqrt(π)
  b,ω = Ω

  πt,z,i,R,μ,W,Y,D,Tr,Ey,C,y,c,μB,μBb,μBϵ = closedFormLinearize(para,Ω,θ)

  y = (1-τ).*W.*(1 .+ ϵ) .+ Tr .+ (1 .+ δ*ϵ)*D
  c = μ.* (y .+ b./(1 .+ πt)) .+ C
  b′ = μB .+ μBb.*b .+ μBϵ.*ϵ
  return ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr
end

function getPoliciesGlobalFull(para,Ω,θ,S=5)
  @unpack τ,δ = para
  x,w = gausshermite(S)
  σ = sqrt(para.Σ_ϵ[1])
  ϵ = sqrt(2)*σ.*x
  ϵ = reshape(ϵ,1,:)
  p = w/sqrt(π)
  b,ω = Ω

  πt,z,i,R,μ,W,Y,D,Tr,Ey,C = map(x->x[1],closedFormSolution(para,Ω,θ))

  y = (1-τ).*W.*(1 .+ ϵ) .+ Tr .+ (1 .+ δ*ϵ)*D
  c = μ.* (y .+ b./(1 .+ πt)) .+ C
  b′ = (y .+ b./(1 .+ πt) .- c).*(1+i)
  return ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr
end


function doSimulationGlobal(para,Ω::AgentDist,θ0,T,σ_Ε = sqrt(para.Σ_Ε[1]))


  df = DataFrame(zeros(T,10),[:π,:Z,:i,:R,:W,:Y,:D,:Tr,:θ,:stdΩ])
  Ωt = Vector{AgentDist}(undef,T+1)
  θ = para.ρ_θ*θ0 + randn()*σ_Ε
  Ωt[1] = Ω
  for t in 1:T
    ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobal(para,Ωt[t],θ)
    Ωt[t+1] = getNextDistribution(Ωt[t],b′,p)
    #renormalize due to simulation error
    Ωt[t+1][1] .-= (dot(Ωt[t+1][1][:],Ωt[t+1][2]) - para.B̄)
    meanΩ=sum(Ωt[t+1][1].*Ωt[t+1][2]);
    stdΩ=(sum(Ωt[t+1][1].^2 .*Ωt[t+1][2])-meanΩ^2)^0.5;

    df.π[t],df.Z[t],df.i[t],df.R[t],df.W[t],df.Y[t],df.D[t],df.Tr[t] = πt,z,i,R,W,Y,D,Tr
    df.θ[t] = θ
    df.stdΩ[t] = stdΩ
    Ε = randn()*σ_Ε#[max(,-0.06-para.ρ_θ*θ[1] )]

    θ = para.ρ_θ*θ +Ε

  end

  return df,Ωt
end

function doSimulationGlobalFull(para,Ω::AgentDist,θ0,T)


  df = DataFrame(zeros(T,10),[:π,:Z,:i,:R,:W,:Y,:D,:Tr,:θ,:stdΩ])
  Ωt = Vector{AgentDist}(undef,T+1)
  θ = para.ρ_θ*θ0 
  Ωt[1] = Ω
  for t in 1:T
    ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobalFull(para,Ωt[t],θ)
    Ωt[t+1] = getNextDistribution(Ωt[t],b′,p)
    #renormalize due to simulation error
    Ωt[t+1][1] .-= (dot(Ωt[t+1][1][:],Ωt[t+1][2]) - para.B̄)
    meanΩ=sum(Ωt[t+1][1].*Ωt[t+1][2]);
    stdΩ=(sum(Ωt[t+1][1].^2 .*Ωt[t+1][2])-meanΩ^2)^0.5;
    df.π[t],df.Z[t],df.i[t],df.R[t],df.W[t],df.Y[t],df.D[t],df.Tr[t] = πt,z,i,R,W,Y,D,Tr
    df.θ[t] = θ
    df.stdΩ[t] = stdΩ
    θ = para.ρ_θ*θ

  end

  return df,Ωt
end


function getPoliciesLocal_linear!(approx,Ω,θ,Ε,S=5)
  para = approx.para
  @unpack τ,δ = para
  comm = MPI.COMM_WORLD

  x,w = gausshermite(S)
  σ = sqrt(para.Σ_ϵ[1])
  ϵ = sqrt(2)*σ.*x
  ϵ = reshape(ϵ,1,:)
  p = w/sqrt(π)
  N = size(Ω[1],1)

  approximate!(approx,Ω)
  my_N = size(approx.x̄,2)
  if MPI.Comm_rank(comm) == 0
    x = zeros(para.nx,N,S)
    X = zeros(para.nX)
  end

  ϵi = zeros(1,my_N)
  for s in 1:S
    ϵi .= ϵ[s]
    if MPI.Comm_rank(comm) == 0
      X,x[:,:,s] = simulatepolicies_linear(approx,θ,Ε,ϵi)
    else
      simulatepolicies_linear(approx,θ,Ε,ϵi)
    end
  end

  if MPI.Comm_rank(comm) == 0
    return p,x,X
  end
end


function doSimulationLocal_linear(para,Ω::AgentDist,θ0,T,σ_Ε = sqrt(para.Σ_Ε[1]))
  comm = MPI.COMM_WORLD
  Σ_Ε = copy(para.Σ_Ε)
  para.Σ_Ε  .= 0 #so they don't contribution to the $σ,σ$ term

  approx = Approximation(para=para)
  df = DataFrame(zeros(T,length(para.Xlab)+1),[para.Xlab;:stdΩ])
  Ωt = Vector{AgentDist}(undef,T+1)
  θ = [θ0]
  Ε = [randn()*σ_Ε]

  Ωt[1] = Ω
  for t in 1:T
    if MPI.Comm_rank(comm) == 0
      println("Linear $t")
      p,x,X = getPoliciesLocal_linear!(approx,Ωt[t],θ,Ε)
      b′ = x[1,:,:]
      Ωt[t+1] = getNextDistribution(Ωt[t],b′,p)
      #renormalize due to simulation error
      Ωt[t+1][1] .-= (dot(Ωt[t+1][1][:],Ωt[t+1][2]) - para.B̄)
      meanΩ=sum(Ωt[t+1][1].*Ωt[t+1][2]);
      stdΩ=(sum(Ωt[t+1][1].^2 .*Ωt[t+1][2])-meanΩ^2)^0.5;
      for (i,sX) in enumerate(para.Xlab)
        df[!,sX][t] = X[i]
      end
      df[!,:stdΩ][t] = stdΩ
      θ = para.IΘX*X
      Ε = [randn()*σ_Ε]#[max(,-0.06-para.ρ_θ*θ[1] )]
      Ωt[t+1],θ,Ε = MPI.bcast((Ωt[t+1],θ,Ε),0,comm)
    else
      getPoliciesLocal_linear!(approx,Ωt[t],θ,Ε)
      Ωt[t+1],θ,Ε = MPI.bcast(nothing,0,comm)
    end
  end
  para.Σ_Ε .= Σ_Ε

  return df,Ωt
end



function getPoliciesLocal!(approxPath,Ω,θ,Ε,S=5)
  comm = MPI.COMM_WORLD
  para = approxPath.para
  @unpack τ,δ = para
  x,w = gausshermite(S)
  σ = sqrt(para.Σ_ϵ[1])
  ϵ = sqrt(2)*σ.*x
  ϵ = reshape(ϵ,1,:)
  p = w/sqrt(π)
  N = size(Ω[1],1)

  approximate!(approxPath,Ω,θ,[0.],25,5)

  my_N = size(approxPath.approx[1].x̄,2)
  if MPI.Comm_rank(comm) == 0
    x = zeros(para.nx,N,S)
    X = zeros(para.nX)
  end

  ϵi = zeros(1,my_N)
  for s in 1:S
    ϵi .= ϵ[s]
    if MPI.Comm_rank(comm) == 0
      X,x[:,:,s] = simulatepolicies(approxPath,Ε,ϵi)
    else
      simulatepolicies(approxPath,Ε,ϵi)
    end
  end

  if MPI.Comm_rank(comm) == 0
    return p,x,X,ϵ
  end
end


function doSimulationLocal(para,Ω::AgentDist,θ0,T,σ_Ε = sqrt(para.Σ_Ε[1]))
  comm = MPI.COMM_WORLD
  approxPath = ApproximationPath{typeof(para)}(para=para)
  df = DataFrame(zeros(T,length(para.Xlab)+1),[para.Xlab;:stdΩ])
  Ωt = Vector{AgentDist}(undef,T+1)
  θ = [θ0]
  Ε = [randn()*σ_Ε]

  Ωt[1] = Ω
  for t in 1:T
    if MPI.Comm_rank(comm) == 0
      println(t)
      println("$θ,$Ε")
      p,x,X = getPoliciesLocal!(approxPath,Ωt[t],θ,Ε)
      b′ = x[1,:,:]
      Ωt[t+1] = getNextDistribution(Ωt[t],b′,p)
      #renormalize due to simulation error
      Ωt[t+1][1] .-= (dot(Ωt[t+1][1][:],Ωt[t+1][2]) - para.B̄)
      meanΩ=sum(Ωt[t+1][1].*Ωt[t+1][2]);
      stdΩ=(sum(Ωt[t+1][1].^2 .*Ωt[t+1][2])-meanΩ^2)^0.5;
      for (i,sX) in enumerate(para.Xlab)
        df[!,sX][t] = X[i]
      end
      df[!,:stdΩ][t] = stdΩ
      θ = para.IΘX*X
      Ε = [randn()*σ_Ε]#[max(randn()*σ_Ε,-0.06-para.ρ_θ*θ[1] )]
      println(Ε)
      println(θ)
      Ωt[t+1],θ,Ε = MPI.bcast((Ωt[t+1],θ,Ε),0,comm)
    else
      getPoliciesLocal!(approxPath,Ωt[t],θ,Ε)
      Ωt[t+1],θ,Ε = MPI.bcast(nothing,0,comm)
    end
  end

  return df,Ωt
end


function computeEEerrorsGlobalFull(para,Ω,θ,S=5)
  @unpack γ,ρ_θ,β = para
  ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobalFull(para,Ω,θ)
  p = reshape(p,1,:)

  Ω′ = (reshape(b′,:,1), (Ω[2].*p)[:]) #compute Ω′
  x,w = gausshermite(S)
  σ_Ε = sqrt(para.Σ_Ε[1])
  Ε′ = sqrt(2)*σ_Ε.*x
  p_Ε = w/sqrt(π)

  c′ = zeros(length(Ω′[2]),length(p),S)
  π′ = zeros(1,1,S)
  for s in 1:S
    θ′ = ρ_θ*θ + Ε′[s]
    c′[:,:,s],π′[s]= getPoliciesGlobalFull(para,Ω′,θ′)[4:5]
  end

  p_Ε = reshape(p_Ε,1,1,:)
  EUc′ = sum((1 .+ i)./(1. .+ π′).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

  ĉ = log.(β*EUc′)./(-γ)

  EE =  (c[:]-ĉ)./ĉ
  EEmax = norm(EE,Inf)
  EEmean = dot(EE[:],Ω′[2])
  return reshape(EE,:,S),EEmax,EEmean,c,p[:]
end





function computeEEerrorsGlobal(para,Ω,θ,S=5)
  @unpack γ,ρ_θ,β = para
  ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobal(para,Ω,θ)
  p = reshape(p,1,:)

  Ω′ = (reshape(b′,:,1), (Ω[2].*p)[:]) #compute Ω′
  x,w = gausshermite(S)
  σ_Ε = sqrt(para.Σ_Ε[1])
  Ε′ = sqrt(2)*σ_Ε.*x
  p_Ε = w/sqrt(π)

  c′ = zeros(length(Ω′[2]),length(p),S)
  π′ = zeros(1,1,S)
  for s in 1:S
    θ′ = ρ_θ*θ + Ε′[s]
    c′[:,:,s],π′[s]= getPoliciesGlobal(para,Ω′,θ′)[4:5]
  end

  p_Ε = reshape(p_Ε,1,1,:)
  EUc′ = sum((1 .+ i)./(1. .+ π′).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

  ĉ = log.(β*EUc′)./(-γ)

  EE =  (c[:]-ĉ)./Y#ĉ
  EEmax = norm(EE,Inf)
  EEmean = dot(EE[:],Ω′[2])
  return reshape(EE,:,S),EEmax,EEmean,c
end

function computeDynamicEEerrorsGlobal(para,b0,Ωt,θt,dfG,S=5)
  @unpack γ,ρ_θ,β,τ = para
  T = length(θt)-1
  N = length(b0)
  Ω = ([b0;Ωt[1][1]],[zeros(N);Ωt[1][2]])
  Ω̂ = Ω
  bt,b̂t,ct,ĉt = zeros(length(b0),T),zeros(length(b0),T),zeros(length(b0),T),zeros(length(b0),T)
  EE = zeros(N,T)
  for t in 1:T
    ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobal(para,Ω,θt[t])
    distp = Categorical(p)
    s_i = rand(distp,N)
    bt[:,t] .= [b′[i,s_i[i]] for i in 1:N]
    ct[:,t] .= [c[i,s_i[i]] for i in 1:N]

    ϵ,p,b′,c,πt,z,i,R,μ,W,Y,D,Tr = getPoliciesGlobal(para,Ω̂,θt[t])
    p = reshape(p,1,:)
    b′ = [b′[i,s_i[i]] for i in 1:N]
    Ω′ = ([b′;Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])
    x,w = gausshermite(S)
    σ_Ε = sqrt(para.Σ_Ε[1])
    Ε′ = sqrt(2)*σ_Ε.*x
    p_Ε = w/sqrt(π)

    c′ = zeros(N,length(p),S)
    π′ = zeros(1,1,S)
    for s in 1:S
      θ′ = ρ_θ*θt[t] + Ε′[s]
      ctemp,π′[s] = getPoliciesGlobal(para,Ω′,θ′)[4:5]
      c′[:,:,s] = ctemp[1:N,:]
    end

    p_Ε = reshape(p_Ε,1,1,:)
    EUc′ = sum((1 .+ i)./(1. .+ π′).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

    ĉt[:,t] = log.(β*EUc′)./(-γ)
    b_ = Ω̂[1][1:N]
    Q = 1/(1+dfG.i[t])
    y = (1-τ).*dfG.W[t].*(1 .+ ϵ[s_i]) .+ dfG.Tr[t] .+ dfG.D[t]
    b̂t[:,t] = (b_./(1+dfG.π[t]) .+ y .- ĉt[:,t])./Q

    Ω̂ = ([b̂t[:,t];Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])
    Ω = ([bt[:,t];Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])

    EE[:,t] .=  (ct[:,t]-ĉt[:,t])./Y
  end
  EEmax = norm(EE,Inf)
  return EE,EEmax,ct,ĉt
end


function computeEEerrorsLocal_linear(approx,Ω,θ,Ε,S=5)
  comm = MPI.COMM_WORLD
  para = approx.para
  @unpack γ,ρ_θ,β = para
  if MPI.Comm_rank(comm) == 0
    p,x,X = getPoliciesLocal_linear!(approx,Ω,[θ],[Ε])
    p = reshape(p,1,:)
    b′ = x[1,:,:]
    c = x[2,:,:]
    Q = X[6]

    Ω′ = (reshape(b′,:,1), (Ω[2].*p)[:]) #compute Ω′
    x,w = gausshermite(S)
    σ_Ε = sqrt(para.Σ_Ε[1])
    Ε′ = sqrt(2)*σ_Ε.*x
    p_Ε = w/sqrt(π)

    c′ = zeros(length(Ω′[2]),length(p),S)
    π′ = zeros(1,1,S)
    θ′ = ρ_θ*θ + Ε
    MPI.bcast((Ω′,θ′,Ε′),0,comm)
    for s in 1:S
      p′,x′,X′ = getPoliciesLocal_linear!(approx,Ω′,[θ′],[Ε′[s]])
      c′[:,:,s] = x′[2,:,:]
      π′[s] = X′[3]
    end

    p_Ε = reshape(p_Ε,1,1,:)
    EUc′ = sum(1 ./(Q*(1. .+ π′)).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

    ĉ = log.(β*EUc′)./(-γ)

    EE = (c[:]-ĉ)./ĉ
    EEmax = norm(EE,Inf)
    EEmean = dot(EE[:],Ω′[2])
    return reshape(EE,:,S),EEmax,EEmean,c
  else
    getPoliciesLocal_linear!(approx,Ω,[θ],[Ε])
    Ω′,θ′,Ε′ = MPI.bcast(nothing,0,comm)
    for s in 1:S
      getPoliciesLocal_linear!(approx,Ω′,[θ′],[Ε′[s]])
    end
    return nothing,nothing,nothing,nothing
  end
end

function computeEEerrorsLocal(approxPath,Ω,θ,Ε,S=5)
  comm = MPI.COMM_WORLD
  para = approxPath.para
  θ,Ε = MPI.bcast((θ,Ε),0,comm)
  if MPI.Comm_rank(comm) == 0
    @unpack γ,ρ_θ,β = para
    p,x,X = getPoliciesLocal!(approxPath,Ω,[θ],[Ε])
    p = reshape(p,1,:)
    b′ = x[1,:,:]
    c = x[2,:,:]
    Q = X[6]

    Ω′ = (reshape(b′,:,1), (Ω[2].*p)[:]) #compute Ω′
    x,w = gausshermite(S)
    σ_Ε = sqrt(para.Σ_Ε[1])
    Ε′ = sqrt(2)*σ_Ε.*x
    p_Ε = w/sqrt(π)

    c′ = zeros(length(Ω′[2]),length(p),S)
    π′ = zeros(1,1,S)
    θ′ = ρ_θ*θ + Ε

    MPI.bcast((Ω′,θ′,Ε′),0,comm)
    for s in 1:S
      p′,x′,X′ = getPoliciesLocal!(approxPath,Ω′,[θ′],[Ε′[s]])
      c′[:,:,s] = x′[2,:,:]
      π′[s] = X′[3]
    end

    p_Ε = reshape(p_Ε,1,1,:)
    EUc′ = sum(1 ./(Q*(1. .+ π′)).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

    ĉ = log.(β*EUc′)./(-γ)

    EE =  (c[:]-ĉ)./X[5]#ĉ
    EEmax = norm(EE,Inf)
    EEmean = dot(EE[:],Ω′[2])
    return reshape(EE,:,S),EEmax,EEmean,c
  else
    getPoliciesLocal!(approxPath,Ω,[θ],[Ε])
    Ω′,θ′,Ε′ = MPI.bcast(nothing,0,comm)
    for s in 1:S
      getPoliciesLocal!(approxPath,Ω′,[θ′],[Ε′[s]])
    end
    return nothing,nothing,nothing,nothing
  end
end

function computeDynamicEEerrorsLocal(approxPath,b0,Ωt,θt,dfG,S=5)
  comm = MPI.COMM_WORLD
  para = approxPath.para
  θt = MPI.bcast(θt,0,comm)
  T = length(θt)-1
  if MPI.Comm_rank(comm) == 0
    @unpack γ,ρ_θ,β,τ = para

    N = length(b0)
    Ω = ([b0;Ωt[1][1]],[zeros(N);Ωt[1][2]])
    Ω̂ = Ω
    bt,b̂t,ct,ĉt = zeros(length(b0),T),zeros(length(b0),T),zeros(length(b0),T),zeros(length(b0),T)
    EE = zeros(N,T)
    for t in 1:T
      println(t)
      MPI.bcast((Ω,Ω̂),0,comm)
      p,x,X = getPoliciesLocal!(approxPath,Ω,[θt[t]],[0.])
      distp = Categorical(p)
      s_i = rand(distp,N)
      b′ = x[1,:,:]
      c = x[2,:,:]
      bt[:,t] .= [b′[i,s_i[i]] for i in 1:N]
      ct[:,t] .= [c[i,s_i[i]] for i in 1:N]
      p,x,X,ϵ = getPoliciesLocal!(approxPath,Ω̂,[θt[t]],[0.])
      b′ = [x[1,i,s_i[i]] for i in 1:N]
      Q = X[6]

      p = reshape(p,1,:)

      Ω′ = ([b′;Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])
      x,w = gausshermite(S)
      σ_Ε = sqrt(para.Σ_Ε[1])
      Ε′ = sqrt(2)*σ_Ε.*x
      p_Ε = w/sqrt(π)

      c′ = zeros(N,length(p),S)
      π′ = zeros(1,1,S)
      θ′ = ρ_θ*θt[t]

      MPI.bcast((Ω′,θ′,Ε′),0,comm)
      for s in 1:S
        p′,x′,X′ = getPoliciesLocal!(approxPath,Ω′,[θ′],[Ε′[s]])
        c′[:,:,s] = x′[2,1:N,:]
        π′[s] = X′[3]
      end
      p_Ε = reshape(p_Ε,1,1,:)
      EUc′ = sum(1 ./(Q*(1. .+ π′)).*exp.(-γ.*c′).*p.*p_Ε,dims=[2,3])

      ĉt[:,t] = log.(β*EUc′)./(-γ)
      b_ = Ω̂[1][1:N]
      Q = 1/(1+dfG.i[t])
      y = (1-τ).*dfG.W[t].*(1 .+ ϵ[s_i]) .+ dfG.Tr[t] .+ dfG.D[t]
      b̂t[:,t] = (b_./(1+dfG.π[t]) .+ y .- ĉt[:,t])./Q

      Ω̂ = ([b̂t[:,t];Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])
      Ω = ([bt[:,t];Ωt[t+1][1]],[zeros(N);Ωt[t+1][2]])

      EE[:,t] .=  (ct[:,t]-ĉt[:,t])./X[5]
      println(norm(EE,Inf))
    end
    EEmax = norm(EE,Inf)
    return EE,EEmax,ct,ĉt
  else
    for t in 1:T
      Ω,Ω̂ = MPI.bcast(nothing,0,comm)
      getPoliciesLocal!(approxPath,Ω,[θt[t]],[0.])
      getPoliciesLocal!(approxPath,Ω̂,[θt[t]],[0.])
      Ω′,θ′,Ε′ = MPI.bcast(nothing,0,comm)
      for s in 1:S
        getPoliciesLocal!(approxPath,Ω′,[θ′],[Ε′[s]])
      end
    end
    return nothing,nothing,nothing,nothing
  end
end



function getDistributionDataFrame(Ω,Ȳ)
  bs = Ω[1]./Ȳ
  ω = Ω[2]./(bs[3]-bs[2])

  return DataFrame(b=bs[:],pdf=ω[:])
end

function plotDistribution(Ω,Ȳ,label)
  bs = Ω[1]./Ȳ
  ω = Ω[2]./(bs[3]-bs[2])

  scatter!(bs,ω,label=label)

  return DataFrame(b=bs[:],pdf=ω[:])
end



### Code to run Comparisons

rank = MPI.Comm_rank(MPI.COMM_WORLD)
size = MPI.Comm_size(MPI.COMM_WORLD)
println(rank)
para = CARAmodelNK()
para.Σ_ϵ .= 0.5^2
para.Σ_Ε .= 0.
para.γ = 1.
para.β = 0.96
para.B̄ = 0.7
para.α = 0.6
para.ψ = 41.6

##Computing Euler Errors along 
for sd in [0.5,0.75,1.,-1.]
  if sd > 0.
    para.Σ_ϵ .= sd^2
    Figurepath = "Figures/sigma$(Int(sd*100))/"
    Datapath = "Data/sigma$(Int(sd*100))/"
  else
    para.Σ_ϵ .= 0.5^2
    para.γ = 3.
    Figurepath = "Figures/gamma3/"
    Datapath = "Data/gamma3/"
  end

  #make the saving directories
  if rank == 0
    try mkdir(Datapath) catch e  end
    try mkdir(Figurepath) catch e  end
  end
  ##### Plots for Anmol

  approx = Approximation(para=para)
  approxPath = ApproximationPath{typeof(para)}(para=para)


  Ω = reshape(collect(LinRange(-6. + para.B̄,6. + para.B̄,150)),:,1)
  ω = pdf.(Normal(para.B̄,4),Ω[:])
  ω ./= sum(ω)
  approximate!(approx,(Ω,ω))
  Ȳ = approx.X̄[5]
  Z̄ = approx.X̄[4]

  println(approx.X̄[7]/approx.X̄[5])

  T = 250
  N_IRF = 25
  shock = 0.0123
  #Construct long simulation of 250 periods and compute Euler Errors
  dfG,ΩGt = doSimulationGlobalFull(para,(Ω,ω),shock,T)
  dfG0,ΩGt0 = doSimulationGlobalFull(para,(Ω,ω),0.,T)
  dfGl,ΩGlt = doSimulationGlobal(para,(Ω,ω),shock,T)
  dfLl,ΩLlt = doSimulationLocal_linear(para,(Ω,ω),shock,T)
  dfL,ΩLt = doSimulationLocal(para,(Ω,ω),shock,T)
  dfL0,ΩLt0 = doSimulationLocal(para,(Ω,ω),0.,T)
  b0 = LinRange(-6. + para.B̄,6. + para.B̄,size)

  #Save data from long simulations
  if rank==0
    Ȳ = dfG.Y[end]

    dfIRF = DataFrame(piG=dfG.π*100,piGl=dfGl.π*100,piL=dfL.Π*100,
                      iG =dfG.i*100,iGl =dfGl.i*100,iL=(1 ./dfL.Q.-1)*100,
                      yG =(dfG.Y.-Ȳ)./Ȳ*100,yGl =(dfGl.Y.-Ȳ)./Ȳ*100,yL=(dfL.Y.-Ȳ)./Ȳ*100)

    CSV.write(Datapath*"IRFLongdata.csv",first(dfIRF,50))
    dfL.i = (1 ./dfL.Q.-1)
    dfError = DataFrame(YGl=norm((dfGl.Y.-dfG.Y)./dfG.Y,Inf)*100,
                        piGl=norm(dfGl.π.-dfG.π,Inf)*100,
                        iGl=norm(dfGl.i .- dfG.i,Inf)*100,
                        YL=norm((dfL.Y.-dfG.Y)./dfG.Y,Inf)*100,
                        piL=norm(dfL.Π .-dfG.π,Inf)*100,
                        iL=norm(dfL.i .- dfG.i,Inf)*100)


    dfΩG = getDistributionDataFrame(ΩGt[T],Ȳ)
    dfΩGl = getDistributionDataFrame(ΩGlt[T],Ȳ)
    dfΩL = getDistributionDataFrame(ΩLt[T],Ȳ)
    dfΩ = DataFrame(bG=dfΩG.b,pdfG=dfΩG.pdf,bGl=dfΩGl.b,pdfGl=dfΩGl.pdf,bL=dfΩL.b,pdfL=dfΩL.pdf)          
    CSV.write(Datapath*"Distribution250.csv",dfΩ)
    CSV.write(Datapath*"AggregateErrors.csv",dfError)
  end

  rank==0 && println("Doing EE")
  if rank == 0
    dfError = DataFrame(zeros(1,7),[:EpolGl,:EpolL,:EEGl,:EEL,:DynGl,:DynL,:DynL0])
  end
  for t in [2,10,20,50]
    EEG,EEmaxG,EEavgG,cG,p = computeEEerrorsGlobalFull(para,ΩGt[t],dfG.θ[t])
    EEGl,EEmaxGl,EEavgGl,cGl = computeEEerrorsGlobal(para,ΩGt[t],dfG.θ[t])
    EEL,EEmaxL,EEavgL,cL = computeEEerrorsLocal(approxPath,ΩGt[t],dfG.θ[t-1],0.)
    
    if rank == 0
      EcG = ((cG.-cG)./Ȳ)
      EcGl =((cGl.-cG)./Ȳ)
      EcL = ((cL.-cG)./Ȳ)
      dfError_t = DataFrame(b = ΩGt[t][1][:]./Ȳ, EpolG = maximum(abs.(EcG),dims=2)[:]*100, EpolGl = maximum(abs.(EcGl),dims=2)[:]*100,
                          EpolL = maximum(abs.(EcL),dims=2)[:]*100,EpolGavg = abs.(EcG*p)[:]*100,EpolGlavg = abs.(EcGl*p)[:]*100,
                          EpolLavg=abs.(EcL*p)[:]*100,EEG = maximum(abs.(EEG),dims=2)[:]*100,EEGl=maximum(abs.(EEGl),dims=2)[:]*100,
                          EEL=maximum(abs.(EEL),dims=2)[:]*100,EEGavg = abs.(EEG*p)[:]*100,EEGlavg = abs.(EEGl*p)[:]*100,EELavg = abs.(EEL*p)[:]*100)

      dfError.EpolGl[1] = max(dfError.EpolGl[1],norm(EcGl,Inf)*100)
      dfError.EpolL[1] = max(dfError.EpolL[1],norm(EcL,Inf)*100)
      dfError.EEGl[1] = max(dfError.EEGl[1],norm(EEGl,Inf)*100)
      dfError.EEL[1] = max(dfError.EEL[1],norm(EEL,Inf)*100)

      CSV.write(Datapath*"Errors$t.csv",dfError_t)
    end
  end

  #do dynamic EE
  rank==0 && println("Doing Dynamic EE")
  EEGl,EEmaxGl,cGl,ĉGl = computeDynamicEEerrorsGlobal(para,b0,ΩGlt,dfG.θ,dfG,1)
  θt = [shock;dfL.θ[1:end-1]]
  EEL,EEmaxL,cL,ĉL = computeDynamicEEerrorsLocal(approxPath,b0,ΩLt,θt,dfG,1)
  EEL0,EEmaxL0,cL0,ĉL0 = computeDynamicEEerrorsLocal(approxPath,b0,ΩLt,θt*0,dfG0,1)
  if rank==0
    writedlm(Datapath*"DynEEGl.csv",EEGl*100,",")
    writedlm(Datapath*"DynEEL.csv",EEL*100,",")
    writedlm(Datapath*"DynEEL0.csv",EEL0*100,",")
    dfError.DynGl[1] = EEmaxGl*100
    dfError.DynL[1] = EEmaxL*100
    dfError.DynL0[1] = EEmaxL0*100
    CSV.write(Datapath*"IndividualErrors.csv",dfError)
  end



  #Next do IRF
  rank==0 && println("Doing IRFs")
  for t in [1,T]
    dfG1,ΩGt1 = doSimulationGlobalFull(para,ΩGt[t],shock,N_IRF)
    dfL1,ΩLt1 = doSimulationLocal(para,ΩLt[t],shock,N_IRF)
    dfGl1,ΩGlt1 = doSimulationGlobal(para,ΩGt[1],shock,N_IRF)


    dfG0,ΩGt0 = doSimulationGlobalFull(para,ΩGt[t],0.0,N_IRF)
    dfL0,ΩLt0 = doSimulationLocal(para,ΩLt[t],0.00,N_IRF)
    dfGl0,ΩGlt0 = doSimulationGlobal(para,ΩGt[1],0.00,N_IRF)


    if rank ==0

      dfIRF = DataFrame(piG=(dfG1.π-dfG0.π)*100,piGl=(dfGl1.π-dfGl0.π)*100,piL=(dfL1.Π-dfL0.Π)*100,
                        iG =(dfG1.i-dfG0.i)*100,iGl =(dfGl1.i-dfGl0.i)*100,iL=(1 ./dfL1.Q.-1 .- (1 ./dfL0.Q.-1))*100,
                        yG =(dfG1.Y.-dfG0.Y)./Ȳ*100,yGl =(dfGl1.Y .-dfGl0.Y)./Ȳ*100,yL=(dfL1.Y.-dfL0.Y)./Ȳ*100,
                        ZG =(dfG1.Z.-dfG0.Z)./Z̄*100,ZGl =(dfGl1.Z .-dfGl0.Z)./Z̄*100,ZL=(dfL1.Z.-dfL0.Z)./Z̄*100,
                        stdG = (dfG1.stdΩ.-dfG0.stdΩ)./Ȳ*100,stdGl = (dfGl1.stdΩ.-dfGl0.stdΩ)./Ȳ*100,stdL = (dfL1.stdΩ-dfL0.stdΩ)./Ȳ*100)

      CSV.write(Datapath*"IRF$t.csv",dfIRF)
    end
  end
end


approx = Approximation(para=para)
approxPath = ApproximationPath{typeof(para)}(para=para)


Ω = reshape(collect(LinRange(-6. + para.B̄,6. + para.B̄,150)),:,1)
ω = pdf.(Normal(para.B̄,4),Ω[:])
ω ./= sum(ω)
approximate!(approx,(Ω,ω))
Ȳ = approx.X̄[5]
Z̄ = approx.X̄[4]


T = 200
sd = 0.5
Figurepath = "Figures/sigma$(Int(sd*100))/"
Datapath = "Data/sigma$(Int(sd*100))/"
para.γ = 1.
para.Σ_ϵ .= sd^2
para.Σ_Ε .= 0.0123^2
Random.seed!(5235623856)
dfGl,ΩGlt = doSimulationGlobal(para,(Ω,ω),0.,T)
Random.seed!(5235623856)
dfLl,ΩLlt = doSimulationLocal_linear(para,(Ω,ω),0.,T)
Random.seed!(5235623856)
dfL,ΩLt = doSimulationLocal(para,(Ω,ω),0.,T)
if rank==0
  dfΩGl = getDistributionDataFrame(ΩGlt[T],Ȳ)
  dfΩL = getDistributionDataFrame(ΩLt[T],Ȳ)
  dfΩLl = getDistributionDataFrame(ΩLlt[T],Ȳ)

  dfΩ = DataFrame(bL=dfΩL.b,pdfL=dfΩL.pdf,bGl=dfΩGl.b,pdfGl=dfΩGl.pdf,bLl=dfΩLl.b,pdfLl=dfΩLl.pdf)
  CSV.write(Datapath*"Distribution_aggshocks.csv",dfΩ)
end

