using LinearAlgebra
using TensorOperations

import Base.*,LinearAlgebra.⋅

function *(A::Matrix{T1},B::Array{T2,3}) where {T1<:Real,T2<:Real}
    return @tensor C[i,k,l] := A[i,j]*B[j,k,l]
end

function *(A::Matrix{T1},B::Array{T2,3}) where {T1<:Real,T2<:Real}
    return @tensor C[i,k,l] := A[i,j]*B[j,k,l]
end

function *(A::Matrix{T1},B::Array{T2,3}) where {T1<:Real,T2<:Real}
    return @tensor C[i,k,l] := A[i,j]*B[j,k,l]
end

function *(A::Matrix{T1},B::Array{T2,3}) where {T1<:Real,T2<:Real}
    return @tensor C[i,k,l] := A[i,j]*B[j,k,l]
end

function ⋅(A::Array{T1,3},Σ::Matrix{T2})  where {T1<:Real,T2<:Real}
    return @tensor C[i] := A[i,j,k]*Σ[j,k]
end

function ⋅(A::Array{T1,3},B::Tuple{Matrix{T2},Matrix{T3}}) where {T1<:Real,T2<:Real,T3<:Real}
    return @tensor C[i,l,m] := A[i,j,k]*B[1][j,l]*B[2][k,m]
end

function ⋅(A::Array{T1,3},B::Tuple{UniformScaling{Bool},Matrix{T2}}) where {T1<:Real,T2<:Real}
    return @tensor C[i,j,m] := A[i,j,k]*B[2][k,m]
end

function ⋅(A::Array{T1,3},B::Tuple{Matrix{T2},UniformScaling{Bool}}) where {T1<:Real,T2<:Real}
    return @tensor C[i,m,k] := A[i,j,k]*B[1][j,m]
end

function ⋅(A::Array{T,3},B::Tuple{UniformScaling{Bool},UniformScaling{Bool}}) where {T<:Real}
    return A
end

function ⋅(A::Array{T1,3},B::Tuple{Vector{T2},Vector{T3}}) where {T1<:Real,T2<:Real,T3<:Real}
    return @tensor C[i] := A[i,j,k]*B[1][j]*B[2][k]
end

function doDottedSum!(A::Array{T1,3},B::Tuple{Matrix{T2},Matrix{T3}},C::Array{T4,3}) where {T1<:Real,T2<:Real,T3<:Real,T4<:Real}
    @tensor C[i,l,m] += A[i,j,k]*B[1][j,l]*B[2][k,m]
end

function doDottedSum!(A::Array{T1,3},B::Tuple{UniformScaling{Bool},Matrix{T2}},C::Array{T3,3}) where {T1<:Real,T2<:Real,T3<:Real}
    @tensor C[i,j,m] += A[i,j,k]*B[2][k,m]
end

function doDottedSum!(A::Array{T1,3},B::Tuple{Matrix{T2},UniformScaling{Bool}},C::Array{T3,3}) where {T1<:Real,T2<:Real,T3<:Real}
    @tensor C[i,m,k] += A[i,j,k]*B[1][j,m]
end

function doDottedSum!(A::Array{T1,3},B::Tuple{UniformScaling{Bool},UniformScaling{Bool}},C::Array{T2,3}) where {T1<:Real,T2<:Real}
    C .+= A
end

macro doinplace(block)
    if block.head != :.+= || block.args[2].head != :call || block.args[2].args[1] != :⋅
        error("Inputed wrong expression.  Needs to be C .+= A⋅(B1,B2)")
    end
    Fcall = :(doDottedSum!(A,B,C))
    Fcall.args[2] = block.args[2].args[2]
    Fcall.args[3] = block.args[2].args[3]
    Fcall.args[4] = block.args[1]
    return esc(Fcall)
end

"""
    MPI_slice(r::Int,s::Int,N::Int)

Returns the slice of a range 1:N  for the process of rank r in a group of size s
"""
function MPI_slice(r::Int,s::Int,N::Int)
    N_s = div(N,s)
    N_mod_s = rem(N,s)
    if r < N_mod_s
        r*(N_s+1)+1:(r+1)*(N_s+1)
    else
        (N_s+1)*N_mod_s+(r-N_mod_s)*N_s +1:(N_s+1)*N_mod_s+(r+1-N_mod_s)*N_s
    end
end


"""
    MPI_index(i::Int,r::Int,s::Int,N::Int)

Returns the associated index (in the space of the individual process) of index i,
for the process of rank r in a group of size s.  Returns -1 if not on that space
"""
function MPI_index(i,r,s,N)
    N_s = div(N,s)
    N_mod_s = rem(N,s)
    if r < N_mod_s
        if r*(N_s+1) < i <= (r+1)*(N_s+1)
            return i - r*(N_s+1)
        else
            return -1
        end
    else
        (N_s+1)*N_mod_s+(r-N_mod_s)*N_s +1:(N_s+1)*N_mod_s+(r+1-N_mod_s)*N_s
        if (N_s+1)*N_mod_s+(r-N_mod_s)*N_s < i <= (N_s+1)*N_mod_s+(r+1-N_mod_s)*N_s
            return i - ((N_s+1)*N_mod_s+(r-N_mod_s)*N_s)
        else
            return -1
        end
    end
end


"""
    MPI_GatherArray(my_x,comm,root=0)

Gathers an array from across multiple process, even when processes have different
size.  Assumes the arrays only differ in the last dimension and are then horizontally stacked
"""
function MPI_GatherArray(my_X::Array,comm,root=0)
    r = MPI.Comm_rank(comm)
    lenX = Vector{Int32}(MPI.Allgather(length(my_X),comm))

    X = MPI.Gatherv(my_X,lenX,root,comm)

    if r == 0
        dims = (size(my_X)[1:end-1]...,:)
        return reshape(X,dims) #reshape to keep last dimensions the same as my_X
    else
        return my_X
    end
end
