# -*- coding: utf-8 -*-
"""
Created on Wed Mar  2 17:13:56 2016

@author: dgevans
"""
import numpy as np
from copy import deepcopy
from scipy.optimize import brentq
from scipy.interpolate import interp1d
from scipy.optimize import minimize
from scipy.optimize import root
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()



def find_decreasing_root(f,x):
    '''
    Finds the root of f given guess x.  Assume f is decreasing
    '''
    if f(x)>0:
        x_up = x+1
        while f(x_up)>0:
            x = x_up
            x_up += 1.
        x0 = x
        x1 = x_up
    else:
        x_down = x - 1.
        while f(x_down) <0:
            x = x_down
            x_down -= 1.
        x0 = x_down
        x1 = x
    return brentq(f,x0,x1)

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    def __getattr__(self, attr):
         return self.get(attr)
    __setattr__= dict.__setitem__
    __delattr__= dict.__delitem__
    def __deepcopy__(self, memo):
        return dotdict(deepcopy(dict(self)))

class NearestDict(dict):
    '''
    For dictionary with getting the nearest element
    '''
    def __init__(self, ndims):
        super(NearestDict, self).__init__()
        self.ndims = ndims

    # Enforce dimensionality
    def __setitem__(self, key, val):
        if not isinstance(key, tuple): key = (key,)
        if len(key) != self.ndims: raise KeyError("key must be %d dimensions" % self.ndims)
        super(NearestDict, self).__setitem__(key, val)

    @staticmethod
    def __dist(ka, kb):
        assert len(ka) == len(kb)
        return sum((ea-eb)**2 for (ea, eb) in zip(ka, kb))

    # Helper method and might be of use
    def nearest_key(self, key):
        if not isinstance(key, tuple): key = (key,)
        nk = min((k for k in self), key=lambda k: NearestDict.__dist(key, k))
        return nk

    def __missing__(self, key):
        if not isinstance(key, tuple): key = (key,)
        if len(key) != self.ndims: raise KeyError("key must be %d dimensions" % self.ndims)
        return self[self.nearest_key(key)]


class EquilibriumResiduals(dotdict):
    '''
    Computes the welfare and residual of equilibrium for a tau and T for a given 
    distriubtion of wages.
    '''
    def __init__(self,Para,log_wages,shares,bonds):
        '''
        Initializes for parameters and a given distriubtion of log wages
        '''
        super(dotdict,self).__init__(deepcopy(Para))
        N = len(log_wages)
        my_n = N // size
        if my_n*size != N:
            raise Exception("size of mpi must divide equally into N")
            
        self.log_wages = log_wages
        self.shares = shares
        self.bonds = bonds
        self.my_log_wages = log_wages[rank*my_n:(rank+1)*my_n]
        self.my_wages = np.exp(self.my_log_wages)
        self.my_shares = shares[rank*my_n:(rank+1)*my_n]
        self.my_bonds = bonds[rank*my_n:(rank+1)*my_n]
        self.logl = NearestDict(2)
        self.offset = np.zeros(4)                              
         #initialize loglf
        self.logl[-100000,-100000] = np.zeros(my_n)
        
        
    def res_l(self,logl,w,s,b,z):
        '''
        Computes residual of agent's optimality constraint
        '''
        T,Y = z
        beta,gamma,sigma,varphi,epsilon= self.beta,self.gamma,self.ra, self.varphi, self.epsilon
        Tau_w,Tau_d,Tau_b = self.Tau_w,self.Tau_d,self.Tau_b
        W = (epsilon-1.)/epsilon #<ok>
        D = (1-W)*Y #<ok>
        l = np.exp(logl) #<ok>
        c = (l**gamma/((1-Tau_w)*W*w))**(-1./sigma)
        return  (c - T - (1-Tau_d)*s*D - (1-Tau_w)*W*w*l)/(1-beta) - b*(1-Tau_b)
        
    def interpolate_logl(self,z):
        '''
        Interpolate for a given policy z
        '''
        #first check to see if already interpolated
        z_tup = tuple(z)
        if z_tup in self.logl:
            return self.logl[z_tup]
        logl0 = self.logl[z_tup]
        
        n = len(self.my_wages)
        logl_vec = np.zeros(n)
        for i in range(n):
            logl_vec[i] = find_decreasing_root(lambda logl:self.res_l(logl,self.my_wages[i],self.my_shares[i]
                                        ,self.my_bonds[i],z),logl0[i])
        
        
        self.logl[z_tup] = logl_vec
        return logl_vec
    
    def getAllocation(self,z):
        '''
        Computes allocation given current 
        '''
        T,Y = z
        beta,gamma,sigma,varphi,epsilon= self.beta,self.gamma,self.ra, self.varphi, self.epsilon
        Tau_w,Tau_d,Tau_b = self.Tau_w,self.Tau_d,self.Tau_b
        logl = self.interpolate_logl(z)
        w = self.my_wages
        W = (epsilon-1.)/epsilon #<ok>
        D = (1-W)*Y #<ok>
        s = self.my_shares
        b = self.my_bonds
        
        
        l = np.exp(logl)
        c = (l**gamma/((1-Tau_w)*W*w))**(-1./sigma)
        
        
        return w,b,s,c,l
    
    def getMultipliers(self,x,z):
        beta,gamma,sigma,varphi,epsilon= self.beta,self.gamma,self.ra, self.varphi, self.epsilon
        Tau_w,Tau_d,Tau_b = self.Tau_w,self.Tau_d,self.Tau_b
        w,b,s,c,l = self.getAllocation(z)
        T,Y = z
        Alpha_w,Alpha_s,Alpha_b,Xi = x
        
        W = (epsilon-1.)/epsilon #<ok>
        D = (1-W)*Y #<ok>
        uc = c**(-sigma)
        ucc = -sigma*c**(-sigma-1)
        ul = -l**gamma
        ull = -gamma*l**(gamma-1)
        omega = ((w)**Alpha_w)+np.exp(Alpha_s*s)+np.exp(Alpha_b*b) 
        omega = np.maximum(np.minimum(omega,10000),0.001)
        my_omega_mean = np.array([np.mean(omega)])
        omega_mean = np.zeros(1)
        comm.Allreduce(my_omega_mean,omega_mean)
        omega_mean /= size
        omega /= omega_mean
        
        
        #first compute multipliers
        a_c = omega*uc - Xi
        a_cmu = -ucc*(c-T-(1-Tau_d)*s*D) - uc + (1-beta)*ucc*(1-Tau_b)*b
        a_cphi = ucc*(1-Tau_w)*w*W
        a_l = omega*ul + w*Xi
        a_lmu = -(ull*l+ul)
        a_lphi = ull
        
        mu = -(a_c - a_cphi/a_lphi * a_l)/(a_cmu - a_cphi/a_lphi * a_lmu)
        phi = -(a_l + a_lmu*mu)/a_lphi
        
        return s,w,b,uc,mu,phi,omega
        
        
    def equilibriumFOCs(self,z):
        '''
        Computes resource at a given tau
        '''
        T,Y = z
        w,b,s,c,l = self.getAllocation(z)
        
        
        #Now do constraints
        my_resids = np.array([
                            (l*w - Y).mean(),
                            (Y - c - self.g).mean()
                            ])
        resids = np.zeros(2)
        comm.Allreduce(my_resids,resids)
        return resids/size
    
    def utilitarianFOCs(self,y):
        T,Y,Tau_w,Tau_d,Tau_b,Xi = y
        self.Tau_w,self.Tau_d,self.Tau_b = Tau_w,Tau_d,Tau_b
        x = np.array([0.,0.,0.,Xi])
        z = np.array([T,Y])
        self.logl = NearestDict(2)
        self.logl[-100000,-100000] = np.zeros(len(self.shares))
        w,b,s,c,l = self.getAllocation(z)
        s,w,b,uc,mu,phi,omega = self.getMultipliers(x,z)
        
        #Now do constraints
        my_resids = np.array([
                            (l*w - Y).mean(),
                            (Y - c - self.g).mean(),
                            (uc*b*mu).mean(),
                            (uc*s*mu).mean(),
                            (uc*w*phi).mean(),
                            (uc*mu).mean()
                            ])
        resids = np.zeros(6)
        comm.Allreduce(my_resids,resids)
        return resids/size
        
    
    def paretoWeightFOCs(self,x,z):
        s,w,b,uc,mu,phi,omega = self.getMultipliers(x,z)
        my_resids = np.array([
                            (uc*b*mu).mean(),
                            (uc*s*mu).mean(),
                            (uc*w*phi).mean(),
                            (uc*mu).mean()
                            ])
        resids = np.zeros(4)
        comm.Allreduce(my_resids,resids)
        return resids/size
        
    
    
def computeUtilitarian(ER,y0):
    fres = lambda y: ER.utilitarianFOCs(y)
    res = root(fres,y0)
    if not res.success:
        raise Exception('Could not find root!')
    else:
        return res.x
    
        
def computeEqulibrium(ER,z0):
    fres = lambda z: ER.equilibriumFOCs(z)
    res = root(fres,z0)
    if not res.success:
        raise Exception('Could not find root!')
    else:
        return res.x
    

def computeParetoWeights(ER,x0,z):
    fres = lambda x: ER.paretoWeightFOCs(x,z)
    res = root(fres,x0)
    if not res.success:
        raise Exception('Could not find root!')
    else:
        return res.x
    
    
def optimal_policy(Para,log_wages,shares,bonds,x0= [1.24320762, 0.17518274, 0.03165968, 0.91513418]):
    '''
    Finds the optimal policy given parameters and log_wages
    '''
    print rank
    ER = EquilibriumResiduals(Para,log_wages,shares,bonds)

    s = shares
    b = bonds

    z = computeEqulibrium(ER,np.ones(2))
    my_logl = ER.interpolate_logl(z)
    x = computeParetoWeights(ER,x0,z)
    T,Y = z
    Alpha_w,Alpha_s,Alpha_b,Xi = x
            
    if rank == 0:
        print("Alpha_w: " + str(Alpha_w))
        print("Alpha_s: " + str(Alpha_s))
        print("Alpha_b: " + str(Alpha_b))
        logl = np.hstack(comm.gather(my_logl))
        beta,gamma,sigma,varphi,epsilon = ER.beta,ER.gamma,ER.ra, ER.varphi, ER.epsilon
        Tau_w,Tau_d,Tau_b = ER.Tau_w,ER.Tau_d,ER.Tau_b
        W = (epsilon-1.)/epsilon #<ok>
        D = (1-W)*Y
        w = np.exp(ER.log_wages)
        
        l = np.exp(logl)
        c = (l**gamma/((1-Tau_w)*W*w))**(-1./sigma)
        uc = c**(-sigma)
        ucc = -sigma*c**(-sigma-1)
        ul = -l**gamma
        ull = -gamma*l**(gamma-1)
        omega = ((w)**Alpha_w)+np.exp(Alpha_s*s)+np.exp(Alpha_b*b) 
        omega = np.maximum(np.minimum(omega,10000),0.0001)
        omega = omega / np.mean(omega)
        
        #first compute multipliers
        a_c = omega*uc - Xi
        a_cmu = -ucc*(c-T-(1-Tau_d)*s*D) - uc + (1-beta)*ucc*(1-Tau_b)*b
        a_cphi = ucc*(1-Tau_w)*w*W
        a_l = omega*ul + w*Xi
        a_lmu = -(ull*l+ul)
        a_lphi = ull
        
        mu = -(a_c - a_cphi/a_lphi * a_l)/(a_cmu - a_cphi/a_lphi * a_lmu)
        phi = -(a_l + a_lmu*mu)/a_lphi
        
        print("check")
        print(np.linalg.norm(omega*uc-ucc*(c-T-(1-Tau_d)*s*D)*mu - uc*mu + ucc*(1-Tau_w)*w*W*phi - Xi + (1-beta)*ucc*(1-Tau_b)*b*mu))
        print(np.linalg.norm(omega*ul-(ull*l+ul)*mu +ull*phi + w*Xi))
        
        returns = np.hstack((z)),np.hstack((x)),c,l,b,omega,T,Tau_w,Tau_d,Tau_b,Y,mu
        comm.bcast(returns)
        return returns
    else:
        comm.gather(my_logl)
        returns = None
        returns = comm.bcast(returns)
        return returns
        

    
def optimal_utilitarian_policy(Para,log_wages,shares,bonds,y0 = np.array([1.,1.,0.65,1.2,1.2,0.9])):
    '''
    Finds the optimal policy given parameters and log_wages
    '''
    print rank
    ER = EquilibriumResiduals(Para,log_wages,shares,bonds)

    s = shares
    b = bonds

    y = computeUtilitarian(ER,y0)
    T,Y,Tau_w,Tau_d,Tau_b,Xi = y
    ER.Tau_w,ER.Tau_d,ER.Tau_b = Tau_w,Tau_d,Tau_b
    x = np.array([0.,0.,0.,Xi])
    z = np.array([T,Y])
    
    my_logl = ER.interpolate_logl(z)
    Alpha_w,Alpha_s,Alpha_b,Xi = x
            
    if rank == 0:
        print("Alpha_w: " + str(Alpha_w))
        print("Alpha_s: " + str(Alpha_s))
        print("Alpha_b: " + str(Alpha_b))
        logl = np.hstack(comm.gather(my_logl))
        beta,gamma,sigma,varphi,epsilon = ER.beta,ER.gamma,ER.ra, ER.varphi, ER.epsilon
        Tau_w,Tau_d,Tau_b = ER.Tau_w,ER.Tau_d,ER.Tau_b
        W = (epsilon-1.)/epsilon #<ok>
        D = (1-W)*Y
        w = np.exp(ER.log_wages)
        
        l = np.exp(logl)
        c = (l**gamma/((1-Tau_w)*W*w))**(-1./sigma)
        uc = c**(-sigma)
        ucc = -sigma*c**(-sigma-1)
        ul = -l**gamma
        ull = -gamma*l**(gamma-1)
        omega = ((w)**Alpha_w)+np.exp(Alpha_s*s)+np.exp(Alpha_b*b) 
        omega = np.maximum(np.minimum(omega,10000),0.0001)
        omega = omega / np.mean(omega)
        
        #first compute multipliers
        a_c = omega*uc - Xi
        a_cmu = -ucc*(c-T-(1-Tau_d)*s*D) - uc + (1-beta)*ucc*(1-Tau_b)*b
        a_cphi = ucc*(1-Tau_w)*w*W
        a_l = omega*ul + w*Xi
        a_lmu = -(ull*l+ul)
        a_lphi = ull
        
        mu = -(a_c - a_cphi/a_lphi * a_l)/(a_cmu - a_cphi/a_lphi * a_lmu)
        phi = -(a_l + a_lmu*mu)/a_lphi
        
        print("check")
        print(np.linalg.norm(omega*uc-ucc*(c-T-(1-Tau_d)*s*D)*mu - uc*mu + ucc*(1-Tau_w)*w*W*phi - Xi + (1-beta)*ucc*(1-Tau_b)*b*mu))
        print(np.linalg.norm(omega*ul-(ull*l+ul)*mu +ull*phi + w*Xi))
        
        returns = np.hstack((z)),np.hstack((x)),c,l,b,omega,T,Tau_w,Tau_d,Tau_b,Y,mu
        comm.bcast(returns)
        return returns
    else:
        comm.gather(my_logl)
        returns = None
        returns = comm.bcast(returns)
        return returns